;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Z80 Abersoft fig-FORTH recompiled
;; Copyright (C) 2024 Ketmar Dark // Invisible Vector
;; Understanding is not required. Only obedience.
;; GPLv3 ONLY
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; various math and compare primitives
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


<zx-definitions>

;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; bitwise logic

;; can be used instead of "DROP 0"
code: 0AND  ( a -- 0 )
  pop   hl
  ld    hl, # 0
;code-push-hl

code: AND  ( a b -- a&b )
  pop   de
  pop   hl
  ld    a, e
  and   l
  ld    l, a
  ld    a, d
  and   h
  ld    h, a
;code-push-hl

1 [IF]
code: ~AND  ( a b -- a&~b )
  pop   de
  pop   hl
  ld    a, e
  cpl
  and   l
  ld    l, a
  ld    a, d
  cpl
  and   h
  ld    h, a
;code-push-hl
[ELSE]
: ~AND  -1 XOR AND ;
[ENDIF]

code: OR  ( a b -- a|b )
  pop   de
  pop   hl
  ld    a, e
  or    l
  ld    l, a
  ld    a, d
  or    h
  ld    h, a
;code-push-hl

code: XOR  ( a b -- a|b )
  pop   de
  pop   hl
  ld    a, e
  xor   l
  ld    l, a
  ld    a, d
  xor   h
  ld    h, a
;code-push-hl


code: LO-BYTE  ( u -- u>>8 )
  pop   hl
  ld    h, # 0
;code-push-hl

code: HI-BYTE  ( u -- u>>8 )
  pop   hl
  ld    l, h
  ld    h, # 0
;code-push-hl


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; add, sub

code: +  ( a b -- a+b )
  pop   de
  pop   hl
  add   hl, de
;code-push-hl

code: -  ( a b -- a-b )
  pop   de
  pop   hl
  or    a
  sbc   hl, de
;code-push-hl

code: D+  ( da db -- da+db )
  ld    hl, # 6
  add   hl, sp
  ld    e, (hl)
  ld    (hl), c
  inc   hl
  ld    d, (hl)
  ld    (hl), b
  pop   bc
  pop   hl
  add   hl, de
  ex    de, hl
  pop   hl
  adc   hl, bc
  pop   bc
;code-push-de-hl


code: 1+  ( a -- a+1 )
  pop   hl
  inc   hl
;code-push-hl

code: 2+  ( a -- a+2 )
  pop   hl
  inc   hl
  inc   hl
;code-push-hl

(* 8 bytes
code: 4+  ( a -- a+4 )
  pop   hl
  ld    de, # 4
  add   hl, de
;code-push-hl
*)
;; 7 bytes
\ : 4+  4 + ;

code: 1-  ( a -- a-1 )
  pop   hl
  dec   hl
;code-push-hl

code: 2-  ( a -- a-2 )
  pop   hl
  dec   hl
  dec   hl
;code-push-hl

(* 8 bytes
code: 4-  ( a -- a-4 )
  pop   hl
  ld    de, # -4
  add   hl, de
;code-push-hl
*)
;; 7 bytes
\ : 4-  4 - ;


code: BSWAP  ( n -- n-nswapped )
  pop   de
  ld    h, e
  ld    l, d
;code-push-hl


code: 2*  ( n -- n*2 )
  pop   hl
  add   hl, hl  ;; *2
;code-push-hl

code: 4*  ( n -- n*4 )
  pop   hl
zx-word-umul4-do:
  add   hl, hl  ;; *2
  add   hl, hl  ;; *4
;code-push-hl

code: 8*  ( n -- n*8 )
  pop   hl
  add   hl, hl  ;; *2
  jr    # zx-word-umul4-do
;code-no-next

code: 16*  ( n -- n*16 )
  pop   hl
zx-word-umul16-do:
  add   hl, hl  ;; *2
  add   hl, hl  ;; *4
  jr    # zx-word-umul4-do
;code-no-next

code: 32*  ( n -- n*16 )
  pop   hl
  add   hl, hl  ;; *2
  jr    # zx-word-umul16-do
;code-no-next

code: 256*  ( n -- n*256 )
  pop   hl
  ld    h, l
  ld    l, # 0
;code-push-hl


zx-code: 2U/  ( n -- n/2 )
  pop   hl
  srl h  rr l   ;; /2u
;code-push-hl

zx-code: 4U/  ( n -- n/4 )
  pop   hl
  srl h  rr l   ;; /2u
  srl h  rr l   ;; /4u
;code-push-hl

zx-code: 8U/  ( n -- n/8 )
  pop   hl
zx-word-8udiv-do:
  srl h  rr l   ;; /2u
  srl h  rr l   ;; /4u
  srl h  rr l   ;; /8u
;code-push-hl

zx-code: 16U/  ( n -- n/16 )
  pop   hl
zx-word-16udiv-do:
  srl h  rr l   ;; /2u
  jr    # zx-word-8udiv-do
;code-no-next

zx-code: 32U/  ( n -- n/32 )
  pop   hl
  srl h  rr l   ;; /2u
  jr    # zx-word-16udiv-do
;code-no-next


zx-code: 2/  ( n -- n/2 )
  pop   hl
  sra h  rr l   ;; /2
;code-push-hl


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; negate

code: MINUS  ( a -- negated-a )
  pop   de
  ld    hl, # 0
  or    a
  sbc   hl, de
;code-push-hl

alias-for MINUS is NEGATE  ( a -- negated-a )


code: DMINUS  ( da -- negated-da )
  pop   hl
  pop   de
  sub   a
  sub   e
  ld    e, a
  ld    a, # 0
  sbc   a, d
  ld    d, a
  ld    a, # 0
  sbc   a, l
  ld    l, a
  ld    a, # 0
  sbc   a, h
  ld    h, a
;code-push-de-hl

alias-for DMINUS is DNEGATE  ( a -- negated-a )


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; C>S, S->D

code: C>S  ( signed-byte -- n )
  pop   hl
  ld    a, l
  rla
  ld    a, # 0
  sbc   a, a
  ld    h, a
;code-push-hl

code: S->D  ( n -- d )
  pop   de
  push  de
  xor   a
  rl    d
  sbc   a, a
  ld    l, a
  ld    h, a
;code-push-hl

alias-for S->D is S>D


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; comparisons with 0

code: 0=  ( a -- a=0? )
  pop   hl
  ld    a, l
  or    h
  ld    hl, # 1
  jp    z, # i-next-push-hl
  dec   l
;code-push-hl

alias-for 0= is NOT  ( a -- a=0? )

code: 0<>  ( a -- a<>0? )
  pop   hl
  ld    a, l
  or    h
  jp    z, # i-next-push-hl
  ld    hl, # 1
;code-push-hl

code: 0<  ( a -- a<0? )
  pop   hl
  xor   a
  rl    h
  ld    h, a
  adc   a, a
  ld    l, a
;code-push-hl

code: 0>  ( a -- a>0? )
  pop   hl
  ld    a, l
  or    h
  jp    z, # i-next-push-hl
  ;; 0>=
zx-word-do-0-gequ:
  xor   a
  rl    h
  ld    h, a
  sbc   a, # $FF
  ld    l, a
;code-push-hl

;; smaller than Forth version (and faster)
code: 0>=  ( a -- a>=0? )
  pop   hl
  jr    # zx-word-do-0-gequ
;code-no-next

: 0<=  ( a -- a<=0? )  0> NOT ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; numeric comparisons

code: 0MAX  ( a -- a // 0 )
  pop   hl
  bit   7, h
  jp    p, # i-next-push-hl
  ld    hl, # 0
;code-push-hl


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; numeric comparisons

code: =  ( a b -- a=b? )
  pop   hl
  pop   de
  xor   a
  sbc   hl, de
  ld    h, a
  ld    l, a
  jp    nz, # i-next-push-hl
  inc   l
;code-push-hl

code: <>  ( a b -- a<>b? )
  pop   hl
  pop   de
  xor   a
  sbc   hl, de
  jp    z, # i-next-push-hl
  ld    hl, # 1
;code-push-hl

\ zx: <>  ( a b -- a<>b? )  - 0<> ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; signed numeric comparisons

code: <  ( a b -- a<b? )
  pop   de
  pop   hl
zx-word-do-less:
  ld    a, d
  xor   h
  jp    m, # .different-sign
  sbc   hl, de
.different-sign:
  xor   a
  rl    h
  ld    h, a
  rla
  ld    l, a
;code-push-hl

code: >  ( a b -- a>b? )
  pop   hl
  pop   de
  jp    # zx-word-do-less
;code-no-next

: <=  ( a b -- a<=b? )  > NOT ;
: >=  ( a b -- a>=b? )  < NOT ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; unsigned numeric comparisons

code: U<  ( ua ub -- ua<ub? )
  pop   de
  pop   hl
zx-word-uless-rest:
  xor   a
  sbc   hl, de
  ld    h, a
  adc   a, a
  ld    l, a
;code-push-hl

code: U>  ( ua ub -- ua>ub? )
  pop   hl
  pop   de
  jp    # zx-word-uless-rest
;code-no-next

: U<=  ( ua ub -- ua<=ub? )  U> NOT ;
: U>=  ( ua ub -- ua>=ub? )  U< NOT ;


: WITHIN  ( test low high -- flag )  OVER - >R - R> U< ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; some useful utils
OPT-UNDER-MATH? [IF]
code: UNDER+  ( a b c -- a+c b )
  pop   de
  pop   af
  ex    af, afx
  pop   hl
  add   hl, de
  ex    af, afx
  push  hl
  push  af
;code

code: UNDER+1  ( a b -- a+1 b )
  pop   hl
  pop   de
  inc   de
;code-push-de-hl

code: UNDER-1  ( a b -- a+1 b )
  pop   hl
  pop   de
  dec   de
;code-push-de-hl

: UNDER-  ( a b c -- a-c b )  NEGATE UNDER+ ;
[ENDIF]


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; shifts

OPT-SHIFT-WORDS? [IF]
code: LSHIFT  ( a count -- a<<count )
  pop   de
  pop   hl
  ld    a, e
  and   # 15
  jp    z, # i-next-push-hl
.loop:
  add   hl, hl
  dec   a
  jp    nz, # .loop
;code-push-hl

code: RSHIFT  ( a count -- a>>count )
  pop   de
  pop   hl
zx-word-do-rshift:
  ld    a, e
  and   # 15
  jp    z, # i-next-push-hl
.loop:
  srl   h
  rr    l
  dec   a
  jp    nz, # .loop
;code-push-hl

code: ARSHIFT  ( a count -- a>>count )
  pop   de
  pop   hl
  ld    a, e
  and   # 15
  jp    z, # i-next-push-hl
.loop:
  sra   h
  rr    l
  dec   a
  jp    nz, # .loop
;code-push-hl

code: ROL  ( a count -- a<r<count )
  pop   de
  pop   hl
  ld    a, e
zx-code-rol-do:
  and   # 15
  jp    z, # i-next-push-hl
.loop:
  add   hl, hl
  jr    nc, # .no-inc-l
  inc   l
.no-inc-l:
  dec   a
  jp    nz, # .loop
;code-push-hl

: ROR  ( a count -- a>r>count )  NEGATE ROL ;

code: ROL8  ( a count -- a<r<count )
  pop   de
  pop   hl
  ld    a, e
zx-code-rol8-do:
  and   # 7
  jp    z, # i-next-push-hl
  ld    e, a
  ld    a, l
.loop:
  rlca
  dec   e
  jp    nz, # .loop
  ld    l, a
;code-push-hl

: ROR8  ( a count -- a>r>count )  NEGATE ROL8 ;
[ENDIF]


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; generate a random number
;; period of 2^32-1, and passes most of the diehard tests
;; this is the preferred PRNG, it is fast, small and good
;;
;; IN:
;;   xte_rnd_seed0, xte_rnd_seed1: 32 bits of the initial seed
;;   WARNING! initial seed must not be zero!
;; OUT:
;;   HL: 16-bit random
;;    A: weak 8-bit random (the same as L); use `add a, h` to get better result
;;   DE: dead
;;   flags: CARRY reset, others dead
;;
;;  taken from http://www.worldofspectrum.org/forums/showthread.php?t=23070
;;  original code by Patrik Rak (2012, based on Einar Saukas version)
;;
code: URND  ( -- u16 )
  call  # xte-rnd
  next-push-hl
xte-rnd:
  ld    hl, # $A29A ;; yw -> zt
$here 2- @def: xte-rnd-seed0
  ld    de, # $C0DE ;; xz -> yw
$here 2- @def: xte-rnd-seed1
  ld    xte-rnd-seed1 (), hl  ;; x = y, z = w
  ld    a, l        ;; w = w^(w<<3)
  add   a, a
  add   a, a
  add   a, a
  xor   l
  ld    l, a
  ld    a, d        ;; t = x^(x<<1)
  add   a, a
  xor   d
  ld    h, a
  rra               ;; t = t^(t>>1)^w
  xor   h
  xor   l
  ld    h, e        ;; y = z
  ld    l, a        ;; w = t
  ld    xte-rnd-seed0 (), hl
  ret
;code-no-next

code: URND8  ( -- u8 )
  call  # xte-rnd
  add   a, h
  ld    l, a
  ld    h, # 0
;code-push-hl

;; two seed addresses
asm-label: xte-rnd-seed0 constant URND-S0
asm-label: xte-rnd-seed1 constant URND-S1


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; generate random number
;; 16-bit xorshift pseudorandom number generator by John Metcalf
;; generates 16-bit pseudorandom numbers with a period of 65535
;; using the xorshift method:
;;
;; hl ^= hl << 7
;; hl ^= hl >> 9
;; hl ^= hl << 8
;;
;; some alternative shift triplets which also perform well are:
;; 6, 7, 13; 7, 9, 13; 9, 7, 13.
;;
;; IN:
;;   rndSeed: 2 bytes of shit (must not be 0)
;; OUT:
;;   HL: 16-bit random
;;   AF: dead
;; WARNING:
;;  be careful to not have all zeroes in rndSeed!
;;
;; ketmar's note: this one can be used to "dissolve" (and "ensolve")
;; the screen; see "samples/rndtest"
code: URNDN  ( -- u16 )
random:
  ld    hl, # $C0DE     ;; seed must not be 0
$here 2- @def: rndn-seed
  ld    a, h
  rra
  ld    a, l
  rra
  xor   h
  ld    h, a
  ld    a, l
  rra
  ld    a, h
  rra
  xor   l
  ld    l, a
  xor   h
  ld    h, a
  ld    rndn-seed (), hl
;code-push-hl

asm-label: rndn-seed constant URNDN-S0


<zx-done>
