;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Z80 Abersoft fig-FORTH recompiled
;; Copyright (C) 2024 Ketmar Dark // Invisible Vector
;; Understanding is not required. Only obedience.
;; GPLv3 ONLY
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; multiply and divide primitives
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


<zx-definitions>

code: U*  ( u1 u2 -- ud )
  pop   de
  pop   hl
  push  bc
  ld    b, h
  ld    a, l
  call  # .sub_606dh
  push  hl
  ld    h, a
  ld    a, b
  ld    b, h
  call  # .sub_606dh
  pop   de
  ld    c, d
  add   hl, bc
  adc   a, # 0
  ld    d, l
  ld    l, h
  ld    h, a
  pop   bc
  next-push-de-hl
.sub_606dh:
  ld    hl, # 0
  ld    c, # 8
.loop:
  add   hl, hl
  rla
  jr    nc, # .skip-add
  add   hl, de
  adc   a, # 0
.skip-add:
  dec   c
  jp    nz, # .loop
  ret
;code-no-next

code: U/MOD  ( ud u1 -- u-rem u-quot )
  ld    hl, # 4
  add   hl, sp
  ld    e, (hl)
  ld    (hl), c
  inc   hl
  ld    d, (hl)
  ld    (hl), b
  pop   bc
  pop   hl
  ld    a, l
  sub   c
  ld    a, h
  sbc   a, b
  jr    c, # .l60a0h
  ld    hl, # -1
  ld    de, hl
  jr    # .l60c0h
.l60a0h:
  ld    a, # 16
.l60a2h:
  add   hl, hl
  rla
  ex    de, hl
  add   hl, hl
  jr    nc, # .skip-inc
  inc   de
  or    a
.skip-inc:
  ex    de, hl
  rra
  push  af
  jr    nc, # .l60b4h
  and   l
  sbc   hl, bc
  jr    # .l60bbh
.l60b4h:
  or    a
  sbc   hl, bc
  jr    nc, # .l60bbh
  add   hl, bc
  dec   de
.l60bbh:
  inc   de
  pop   af
  dec   a
  jr    nz, # .l60a2h
.l60c0h:
  pop   bc
  push  hl
  push  de
;code


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; faster mul and div

OPT-16BIT-MUL/DIV? [IF]

;; fast 16-bit unsigned multiply.
;; uses AF'.
code: UU* ( ua ub -- ua*ub )
  pop   de
  pop   hl
  call  # vmx-umul-hl-de
  jp    # i-next-push-hl

vmx-umul-hl-de:
  ld    a, h
  or    a
  jr    z, # .umul-hl-8
  ;; HL is 16 bit, check DE
  ld    a, d
  or    a
  jr    z, # .umul-de-8

  ld    a, l
  or    a
  jr    z, # .umul-hl-l0

  ld    a, e
  or    a
  jr    z, # .umul-de-e0

  ;; both values are 16 bit
  push  bc
;; Multiply 16-bit values (with 16-bit result)
;; In: Multiply HL with DE
;; Out: HL = result
.umul-16-16-16:
  ld    a, h
  ld    c, l
  call  # .umul-16-16-32-unroll-8
  call  # .umul-16-16-32-unroll-8
  pop   bc
  \ ex    de, hl
  \ jp    # vmf-prim-next
  ret

.umul-de-8:
  ex    de, hl
.umul-hl-8:
  ld    a, l
  ld    l, h      ;; H is 0 here
  \ call  # .umul-8-16-16
  \ ret

;; Multiply 8-bit value with a 16-bit value
;; In: Multiply A with DE
;; Out: HL = result
.umul-8-16-16:
 OPT-16BIT-MUL/DIV-UNROLLED? [IF]
  ;; 1
  add   hl, hl
  add   a, a
  jr    nc, # $ 3 +
  add   hl, de
  ;; 2
  add   hl, hl
  add   a, a
  jr    nc, # $ 3 +
  add   hl, de
  ;; 3
  add   hl, hl
  add   a, a
  jr    nc, # $ 3 +
  add   hl, de
  ;; 4
  add   hl, hl
  add   a, a
  jr    nc, # $ 3 +
  add   hl, de
  ;; 5
  add   hl, hl
  add   a, a
  jr    nc, # $ 3 +
  add   hl, de
  ;; 6
  add   hl, hl
  add   a, a
  jr    nc, # $ 3 +
  add   hl, de
  ;; 7
  add   hl, hl
  add   a, a
  jr    nc, # $ 3 +
  add   hl, de
  ;; 8
  add   hl, hl
  add   a, a
  ret   nc
  add   hl, de
 [ELSE]
  push  bc
  ld    b, # 7
.loop-div8:
  add   hl, hl
  add   a, a
  jr    nc, # $ 3 +
  add   hl, de
  djnz  # .loop-div8
  pop   bc
  ;; 8
  add   hl, hl
  add   a, a
  ret   nc
  add   hl, de
 [ENDIF]
  ret

.umul-de-e0:
  ex    de, hl
.umul-hl-l0:
  ld    a, h
  ld    h, l    ;; L is 0 here
  call  # .umul-8-16-16
  ld    h, l
  ld    l, # 0
  ret

.umul-16-16-32-unroll-8:
 OPT-16BIT-MUL/DIV-UNROLLED? [IF]
  ;; 1
  add   hl, hl
  sla   c
  rla
  jr    nc, # $ 3 +
  add   hl, de
  ;; 2
  add   hl, hl
  sla   c
  rla
  jr    nc, # $ 3 +
  add   hl, de
  ;; 3
  add   hl, hl
  sla   c
  rla
  jr    nc, # $ 3 +
  add   hl, de
  ;; 4
  add   hl, hl
  sla   c
  rla
  jr    nc, # $ 3 +
  add   hl, de
  ;; 5
  add   hl, hl
  sla   c
  rla
  jr    nc, # $ 3 +
  add   hl, de
  ;; 6
  add   hl, hl
  sla   c
  rla
  jr    nc, # $ 3 +
  add   hl, de
  ;; 7
  add   hl, hl
  sla   c
  rla
  jr    nc, # $ 3 +
  add   hl, de
  ;; 8
  add   hl, hl
  sla   c
  rla
  ret   nc
  add   hl, de
 [ELSE]
  ex    af, afx
  ld    a, b
  ex    af, afx
  ld    b, # 7
.loop-div16:
  add   hl, hl
  sla   c
  rla
  jr    nc, # $ 3 +
  add   hl, de
  djnz  # .loop-div16
  ex    af, afx
  ld    b, a
  ex    af, afx
  ;; 8
  add   hl, hl
  sla   c
  rla
  ret   nc
  add   hl, de
 [ENDIF]
  ret
;code-no-next


;; fast 16-bit unsigned division.
;; doesn't use alternate registers.
code: UU/MOD  ( ua ub -- ua/ub ua%ub )
  pop   de
  ld    hl, bc
  ex    (sp), hl
  ;; HL=a
  ;; DE=b
  call  # vmx-do-udivmod
  ;; HL=remainder
  ;; CA=quotient
  ld    d, c
  ld    e, a
  pop   bc
  jp    # i-next-push-de-hl

;; in:
;;   HL=a
;;   DE=b
;; out:
;;   CA=quotient
;;   HL=remainder
vmx-do-udivmod:
  \ ld    a, h
  \ or    l
  \ jr    z, # vmx-do-udivmod-0-hl
  ld    a, d
  or    e
  jr    z, # vmx-do-udivmod-0-de

  ld    bc, hl
  ld    hl, # 0
  ld    a, b
  call  # vmx-udiv-part
  ld    b, a
  ld    a, c
  ld    c, b
  call  # vmx-udiv-part
  \ ld    b, c
  \ ld    c, a
  ret

vmx-do-udivmod-0-de:
  ex    de, hl
vmx-do-udivmod-0-hl:
  ld    c, l
  ld    a, l
  ret

vmx-udiv-part:
 OPT-16BIT-MUL/DIV-UNROLLED? [IF]
  ;; 1
  rla
  adc   hl, hl
  sbc   hl, de
  jr    nc, # $ 3 +
  add   hl, de
  ;; 2
  rla
  adc   hl, hl
  sbc   hl, de
  jr    nc, # $ 3 +
  add   hl, de
  ;; 3
  rla
  adc   hl, hl
  sbc   hl, de
  jr    nc, # $ 3 +
  add   hl, de
  ;; 4
  rla
  adc   hl, hl
  sbc   hl, de
  jr    nc, # $ 3 +
  add   hl, de
  ;; 5
  rla
  adc   hl, hl
  sbc   hl, de
  jr    nc, # $ 3 +
  add   hl, de
  ;; 6
  rla
  adc   hl, hl
  sbc   hl, de
  jr    nc, # $ 3 +
  add   hl, de
  ;; 7
  rla
  adc   hl, hl
  sbc   hl, de
  jr    nc, # $ 3 +
  add   hl, de
  ;; 8
  rla
  adc   hl, hl
  sbc   hl, de
  jr    nc, # $ 3 +
  add   hl, de
  ;; done
  rla
  cpl
  ret
 [ELSE]
  ld  b, # 8
.loop0:
  rla
  adc   hl, hl
  sbc   hl, de
  jr    nc, # .skip-add-0
  add   hl, de
.skip-add-0:
  djnz  # .loop0
  rla
  cpl
  ret
 [ENDIF]
;code-no-next


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; fast 16-bit signed mutliplication and division

;; the remainder takes its sign from the dividend.
;; k8: don't even ask me!
code: /MOD  ( a b -- a%b a/b )
  pop   de
  ld    hl, bc
  ex    (sp), hl
  ;; HL=a
  ;; DE=b

  ld    a, d
  or    e
  jr    z, # vmx-do-divmod-0-de

  ld    a, h
  xor   d
  ex    af, afx ;; store flags
  push  hl      ;; save dividend

  call  # vmx-abs-de-hl
  call  # vmx-do-udivmod
  ;; HL=remainder
  ;; CA=quotient
  ld    b, c
  ld    c, a
  ;; HL=remainder
  ;; BC=quotient

  pop   af      ;; dividend high byte is in A
  rla
  jr    nc, # .remainder-is-ok
  ;; negate HL
  ex    de, hl
  ld    hl, # 0
  or    a
  sbc   hl, de
.remainder-is-ok:
  ex    de, hl
  ;; DE=remainder
  ;; BC=quotient
  ex    af, afx ;; sign flags
  ld    hl, bc
  ;; DE=remainder
  ;; HL=quotient
  jp    p, # .quot-is-ok
  ld    hl, # 0
  ;; carry is reset here
  sbc   hl, bc
.quot-is-ok:
  pop   bc
  jp    # i-next-push-de-hl

@vmx-do-divmod-0-de:
  ld    hl, de
  jr    # .quot-is-ok

vmx-abs-de-hl:
  ex    de, hl
  bit   7, h
  call  nz, # vmx-neg-hl
  ex    de, hl
  bit   7, h
  ret   z

vmx-neg-hl:
  ld    a, h
  cpl
  ld    h, a
  ld    a, l
  cpl
  ld    l, a
  inc   hl
  ret
;code-no-next
[ENDIF]


<zx-done>
