;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Z80 Abersoft fig-FORTH recompiled
;; Copyright (C) 2024 Ketmar Dark // Invisible Vector
;; Understanding is not required. Only obedience.
;; GPLv3 ONLY
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; 42/64 columns driver, common code
;; directly included from "zx-22-prims-io.f"
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


<zx-definitions>

code|: 6Y@  ( -- y )
  ld    a, () emit6y
zx-word-code6xy-load-finish:
  ld    l, a
  ld    h, # 0
;code-push-hl

code|: 6X@  ( -- y )
  ld    a, () emit6x
  jr    # zx-word-code6xy-load-finish
;code-no-next

code|: 6Y!  ( y )
  pop   hl
  ld    a, l
  ld    emit6y (), a
;code

code|: 6X!  ( y )
  pop   hl
  ld    a, l
  ld    emit6x (), a
;code


|: XCLS  0 6X! 0 6Y! (CLS) ;

: (AT@)  ( -- y x )  6Y@ 6X@ ;

: (AT)  ( y x -- )
  0MAX [ #COLS 1- ] {#,} MIN 6X!
  0MAX 23 MIN 6Y! ;


|: (.CCX)  (EMIT-RAW) MASK-P C! TINV TOVER 0 TO OBL? ;

;; draw input cursor
|: (.CUR)
  KCUR C@ 1 AND NOT?EXIT
  TOVER? TINV? MASK-P C@
  ;; emulate FLASH
  KCUR C@  SYS: FRAMES C@
  ( tover tinv mask-p kcur frames )
  OVER $80 AND IF ;; blinking bar
    OVER XOR $10 AND 0IF 2DROP 2DROP EXIT ENDIF
    $10 XOR KCUR C!
    1 TOVER 1 TINV
    $FF MASK-P C!
    BL
  ELSE  ;; C/L
    $10 AND TINV 0 TOVER
    $10 OR KCUR C!
    ;; use L or C (sadly, the difference is 9)
    \ SYS: FLAGS2 C@ 8 AND IF $43 ELSE $4C ENDIF (EMIT-RAW)
    ;; 2 bytes smaller
    $4C SYS: FLAGS2 C@ 8 AND DUP 1 MIN + -
  ENDIF (.CCX) ;

;; erase input cursor
|: (.CUR0)
  KCUR C@ $11 AND $11 - ?EXIT
  $10 KCUR XOR!C
  TOVER? TINV? MASK-P C@
  KCUR C@ $80 AND DUP IF $FF MASK-P C! ENDIF
  0<> DUP TOVER TINV
  BL (.CCX) ;


|: (CR?)
  6Y@ 6X@ #COLS - 0IF 0 6X! 1+ ENDIF
  DUP 23 U> IF DROP SCROLL 23 ENDIF 6Y! ;

|: (.CR)  0 6X! 6Y@ 1+ 6Y! (CR?) ;

|: (.>>)  6X@ 1+ #COLS MIN 6X! ;


|: (.<)
  6X@ -DUP IF 1-
  ELSE 6Y@ DUP 1- 0MAX 6Y! IF [ #COLS 1- ] {#,} ELSE 0 ENDIF ENDIF 6X! ;

|: (.>)  6Y@ 23 > IF (CR?) EXIT ENDIF (.>>) ;
|: (.V)  6Y@ 1+ 23 MIN 6Y! ;
|: (.^)  6Y@ 1- 0MAX 6Y! ;

|: (.XX)  (CR?) (EMIT-RAW) (.>>) OUT 1+! ;
|: (.XY)  0 (.XX) ;

\ |: (.TAB)   BEGIN BL (.XX) 6X@ 3 AND 0UNTIL ;
|: (.TAB)   4 6X@ 3 AND - SPACES ;
|: (.0<<)   0 6X! ;
|: (.ENDCR) 6X@ IF (.CR) ENDIF ;

<zx-hidden>
CREATE (CTLS)
  ['] (.ENDCR) zx-w,  ;; 4
  ['] CLS      zx-w,  ;; 5
  ['] (.TAB)   zx-w,  ;; 6
  ['] (.0<<)   zx-w,  ;; 7
  ['] (.<)     zx-w,  ;; 8
  ['] (.>)     zx-w,  ;; 9
  ['] (.V)     zx-w,  ;; 10
  ['] (.^)     zx-w,  ;; 11
  ['] (.XY)    zx-w,  ;; 12
  ['] (.CR)    zx-w,  ;; 13
create;

|: (.INK)    ( ch )  ['] (CEMIT) TO EMIT INK ;
|: (.PAPER)  ( ch )  ['] (CEMIT) TO EMIT PAPER ;
|: (.FLASH)  ( ch )  ['] (CEMIT) TO EMIT FLASH ;
|: (.BRIGHT) ( ch )  ['] (CEMIT) TO EMIT BRIGHT ;
|: (.INV)    ( ch )  ['] (CEMIT) TO EMIT INVERSE ;
|: (.OVER)   ( ch )  ['] (CEMIT) TO EMIT OVER ;

CREATE (X-CTLS)
  ['] (.INK)    zx-w, ;; 16
  ['] (.PAPER)  zx-w, ;; 17
  ['] (.FLASH)  zx-w, ;; 18
  ['] (.BRIGHT) zx-w, ;; 19
  ['] (.INV)    zx-w, ;; 20
  ['] (.OVER)   zx-w, ;; 21
create;

;; low byte: 2nd arg for 2-arg control codes
;; high byte: control code for 2-arg control codes
0 quan EARG

;; AT or TAB
|: (E1A)  ( ch )
  ['] (CEMIT) TO EMIT
  EARG HI-BYTE 22 - 0IF EARG LO-BYTE SWAP AT EXIT ENDIF ;; AT
  ;; TAB
  DROP 6X@ EARG LO-BYTE - SPACES ;

|: (E2A)  ( ch )  QADDR EARG OR!C ['] (E1A) TO EMIT ;

|: (CEMIT)  ( char )
  LO-BYTE
  DUP 31 U> IF (.XX) EXIT ENDIF  ;; fastest case for normal chars
  DUP 4 14 WITHIN IF 4 - 2* (CTLS) + @ EXECUTE-TAIL ENDIF
  DUP 16 22 WITHIN IF 16 - 2* (X-CTLS) + @ TO EMIT EXIT ENDIF
  DUP 22 24 WITHIN IF 256* TO EARG ['] (E2A) TO EMIT EXIT ENDIF
  (.XX) ;

<zx-normal>

['] (XCLS) TO CLS
['] (.CUR) TO .CUR
['] (.CUR0) TO .CUR0
['] (AT) TO AT
['] (AT@) TO AT@

;; ROM codes:
;;    4 -- endcr
;;    5 -- cls
;;    6 -- next tab stop
;;    7 -- ? (edit) -- move to the line start
;;    8 -- left
;;    9 -- right
;;   10 -- down
;;   11 -- up
;;   12 -- ? (del)
;;   13 -- cr
;;   14 -- ? (nothing)
;;   15 -- ? (nothing)
;;   16 -- INK
;;   17 -- PAPER
;;   18 -- FLASH
;;   19 -- BRIGHT
;;   20 -- INVERSE
;;   21 -- OVER
;;   22 -- AT
;;   23 -- TAB (position in x, but wants 2 args)

<zx-done>
