;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Z80 Abersoft fig-FORTH recompiled
;; Copyright (C) 2024 Ketmar Dark // Invisible Vector
;; Understanding is not required. Only obedience.
;; GPLv3 ONLY
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; input/output primitives
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


<zx-definitions>
zx-here

0 quan OBL? -- last printed char was BL?
0 quan SC#  -- incremented on scroll
;; this has to be a VARIABLE for compatibility
0 variable OUT
;; show cursor while waiting a key?
;; bit 0: show cursor?
;; bit 7: simple inverted block, no C/L char.
;; bit 4 is used to tell if the cursor was printed.
;; it is set by ".CUR", and reset by ".CUR0".
1 variable KCUR
;; allow CS+9 "GRAPH" switch? has sense only for 32-columns mode.
1 variable ALLOW-G?


(*
code|: OPEN-CHAN-#2
  push  bc
  ld    a, # 2
  call  # $1601
  pop   bc
;code
*)

code|: (ROM-RAW-EMIT)  ( char )
  pop   hl    ;; code
  push  bc
  push  ix    ;; dunno if i need it
zx-word-raw-emit-pop-ix-finish:
  ld    a, l
  rst   # $10
  pop   ix
  pop   bc
  ;; make temporary attributes permanent
  ;; we have channel #2 (main screen) open, so just call ROM
  call  # $1CAD
;code

code|: (ROM-EMIT-ATTR)  ( n code )
  pop   de    ;; code
  pop   hl    ;; n
  push  bc
  push  ix    ;; dunno if i need it
  ld    a, e
  push  hl
  rst   # $10
  pop   hl
  jr    # zx-word-raw-emit-pop-ix-finish
;code-no-next


;; common "CLS" code
code|: (CLS)
  push  bc
zx-word-pr-cls-entry: ;; used in 32-cols driver
  ;; fix attrs
  ld    hl, () sysvar-attr-p
  ld    sysvar-attr-t (), hl
  ;; reset scroll count
  ld    hl, # 0
  ld    zx-['pfa] SC# (), hl
  ;; reset "cursor printed" bit
  ld    hl, # zx-['pfa] KCUR
  res   4, (hl)
  ;; clear scr$
  ld    hl, # $4000
  ld    de, # $4001
  ld    bc, # 6144
  ld    (hl), # 0
  ldir
  ;; set attr$
  ld    bc, # 767
  ld    a, () sysvar-attr-p
  ld    (hl), a
  ldir
  ;; we need to have at least one line in the bottom area
  ld    a, # 1
  ld    sysvar-defsz (), a
  ;; open chan #2
  ld    a, # 2
  call  # $1601
  ;; print whole 24 lines
  xor   a
  ld    sysvar-defsz (), a
  ;; no scroll prompt
  dec   a
  ld    sysvar-scr-ct (), a
  ;; AT 0, 0
  ld    a, # 22
  rst   # $10
  xor   a
  rst   # $10
  xor   a
  rst   # $10
  pop   bc
;code

;; scroll up
code|: SCROLL
  push  bc
  call  # $0DFE ;; CL-SC-ALL
  ;; set attribute for the lower part
  ld    hl, # $5800 23 32 * +
  ld    de, # $5800 23 32 * + 1+
  ld    a, () sysvar-attr-p
  ld    (hl), a
  ld    bc, # 31
  ldir
  ld    hl, # zx-['pfa] SC#
  inc   (hl)
  pop   bc
;code

;; this is defined for all drivers, because why not?
code: ATTR  ( y x -- attr )
  (*
  pop   hl
  pop   de
  push  bc
  push  ix
  ld    c, e
  ld    b, l
  call  # $2583
  call  # $1E94
  pop   ix
  pop   bc
  ld    h, # 0
  ld    l, a
  *)
  pop   hl
  pop   de
  ;; E=y
  ;; L=x
  ld    a, e    ;; y
  rrca
  rrca
  rrca
  ld    e, a
  and   # $E0
  xor   l       ;; x
  ld    l, a
  ld    a, e    ;; y
  and   # $03
  xor   # $58
  ld    h, a
  ld    l, (hl)
  ld    h, # 0
;code-push-hl


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; load print driver

0 vect CLS
0 vect .CUR
0 vect .CUR0
0 vect AT
0 vect AT@

<zx-done>
OPT-64-COLS? [IF]
$include "zx-28-prims-io-25-emit4.f"
[ENDIF]
OPT-42-COLS? [IF]
$include "zx-28-prims-io-21-emit6.f"
[ENDIF]
OPT-32-COLS? [IF]
$include "zx-28-prims-io-10-emit8.f"
[ELSE]
$include "zx-28-prims-io-28-emit6or4-common.f"
[ENDIF]
<zx-definitions>


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; keyboard input words

code: ?TERMINAL  ( -- break? )
  call  # $1F54   ;; this doesn't modify BC
  ld    a, # 0
  ld    h, a
  ccf
  adc   a, h
  ld    l, a
;code-push-hl

;; read key, no wait, no translation, no special processing.
;; zeroes LAST-K.
code: INKEY?  ( -- char )
  \ push  bc
  \ ld    a, # 2
  \ call  # $1601
  \ pop   bc
  \ ei    ;; just in case
  ;; bit 5 of (iy+1) is set when a new key is pressed
  ;; just in case
  ld    hl, # sysvar-last-k
  ld    a, (hl)
  ;; zero LAST-K
  ld    (hl), # 0
  ld    h, # 0
  ld    l, a
  cp    a, # 13
  jr    nz, # .not-cr
  ;; ENTER produces #13
  ;; SS+ENTER produces #2
  ;; check for SS pressed
  ld    a, # $7F
  in    a, () $FE
  and   # $02
  jr    nz, # .not-cr
  ld    l, # 2
.not-cr:
  ;; just in case
  OPT-32-COLS? [IF]
  ld    a, (iy+) 7
  and   # $02
  ld    7 (iy+), a
  [ELSE]
  ld    7 (iy+), # 00 ;; #5C41
  [ENDIF]
;code-push-hl

;; translate pressed key
code: TR-KEY  ( char -- char )
  pop   hl
  ld    a, l
  ld    e, a
  or    a
  jr    z, # .done

  cp    # 128
  jr    nc, # .translate

  cp    # 32
  jr    nc, # .done

  ;; something's pressed
  cp    # 6           ;; CS+2?
  jr    nz, # .not-caps
  ;; toggle CAPS
  ld    hl, # sysvar-flags2
  ld    a, # $08      ;; CAPS flag
.do-graph:
  xor   (hl)
  ld    (hl), a
  ld    e, # 0
  jr    # .done

.not-caps:
  OPT-32-COLS? [IF]
  ;; CS+9 should switch to "G"
  cp    # 15
  jr    nz, # .not-graph
  ld    a, () zx-['pfa] ALLOW-G?
  or    a
  jr    z, # .not-graph
  ld    hl, # sysvar-tv-flag
  ld    a, # $02      ;; GRAPH flag
  jr    # .do-graph
.not-graph:
  ld    a, # 15
  [ENDIF]

.done:
  ld    l, e
  ld    h, # 0
  next-push-hl

.translate:
  ;; translate codes
  ld    hl, # key-translation-table 1-
.table-loop:
  inc   hl
  ld    a, (hl)       ;; src
  or    a             ;; end?
  jr    z, # .done
  inc   hl
  cp    e
  ld    a, (hl)       ;; dest
  jr    nz, # .table-loop
  ld    e, a
  jr    # .done

flush!

key-translation-table:
  198 db,  91 db, ;; AND->[
  197 db,  93 db, ;; OR->]
  172 db, 127 db, ;; AT->(c)
  226 db, 126 db, ;; STOP->~
  195 db, 124 db, ;; NOT->|
  205 db,  92 db, ;; STEP->slash
  204 db, 123 db, ;; TO->{
  203 db, 125 db, ;; THEN->}
  ;; SS+Q, SS+W, SS+E
  199 db, 29 db,  ;; SS+Q
  201 db, 30 db,  ;; SS+W
  200 db, 31 db,  ;; SS+E
0 db,
;code-no-next
\ <zx-forth>

;; bit 0 set for CS, bit 1 set for SS
code: CS/SS?  ( -- shift-flags )
  ld    h, # 0
  ;; check for SS pressed
  ld    a, # $7F
  in    a, () $FE
  cpl
  and   # $02
  ld    l, a
  ;; check for CS pressed
  ld    a, # $FE
  in    a, () $FE
  cpl
  and   # $01
  or    l
  ld    l, a
;code-push-hl


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; high-level I/O words

;; "EMIT" is vectored
['] (CEMIT) VECT EMIT

: CR     13 EMIT ;
: ENDCR  4 EMIT ;
: SPACE  BL EMIT ;
OPT-32-COLS? [IF]
: ?SPACE OBL? ?EXIT SYS: S-POSN-X C@ 33 - NOT?EXIT SPACE ;
[ELSE]
: ?SPACE OBL? ?EXIT AT@ DROP NOT?EXIT SPACE ;
[ENDIF]


code|: KEY-BEEP
  push  bc
  push  ix
  ld    hl, # $C8   ;; pitch
  ld    a, () sysvar-pip
zx-word-kbeep-finish:
  ld    e, a
  ld    d, # 0
zx-word-kbeep-do-beep:
  call  # $03B5
  pop   ix
  pop   bc
;code

code|: ERR-BEEP
  push  bc
  push  ix
  ld    hl, # $1A90 ;; pitch
  ld    a, () sysvar-rasp
  jr    # zx-word-kbeep-finish
;code-no-next


code: BLEEP  ( duration pitch )
  pop   hl
  pop   de
  push  bc
  push  ix
  \ call  # $03B5
  \ pop   ix
  \ pop   bc
  jp    # zx-word-kbeep-do-beep
;code-no-next

;; calculate values for BLEEP
: (BLC)  ( ms hz -- duration pitch )
  >R R@ 1000 */ ( 3500.000) $67E0 $35 R> U/MOD
  241 - 8u/ SWAP DROP ;

: BEEP  ( ms hz )  (BLC) BLEEP ;


: KEY-NC  ( -- code )
  $EF KCUR AND!C  ;; reset "cursor printed" bit (just in case)
  BEGIN .CUR INKEY? TR-KEY -DUP UNTIL KEY-BEEP .CUR0 ;

: KEY  ( -- code )  0 SYS: LAST-K C!  0 SYS: TV-FLAG C!  KEY-NC ;


code|: (FIX-ATTRS)
  (*
  ld    hl, () sysvar-attr-p
  ;; we need Z flag to be set
  xor   a
  call  # $0D5B
  *)
  ;; we have channel #2 (main screen) open, so just call ROM
  call  # $0D4D
;code

;; make temporary attributes permanent
\ code|: (ATTRS-T-PERM)
\   ;; we have channel #2 (main screen) open, so just call ROM
\   call  # $1CAD
\ ;code

: INK@     ( -- n )  SYS: P-FLAG $10 AND IF 9 EXIT ENDIF SYS: ATTR-P C@ 7 AND ;
: PAPER@   ( -- n )  SYS: P-FLAG $40 AND IF 9 EXIT ENDIF SYS: ATTR-P C@ 8u/ 7 AND ;
: BRIGHT@  ( -- n )  SYS: ATTR-P C@ $40 AND 0<> ;
: FLASH@   ( -- n )  SYS: ATTR-P C@ $7F U> ;
: GOVER@   ( -- flag )  SYS: P-FLAG C@ $03 AND 0<> ;
: INVERSE@ ( -- flag )  SYS: P-FLAG C@ $0C AND 0<> ;

: INK     ( n )  DUP 9 U> IF DROP EXIT ENDIF 16 (ROM-EMIT-ATTR) ;
: PAPER   ( n )  DUP 9 U> IF DROP EXIT ENDIF 17 (ROM-EMIT-ATTR) ;
: FLASH   ( n )  18 (ROM-EMIT-ATTR) ;
: BRIGHT  ( n )  19 (ROM-EMIT-ATTR) ;
: INVERSE ( n )  0<> 0<> [ OPT-32-COLS? ] [IFNOT] DUP TINV [ENDIF] 20 (ROM-EMIT-ATTR) ;
: GOVER   ( n )  0<> [ OPT-32-COLS? ] [IFNOT] DUP TOVER [ENDIF] 0<> 21 (ROM-EMIT-ATTR) ;


code: INKEY  ( -- char // 255 )
  push  bc
  call  # $028E
  jr    nz, # .done
  call  # $031E
  jr    nc, # .done
  dec   d
  ld    e, a
  call  # $0333
.done:
  ld    l, a
  ld    h, # 0
  pop bc
;code-push-hl


zx-here swap - endcr ." print driver size: " .bytes cr
<zx-done>
