;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Z80 Abersoft fig-FORTH recompiled
;; Copyright (C) 2024 Ketmar Dark // Invisible Vector
;; Understanding is not required. Only obedience.
;; GPLv3 ONLY
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; high-level low-level words ;-)
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


<zx-definitions>

<zx-system>
\ : ID-TYPE  ( addr count )
\   -DUP IF OVER + SWAP DO I C@ 127 AND EMIT LOOP ELSE DROP ENDIF ;
: ID-TYPE  ( addr count )
  BEGIN DUP WHILE SWAP C@++ 127 AND
    DUP BL < IF DROP [CHAR] ? ENDIF
    EMIT SWAP 1-
  REPEAT 2DROP ;
<zx-forth>

;; k8: this was a copying one. rewritten.
: ID.  ( nfa )  SYS: COUNT $1F AND SYS: ID-TYPE ;

|: WRAP-.ID  ( nfa )
  OUT C@ IF
    DUP C@ 31 AND OUT C@ + 2+ #COLS > IF CR OUT OFF ELSE SPACE ENDIF
  ENDIF ID. ;

: VLIST
  ENDCR OUT OFF CONTEXT @ @
  BEGIN
    DUP
    [ OPT-NEW-ART-HEADER? ] [IF]
      2+ WRAP-.ID
    [ELSE]
      WRAP-.ID PFA LFA
    [ENDIF]
  @ DUP 0= ?TERMINAL OR UNTIL DROP ;


: NOOP  ;
0 VECT (ABORT)  -- will be set later

: SIZE  ( -- n )  HERE ( 0 +ORIGIN) [ zx-base-addr ] {#,} - ;
: FREE  ( -- n )  SYS: SP@ HERE - ;

: TRAVERSE  ( addr1 n -- addr2 )  SWAP BEGIN OVER + 127 OVER C@ < UNTIL NIP ;

: -TRAILING  ( addr count -- addr1 count1 )
  DUP FOR 2DUP + 1- C@ BL - IF LEAVE ELSE 1- ENDIF LOOP ;


<zx-system>
: ?LOADING  BLK @ 0= 22 ?ERROR ;

: ?COMP   STATE@ 0= 17 ?ERROR ;
: ?EXEC   STATE@ 18 ?ERROR ;
: ?CSP    SP@ CSP @ - 20 ?ERROR ;
: ?PAIRS  ( n1 n2 )  - 19 ?ERROR ;
: ?CPAIRS ( n1 n2 )  ?COMP ?PAIRS ;


;; compile branch offset from HERE
;; kept for compatibility with fig-Forth
: BACK  ( addr )  [ OPT-REL-BRANCH? ] [IF] HERE - [ENDIF] , ;

;; patch forward branch at addr to HERE
: FWRD  ( addr )  HERE [ OPT-REL-BRANCH? ] [IF] OVER - [ENDIF] SWAP ! ;
;; compile dummy branch offset, return offset address
: BMARK  ( -- addr )  HERE 0 , ;

\ OPT-ADVANCED-CFLOW? [IF]
|: BRDEST!  ( dest addr )  [ OPT-REL-BRANCH? ] [IF] TUCK - SWAP [ENDIF] ! ;
;; compile dummy branch offset, return offset address.
;; used to chain forward jumps; start with "0" `old-chain`.
|: CHAIN>   ( old-chain -- new-chain )  HERE SWAP , ;
|: MARK>    ( -- chain-new )  0 CHAIN> ;
|: <MARK    ( -- dest-addr )  HERE ;
|: <RESOLVE ( dest-addr )     HERE 0 , BRDEST! ;
|: RESOLVE> ( chain )  HERE >R BEGIN DUP WHILE DUP @ R@ ROT BRDEST! REPEAT DROP RDROP ;

|: ?2PAIRS  ( pn n0 n1 )  ?COMP ROT TUCK ( n0 pn n1 pn) - NROT - AND 19 ?ERROR ;
\ [ENDIF]

|: CSP@  ( -- value )  CSP @ ;
|: CSP!  ( value)      CSP ! ;
: !CSP  SP@ CSP! ;
<zx-forth>


|: LATEST-VLINK  ( -- link )  CURRENT @ @ ;

OPT-NEW-ART-HEADER? [IF]
;; new header format
OPT-DTC? [IF]
: CFA  ( pfa -- cfa )  3 - ;
: NFA  ( pfa -- nfa )  4 - -1 TRAVERSE ;
: PFA  ( nfa -- pfa )  1 TRAVERSE 4 + ;
|: CFA>PFA  3 + ;
[ELSE]
\ : CFA  ( pfa -- cfa )  2- ;
alias-for 2- is CFA
: NFA  ( pfa -- nfa )  3 - -1 TRAVERSE ;
: PFA  ( nfa -- pfa )  1 TRAVERSE 3 + ;
|: CFA>PFA  2+ ;
[ENDIF]
: LFA  ( pfa -- lfa )  NFA 2- ;

: LATEST  ( -- nfa )  LATEST-VLINK 2+ ;

[ELSE]

;; old header format
OPT-DTC? [IF]
: LFA  ( pfa -- lfa )  5 - ;
: CFA  ( pfa -- cfa )  3 - ;
: NFA  ( pfa -- nfa )  6 - -1 TRAVERSE ;
: PFA  ( nfa -- pfa )  1 TRAVERSE 6 + ;
|: CFA>PFA  3 + ;
[ELSE]
: LFA  ( pfa -- lfa )  4 - ;
\ : CFA  ( pfa -- cfa )  2- ;
alias-for 2- is CFA
: NFA  ( pfa -- nfa )  5 - -1 TRAVERSE ;
: PFA  ( nfa -- pfa )  1 TRAVERSE 5 + ;
|: CFA>PFA  2+ ;
[ENDIF]

: LATEST  ( -- nfa )  LATEST-VLINK ;
[ENDIF]

: LATEST-PFA  ( -- cfa )  LATEST PFA ;
: LATEST-CFA  ( -- cfa )  LATEST-PFA CFA ;


: STATE@  ( -- state )  SYS: STATE @ ;

: SMUDGE     LATEST $20 TOGGLE ;
: IMMEDIATE  LATEST $40 TOGGLE ;

: HEX      16 BASE ! ;
: DECIMAL  10 BASE ! ;

: HERE   ( n )  SYS: (DP) @ ;
: ALLOT  ( n )  SYS: (DP) +! ;

: ,  ( n )  HERE ! 2 ALLOT ;
: C, ( n )  HERE C! 1 ALLOT ;

<zx-system>
: CALL,  ( cfa )
  [ OPT-DTC? ] [IF]
    ;; follow JP chain. this makes primitive calls faster.
    BEGIN DUP C@ $C3 - 0WHILE 1+ @ REPEAT
  [ENDIF] , ;

OPT-DTC? [IF]
: CFA@  ( cfa -- [cfa] )  1+ @ ;
[ELSE]
ALIAS-FOR @ IS CFA@
[ENDIF]

|: [CCX]  (  pfa len-byte TRUE // FALSE )
  0= 0 ?ERROR DROP CFA STATE@ IF CALL, EXIT ENDIF EXECUTE-TAIL ;noexit
<zx-forth>

;; added by k8
*: RECURSE  LATEST-CFA SYS: CALL, ;

;; added by k8
*: RECURSE-TAIL  COMPILE SYS: BRANCH LATEST-PFA SYS: BACK ;

;; k8: wow, it is not immediate!
;; it is ok to use "," here
: COMPILE  SYS: ?COMP R> DUP 2+ >R @ , ;

OPT-CLIT? OPT-BLIT? forth:or [IF]
*: LITERAL
  STATE@ NOT?EXIT
  DUP HI-BYTE
  [ OPT-CLIT? ] [IF]
    [ OPT-BLIT? ] [IF]
      DUP 0IF DROP COMPILE SYS: CLIT C, EXIT ENDIF
    [ELSE]
      0IF COMPILE SYS: CLIT C, EXIT ENDIF
    [ENDIF]
  [ENDIF]
  [ OPT-BLIT? ] [IF]
    $FF - 0IF COMPILE SYS: BLIT C, EXIT ENDIF
  [ENDIF]
  COMPILE SYS: LIT , ;
[ELSE]
*: LITERAL  STATE@ NOT?EXIT COMPILE SYS: LIT , ;
[ENDIF]

*: DLITERAL   STATE@ NOT?EXIT SWAP [COMPILE] LITERAL [COMPILE] LITERAL ;


\ : TYPE  ( addr count )
\   -DUP IF OVER + SWAP DO I C@ EMIT LOOP ELSE DROP ENDIF ;
;; this is smaller
: TYPE  ( addr count )
  \ DUP IF SWAP COUNT EMIT SWAP 1- RECURSE-TAIL ENDIF 2DROP ;
  BEGIN DUP WHILE SWAP C@++ EMIT SWAP 1- REPEAT 2DROP ;

: SPACES  ( n )  0MAX FOR SPACE LOOP ;

<zx-system>
: (")   ( -- addr count )  R> COUNT 2DUP + >R ;
\ : (.")  R@ COUNT DUP 1+ R> + >R TYPE ;
: (.")  R> COUNT 2DUP + >R TYPE ;
<zx-forth>

: ERASE  ( addr count )  0 FILL ;
: BLANKS ( addr count )  BL FILL ;
\ : BLANK ( addr count )  BLANK ;
alias-for BLANKS is BLANK

: PAD  ( -- pad^ )  HERE 68 + ;


: +-  ( n1 n2 -- n3 )  0< IF MINUS ENDIF ;
: D+- ( d1 n -- d2 )   0< IF DMINUS ENDIF ;

: ABS  ( n1 -- n2 )  DUP +- ;
: DABS ( d1 -- d2 )  DUP D+- ;

: MIN  ( a b -- c )  2DUP > IF SWAP ENDIF DROP ;
: MAX  ( a b -- c )  2DUP < IF SWAP ENDIF DROP ;

: M*  ( nl n2 -- d )  2DUP XOR >R ABS SWAP ABS U* R> D+- ;
;; this is SM/REM -- symmetric
;; and it should be called "M/MOD"!
;; divides a double into n giving quotient q and remainder r.
;; the remainder has the sign of dividend d.
: M/  ( d n -- r q )  OVER >R >R DABS R@ ABS U/MOD R> R@ XOR +- SWAP R> +- SWAP ;

;; this works, but i don't think anybody will need it
\ : FM/MOD ( d n -- r q )
\   DUP >R M/ OVER DUP 0<> SWAP 0<
\   R@ 0< XOR AND IF 1- SWAP R> + SWAP ELSE R> DROP ENDIF ;

: */MOD  ( a b -- a%b a/b )  >R M* R> M/ ;
: */     ( a b -- c )  */MOD NIP ;
: M/MOD  ( ud1 u2 -- u3 ud4 )  >R 0 R@ U/MOD R> SWAP >R U/MOD R> ;

OPT-16BIT-MUL/DIV? [IF]
: U/    ( ua ub -- ua/ub )  UU/MOD DROP ;
: UMOD  ( ua ub -- ua%ub )  UU/MOD NIP ;
;; for truncated result, there is no difference between signed and unsigned multiplication
\ : *  ( a b -- c )  UU* ;
ALIAS-FOR UU* IS *
: /    ( a b -- a/b )  /MOD NIP ;
: MOD  ( a b -- a%b )  /MOD DROP ;
[ELSE]
: U/   ( ua ub -- ua/ub )  0 SWAP U/MOD NIP ;
: UMOD ( ua ub -- ua%ub )  0 SWAP U/MOD DROP ;
: *    ( a b -- c )  M* DROP ;
: /MOD ( a b -- a%b a/b )  >R S>D R> M/ ;
: /    ( a b -- c )  /MOD NIP ;
: MOD  ( a b -- c )  /MOD DROP ;
[ENDIF]

;; square root of d between 0 and 268435455
\ : DSQRT  ( d -- n )
\   2DUP D+ 2DUP D+ 0 $8000 $0E FOR
\     SWAP DROP >R 2DUP R@ U/MOD R> + 2U/
\   LOOP 2U/ >R DROP 2DROP R> ;

|: ESC-CHAR  ( char -- char )
  UPCHAR CASE
    [CHAR] N OF 13 ENDOF
    [CHAR] R OF  7 ENDOF
    [CHAR] C OF  4 ENDOF
    [CHAR] T OF  6 ENDOF
    [CHAR] S OF BL ENDOF
    [CHAR] I OF 16 ENDOF
    [CHAR] P OF 17 ENDOF
    [CHAR] F OF 18 ENDOF
    [CHAR] R OF 19 ENDOF
    [CHAR] V OF 20 ENDOF
    [CHAR] X OF 21 ENDOF
  OTHERWISE DUP [CHAR] 0 58 WITHIN IF [CHAR] 0 - ENDIF
  ENDCASE ;

|: UNESCAPE  ( caddr )
  DUP 1+ OVER COUNT ( caddr waddr raddr count )
  OVER + SWAP DO  ( caddr waddr )
    I C@ [CHAR] \ - IF I C@  1 ELSE I 1+ C@ ESC-CHAR  2 ENDIF
    >R OVER C! 1+ R>
  +LOOP  ( caddr eaddr )
  SWAP DUP >R 1+ - R> C! ;

|: -STR,  WORD HERE DUP UNESCAPE C@ 1+ ALLOT ;

*: "
  SYS: ?COMP COMPILE SYS: (") 34 -STR, ;

*: ."
  34 STATE@ IF COMPILE SYS: (.") -STR, EXIT ENDIF
  WORD HERE COUNT TYPE ;

*: ,"
  SYS: ?EXEC 34 -STR, ;

*: [CHAR]
  BL WORD HERE 1+ C@ [COMPILE] LITERAL ;


: ERROR  ( n -- in blk )
  SYS: WARNING @ 0< IF (ABORT) ENDIF
  HERE COUNT TYPE ." ? " MESSAGE
  SYS: SP! [ OPT-ENABLE-FP? ] [IF] SYS: FP0! [ENDIF]
  ;; k8: this is required for "WHERE"
  BLK @ -DUP IF TO ERR-BLK IN @ TO ERR-POS ENDIF
  \ [ OPT-EDITOR-CODE? OPT-FS-EDITOR-CODE? forth:or ] [IF]
  \   BLK @ -DUP IF IN @ SWAP ENDIF
  \ [ENDIF]
  QUIT ;noexit

\ : ?ERROR  ( flag err# )  SWAP IF ERROR ELSE DROP ENDIF ;
;; k8: "ERROR" never returns anyway
: ?ERROR  ( flag err# )  SWAP IF ERROR ENDIF DROP ;

: MESSAGE  ( n )
  SYS: WARNING @ IF
    \ -DUP IF 4 SYS: OFFSET @ B/SCR / - .LINE SPACE ENDIF EXIT
    -DUP NOT?EXIT \ 4 SYS: OFFSET @ B/SCR / - .LINE SPACE EXIT
    ;; screen 0 contains messages, 32 bytes per each
    (MSGLINE) -TRAILING TYPE SPACE EXIT
  ENDIF
  ." MSG #" . ;


<zx-system>
;; always skips the first char
: (NUMBER)  ( d1 addr1 -- d2 addr2 )
  BEGIN 1+ DUP >R C@ BASE @ DIGIT WHILE
  SWAP BASE @ U* DROP ROT BASE @ U* D+ DPL @ 1+ IF DPL 1+! ENDIF
  R> REPEAT R> ;
<zx-forth>

|: NUM-CHECK-PFX  ( addr -- addr base )
  DUP 1+ C@ CASE
    [CHAR] $ OF 1+ 16 ENDOF
    [CHAR] % OF 1+ 2 ENDOF
    [CHAR] @ OF 1+ 8 ENDOF
  OTHERWISE DROP BASE @ ENDCASE ;

;; k8: added "$", "%", "@" prefixes.
;; k8: this code cannot actualy parse arbitrary strings.
;; k8: it cannot even parse a counted string -- it expects
;; k8: non-digit-terminated string, and ignores the length.
: NUMBER  ( addr -- d )
  0 0 ROT ( 0 0 addr )
  NUM-CHECK-PFX ( 0 0 addr base ) BASE @ >R BASE !
  DUP 1+ C@ [CHAR] - = DUP >R +
  -1 BEGIN DPL ! SYS: (NUMBER) DUP C@ BL - WHILE
    DUP C@ [CHAR] . - ( 0 ?ERROR)
    IF RDROP R> BASE ! 0 ERROR ENDIF
  0 REPEAT
  DROP R> IF DMINUS ENDIF
  R> BASE ! ;


: WORD  ( ch )
  BLK @ IF BLK @ BLOCK ELSE TIB @ ENDIF
  IN @ + SWAP SYS: ENCLOSE
  HERE 34 BLANKS
  IN +!
  OVER - >R R@ HERE C! +
  HERE 1+ R> CMOVE ;

;; Accept following text to PAD. c is the text delimiter.
: TEXT  ( c )  HERE C/L 1+ BLANKS WORD HERE PAD C/L 1+ CMOVE ;


0 [IF]
;; demo game loads in ~43.3 seconds (~44.94 for ITC).
;; demo game loads in ~26.0 seconds (~27.6 for ITC).
: -FIND  ( -- pfa len-byte TRUE // FALSE )
  BL WORD HERE
  CSENS? 0IF DUP UPSTR ENDIF
  CONTEXT @ @ SYS: (FIND) DUP 0=
  IF DROP HERE LATEST-VLINK (FIND) ENDIF ;
[ELSE]
;; demo game loads in ~32.9 seconds (~34.5 for ITC).
;; demo game loads in ~21.0 seconds (~22.5 for ITC).
: -FIND  ( -- pfa len-byte TRUE // FALSE )
  BL WORD HERE
  CSENS? 0IF DUP UPSTR ENDIF
\ BLK @ IF ." W=|" DUP COUNT TYPE ." |l=" DUP C@ . ." ch=" DUP 1+ C@ . ." |" ENDIF
  CONTEXT @ @ SYS: (FIND) DUP ?EXIT
  DROP CURRENT @ DUP CONTEXT @ - 0IF 0AND EXIT ENDIF
  HERE SWAP @ SYS: (FIND) ;
[ENDIF]

|: (-FIND-ONE)  ( vocab-cfa -- pfa len-byte TRUE // FALSE )
  >R
  BL WORD HERE CSENS? 0IF DUP UPSTR ENDIF
  CONTEXT @  R> EXECUTE CONTEXT @
  SWAP CONTEXT !
  @ SYS: (FIND) ;

: -FIND-SYS  ( -- pfa len-byte TRUE // FALSE )
  [ OPT-SYS-VOCAB? ] [IF] ['] (SYS) [ELSE] ['] FORTH [ENDIF]
  (-FIND-ONE) ;

: -FIND-EDT  ( -- pfa len-byte TRUE // FALSE )
  ['] EDITOR (-FIND-ONE) ;


*: [COMPILE]  -FIND SYS: [CCX] ;
ALIAS-FOR [COMPILE] IS [\\] ZX-IMMEDIATE
ALIAS-FOR COMPILE IS \,
*: SYS:  -FIND-SYS SYS: [CCX] ;
*: EDT:  -FIND-EDT SYS: [CCX] ;

: -FIND-RQ  ( -- pfa )  \ name
  -FIND 0= 0 ?ERROR DROP ;

*: '  ( -- pfa )  \ name
  -FIND-RQ [COMPILE] LITERAL ;

*: [']  ( -- cfa )  \ name
  -FIND-RQ CFA [COMPILE] LITERAL ;


\ FIXME: this is broken -- it should fix voc-link!
: FORGET  \ name
  CURRENT @ CONTEXT @ - 24 ?ERROR
  [COMPILE] ' DUP SYS: FENCE @ U< 21 ?ERROR
  DUP NFA SYS: (DP) !
  LFA @ CURRENT @ ! ;


: HOLD  ( ch )  -1 HLD +! HLD @ C! ;
: SIGN  ( n d -- d )  ROT 0< IF 45 HOLD ENDIF ;
: <#  PAD HLD ! ;
: #>  ( d -- addr count )  2DROP HLD @ PAD OVER - ;
\ : #   ( d1 -- d2 )  BASE @ M/MOD ROT 9 OVER < IF 7 + ENDIF 48 + HOLD ;
;; k8: "M/MOD"? why?
;; mixed operation: it leaves the remainder u3 and the quotient ud4 of ud1 / u1.
;; used by # during number representation.
|: #/MOD  ( ud1 u2 -- u3 ud4 )  >R 0 R@ U/MOD R> SWAP >R U/MOD R> ;
: #   ( d1 -- d2 )  BASE @ #/MOD ROT 9 OVER < IF 7 + ENDIF 48 + HOLD ;
: #S  ( d1 -- d2 )  BEGIN # 2DUP OR 0UNTIL ;

: (D.R)  ( d n -- addr count spaces# )  >R SWAP OVER DABS <# #S SIGN #> R> OVER - ;

: D.R ( d n )    (D.R) SPACES TYPE ;
: .R  ( n1 n2 )  >R S>D R> D.R ;

: D.  ( d )  0 D.R SPACE ;
: .   ( n )  S>D D. ;
: ?   ( addr )  @ . ;
: U.  ( u )  0 D. ;

: U.R ( u n )  >R 0 R> D.R ;

OPT-ZXEMUT-DPRINT? [IF]
: ZTYPE  ( addr count )  BEGIN DUP WHILE SWAP C@++ ZEMIT SWAP 1- REPEAT 2DROP ;
: ZENDCR  2 ZEMIT ;
: ZSPACE  BL ZEMIT ;
: ZCR 10 ZEMIT ;

<zx-system>
: (Z.")  R> COUNT 2DUP + >R ZTYPE ;
<zx-forth>

*: Z."
  34 STATE@ IF COMPILE SYS: (Z.") -STR, EXIT ENDIF
  WORD HERE COUNT ZTYPE ;

: zSPACES  ( n ) BEGIN DUP 0> WHILE zSPACE 1- REPEAT DROP ;
: zD.R ( d n )  (D.R) zSPACES zTYPE ;
: z.R  ( n1 n2 )  >R S>D R> zD.R ;

: zD.  ( d )  0 zD.R zSPACE ;
: z.   ( n )  S>D zD. ;
: zU.  ( u )  0 zD. ;

: zU.R ( u n )  >R 0 R> zD.R ;

: z.HEX2  ( u )  BASE @ HEX SWAP 0 <# # # #> zTYPE BASE ! ;
: z.HEX4  ( u )  BASE @ HEX SWAP 0 <# # # # # #> zTYPE BASE ! ;
[ENDIF]


<zx-done>
