;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Z80 Abersoft fig-FORTH recompiled
;; Copyright (C) 2024 Ketmar Dark // Invisible Vector
;; Understanding is not required. Only obedience.
;; GPLv3 ONLY
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; more high-level words
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(*
buffer format:
  dw index
  B/BUF bytes of data
  dw 0

index: bit 15 set if the buffer is dirty.
index 32767 is "none".
*)

<zx-definitions>

OPT-BLOCK-I/O-TYPE R/W-TAPE = [IF]
<zx-system>
code: (TAPE)  ( opcode )
  pop   hl
  push  bc
  push  ix
  ld    a, l
  ld    hl, () zx-['pfa] LO
  ld    tape-load-header-address (), hl
  ld    tape-save-header-address (), hl
  ld    ix, # tape-load-header
  ld    $5C74 (), a
  call  # $075A
  pop   ix
  pop   bc
  next
flush!

tape-load-header:
  3 db,   ;; filetype: code file
  ;; filename: "DISC      "
  68 db, 73 db, 83 db, 67 db, 32 db,
  32 db, 32 db, 32 db, 32 db, 32 db,
tape-load-header-length:
  zx-#ramdisc dw,
tape-load-header-address:
  zx-ramdisc^ dw,
tape-load-header-not-used:
  $2020 dw,

tape-save-header:
  3 db,   ;; filetype: code file
  ;; filename: "DISC      "
  68 db, 73 db, 83 db, 67 db, 32 db,
  32 db, 32 db, 32 db, 32 db, 32 db,
tape-save-header-length:
  zx-#ramdisc dw,
tape-save-header-address:
  zx-ramdisc^ dw,
tape-save-header-not-used:
  $2020 dw,
;code-no-next
<zx-forth>

: LOADT  1 SYS: (TAPE) ;
: SAVET  FLUSH 0 SYS: (TAPE) ;
: VERIFY 2 SYS: (TAPE) ;

;; bugfixed by k8
;; Wipes all information off the RAM disc prior to first use.
: INIT-DISC  [ zx-ramdisc^ ] {#,} [ zx-#ramdisc ] {#,} BLANKS ;
[ENDIF]


<zx-system>
|: DR0  SYS: OFFSET OFF ;
<zx-forth>


;; Mark all block-buffers as empty, not necessarily affecting the contents.
;; Updated blocks are not written to the external storage. This is also an
;; initialisation procedure before first use of the external storage.
: EMPTY-BUFFERS
  SYS: FIRST SYS: LIMIT OVER - ERASE
  SYS: LIMIT SYS: FIRST DO $7FFF I ! B/BUF-FULL +LOOP ;

;; The fig-FORTH standard disc read-write linkage. addr specifies the source
;; or destination block buffer. blk is the sequential number of the referenced
;; block; and f is a flag for f=0 write and f=1 read. R/W determines the
;; location on the external storage, performs the read-write and performs any
;; error checking.
OPT-BLOCK-I/O-TYPE R/W-TAPE = [IF]
: R/W  ( addr blk read-flag )
  OVER $7FFE U> IF DROP 2DROP EXIT ENDIF  \ FIXME
  >R B/BUF * SYS: LO + DUP SYS: HI > 6 ?ERROR R>
  IF SWAP ENDIF B/BUF CMOVE ;
[ENDIF]

OPT-BLOCK-I/O-TYPE R/W-ZXEMUT = [IF]
code|: (R/W)  ( addr read-flag -- read/written/negative-on-err )
  pop   hl
  ld    a, l
  or    a
  ld    a, # 0
  jr    nz, # .do-read
  ld    a, # 2    ;; write, create if necessary
.do-read:
  exx
  ld    de, # 0
$here 2- @def: zio-fofs-hi
  ld    hl, # 0
$here 2- @def: zio-fofs-lo
  exx
  pop   de
  push  bc
  ld    hl, # zxemut-blk-fname
  ld    bc, # zx-i/o-buf-size
flush!
  ;; ZXEmuT trap
  $ED db, $FE db, $18 db,
  2 db, $13 db, $20 db,
flush!
  nop
  ld    hl, bc
  pop   bc
  jr    nc, # .done
  ld    hl, # -1
.done:
  next-push-hl
flush!
zxemut-blk-fname:
  [char] D db,
  [char] I db,
  [char] S db,
  [char] C db,
  [char] . db,
  [char] b db,
  [char] l db,
  [char] k db,
  0 db,
;code
<zx-normal>

\ TODO: better errors!
: R/W  ( addr blk read-flag )
  OVER $7FFE U> IF DROP 2DROP EXIT ENDIF  \ FIXME
  SWAP B/BUF U* asm-label: zio-fofs-hi ! asm-label: zio-fofs-lo !
  IF DUP B/BUF ERASE TRUE ELSE FALSE ENDIF
  (R/W) 0< 8 ?ERROR ;
[ENDIF]

OPT-BLOCK-I/O-TYPE R/W-P3DOS = [IF]
;; Pages 1, 3, 4, 6 are considered as an array
;; of 128 sector buffers (numbered 0...127),
;; each of 512 bytes.
;; The cache and RAMdisk occupy two separate
;; (contiguous) areas of this array.

;; registers for +3DOS calls
;; loaded before calling a function, and the result will be stored here
(*
0 quan +3DOS-AF
0 quan +3DOS-BC
0 quan +3DOS-DE
0 quan +3DOS-HL
*)

;; call +3DOS
;; WARNING! SP should not be at TSTACK!
code|: +3DOS-CMD  ( fn-call-address )
dop3dos:
  pop   hl

  ld    a, () zx-['pfa] +3DOS?
  or    a
  jr    z, # .no-p3dos-present

  push  bc
  push  ix          ;; some +3DOS functions destroys it

  ld    p3dos-call-addr (), hl

  di
  ;; prepare stack and RAM/ROM
  ld    a, () sysvar-bankm  ;; RAM/ROM switching system variable
  and   # $DF               ;; reset bit 5 (lock), just in case
  ld    p3dos-saved-bankm (), a ;; save it to be restored later
  or    # $07               ;; want RAM page 7
  and   # $EF               ;; reset bit 4 (ROM) and bit 5 (lock)
  ld    bc, # $7FFD         ;; port used for horiz ROM switch and RAM paging
  ld    sysvar-bankm (), a  ;; keep system variables up to date
  out   (c), a              ;; RAM page 7 to top and DOS ROM

  ;; set +2A/+3 port #1FFD
  ld    a, () sysvar-bank678
  ld    p3dos-saved-bank678 (), a
  or    # $04               ;; switch on +3DOS ROM
  ld    sysvar-bank678 (), a
  ld    b, # $1F
  out   (c), a

  ld    p3dos-saved-sp (), sp

  ;; switch to IM1
  ld    a, # $3F
  ld    i, a
  im    # 1

  ;; load registers
  ld    sp, # p3dos-reg-af
  pop   af
  pop   bc
  pop   de
  pop   hl

  ;; switch stack (because it is in a switched page).
  ;; this is the official TSTACK used by BASIC when calling +3DOS.
  ;; according to the official manual, it can extend to $5B7C (inclusive).
  ld    sp, # $5C00
$here 2- @def: p3dos-sp0

  ei
  call  # 0   ;; self-modifying code
$here 2- @def: p3dos-call-addr

@p3dosret:
  di
  ;; save all registers
  ld    sp, # p3dos-reg-hl 2+
  push  hl
  push  de
  push  bc
  push  af

  ;; restore 128K banking
  ld    a, () p3dos-saved-bankm
  ld    bc, # $7FFD
  ld    sysvar-bankm (), a
  out   (c), a      ;; switch back to RAM page 0 and 48 BASIC
  ;; restore +2A/+3 port #1FFD
  ld    a, () p3dos-saved-bank678
  ld    b, # $1F
  ld    sysvar-bank678 (), a
  out   (c), a

  ;; restore stack
  ld    sp, () p3dos-saved-sp

  ;; as arrow library is using the TSTACK area to save scr$,
  ;; notify it that there is nothing interesting there.
  ld    hl, # $5C00 1-
  ld    (hl), # 0
  dec   l
  ld    $5B7C (), hl

  ;; switch back to IM2
  OPT-IM2-HANDLER? [IF]
  ld    a, # $3B  ;; ROM contains a lot of #FF there
  ;; I register points to area filled with #FF here
  ld    i, a
  im    # 2
  [ENDIF]

  ei
  pop   ix
  pop   bc
  next

.no-p3dos-present:
  ld    hl, # $0000   ;; reset carry, error #0 (drive not ready)
  ld    p3dos-reg-af (), hl
  next

p3dos-saved-bankm: 0 db,
p3dos-saved-bank678: 0 db,
p3dos-saved-sp: 0 dw,

p3dos-reg-af: 0 dw,
p3dos-reg-bc: 0 dw,
p3dos-reg-de: 0 dw,
p3dos-reg-hl: 0 dw,
;code-no-next

asm-label: p3dos-reg-af constant +3DOS-AF
asm-label: p3dos-reg-bc constant +3DOS-BC
asm-label: p3dos-reg-de constant +3DOS-DE
asm-label: p3dos-reg-hl constant +3DOS-HL

asm-label: p3dos-sp0 constant +3DOS-SP0

code: +3DOS-ERR?  ( -- some-error? )
  ld    a, () p3dos-reg-af
  cpl
zx-word-p3dos-c-flag-push-a:
  and   # $01
zx-word-p3dos-flag-push-a:
  ld    l, a
  ld    h, # 0
;code-push-hl

code: +3DOS-CF?  ( -- carry-set? )
  ld    a, () p3dos-reg-af
  jr    # zx-word-p3dos-c-flag-push-a
;code-no-next

code: +3DOS-ZF?  ( -- zero-set? )
  ld    a, () p3dos-reg-af
  and   # $40
  cp    # $40
  ld    a, # 1
  sbc   a, # 0
  jr    # zx-word-p3dos-flag-push-a
;code-no-next

;; turn off ramdisk
: +3DOS-NO-RAMDISK
  8 +3DOS-DE !  ;; cache: 8 buffers by 512 bytes
  4 +3DOS-HL !  ;; ramdisk: minimum possible size is 4 sectors
  $13F +3DOS-CMD ;

|: INIT+3DOS
  ;; disable alert routine
  0 +3DOS-AF !
  $14E +3DOS-CMD
  ;; start "motor off" countdown
  $199 +3DOS-CMD ;

<zx-hidden>
;; current block file name
create +3BLK-FNAME 13 allot create;
false quan +3BLK-OPEN?
<zx-normal>

|: +3BLK-FNAME!  ( addr count )
  DUP 1 < IF 2DROP " BLOCKS.BLK" ENDIF
  12 MIN
  +3BLK-FNAME 13 $FF FILL
  +3BLK-FNAME SWAP CMOVE ;

;; close opened block file
: +3BLK-CLOSE
  +3BLK-OPEN? IF
    +3DOS-BC OFF
    $109 +3DOS-CMD
    +3DOS-ERR? IF
      ;; abandon it
      +3DOS-BC OFF
      $10C +3DOS-CMD
    ENDIF
    FALSE TO +3BLK-OPEN?
  ENDIF ;

: +3BLK-OPEN  ( addr count )
  +3BLK-CLOSE
  +3BLK-FNAME!
  $0003 +3DOS-BC !  ;; file #0, exclusive r/w
  $0202 +3DOS-DE !  ;; create/open, ignore header
  +3BLK-FNAME +3DOS-HL !
  $106 +3DOS-CMD
  +3DOS-ERR? 8 ?ERROR ;; disk i/o error
  TRUE TO +3BLK-OPEN? ;

: +3DOS-MOTOR-OFF
  $19C +3DOS-CMD ;

: R/W  ( addr blk read-flag )
  +3BLK-OPEN? 0= 8 ?ERROR
  ;; seek
  SWAP B/BUF U* +3DOS-DE ! +3DOS-HL !
  +3DOS-BC OFF
  $136 +3DOS-CMD
  +3DOS-ERR? 8 ?ERROR ;; disk i/o error
  ;; read or write bytes
  DUP IF OVER B/BUF ERASE ENDIF
  SWAP +3DOS-HL !   ;; addr
  B/BUF +3DOS-DE !  ;; length
  $0000 +3DOS-BC !  ;; file # and page #0 for $C000 -- FIXME!
  IF $112 ELSE $115 ENDIF +3DOS-CMD
  +3DOS-ERR? 8 ?ERROR ;; disk i/o error
;

$5B00 constant (CAT-BUF)
\ : (CAT-BUF)  ( -- addr )  PAD ;

: CAT
  (CAT-BUF) 26 ERASE
  BEGIN
    (CAT-BUF) 13 + (CAT-BUF) 13 CMOVE   ;; continue from this entry
    $0201 +3DOS-BC !
    (CAT-BUF) +3DOS-DE !
    " *.*\xff" DROP +3DOS-HL !
    $11E +3DOS-CMD
    +3DOS-ERR? 8 ?ERROR ;; disk i/o error
  +3DOS-BC 1+ C@ 2- 0WHILE
    (CAT-BUF) 13 +
    DUP 8 TYPE [CHAR] . EMIT DUP 8 + 3 TYPE
    SPACE 11 + @ 0 U.R ." K\n"
  REPEAT
  $4100 +3DOS-AF !
  $121 +3DOS-CMD
  +3DOS-ERR? 0IF +3DOS-HL @ 0 U.R ." K free\n" ENDIF ;
[ENDIF]

;; Advance the disc buffer address addr1 to the address of the next buffer
;; addr2. Boolean f is false when addr2 is the buffer presently pointed to by
;; variable PREV.
: +BUF  ( addr1 -- addr2 0-if-prev )
  B/BUF-FULL +
  DUP SYS: LIMIT - 0IF DROP SYS: FIRST ENDIF
  DUP SYS: PREV @ - ;

;; Flush buffer pointer by addr, if necessary.
;; `addr` is system addr, not data addr.
|: (BUF-FLUSH)  ( addr )
  DUP @ DUP $8000 AND 0IF 2DROP EXIT ENDIF
  ( addr blk# )
  $7FFF AND SWAP  ( blk# addr )
  2DUP !  2+ SWAP 0 R/W ;

;; Obtain the next memory buffer, assigning it to block n. If the contents of
;; the buffer is marked as updated, it is written to the external storage.
;; The block is not read from the external storage. The address left is the
;; first cell within the buffer to data storage.
: BUFFER  ( n -- addr )
  SYS: USE @ DUP >R BEGIN +BUF UNTIL SYS: USE !
  R@ @ 0< IF R@ 2+ R@ @ $7FFF AND 0 R/W ENDIF
  R@ ! R@ SYS: PREV ! R> 2+ ;

;; Leave the memory address of the block buffer containing block n. If the
;; block is not already in memory, it is transferred from external storage to
;; which ever buffer was least recently written. If the block occupying that
;; buffer has been marked as updated, it is rewritten to external storage
;; before block n is read into the buffer. See also: BUFFER R/W UPDATE FLUSH
: BLOCK  ( n -- addr )
  SYS: OFFSET @ + >R
  SYS: PREV @ DUP @ R@ - DUP + IF
    BEGIN
      +BUF 0IF DROP R@ BUFFER DUP R@ 1 R/W 2- ENDIF
      DUP @ R@ - DUP +
    0UNTIL
    DUP SYS: PREV !
  ENDIF RDROP 2+ ;

;; Marks the most recently referenced block (pointed to by PREV) as altered.
;; The block will subsequently be transferred automatically to disk should its
;; buffer be required for storage of a different block.
\ : UPDATE  SYS: PREV @ @ $8000 OR SYS: PREV @ ! ;
: UPDATE  $80 SYS: PREV @ 1+ OR!C ;

;; Write all updated disc buffers to the external storage.
\ : FLUSH  #BUFF 1+ FOR 0 BUFFER DROP LOOP ;
;; k8: this is faster, and doesn't "uncache" buffers
: FLUSH  SYS: LIMIT SYS: FIRST DO I (BUF-FLUSH) B/BUF-FULL +LOOP ;


;; Load line with error message.
: (MSGLINE)  ( msg -- addr count )  32 B/BUF */MOD BLOCK + 32 ;

OPT-SCREEN-WORDS? [IF]
;; Convert the line number line# and the screen scr# to the disc buffer address
;; containing the data. A count of 64 indicates the full line text length.
: (LINE)  ( line# scr# -- addr count )  >R 64 B/BUF */MOD R> B/SCR * + BLOCK + 64 ;

;; Print on the screen, a line of text from the RAM disc by its line and
;; screen number. Trailing blanks are suppressed.
: .LINE  ( line scr )  (LINE) -TRAILING TYPE ;

;; Leave address of line n of current screen. This address will be in the RAM-
;; disc buffer area.
: LINE  ( n -- addr )  DUP $FFF0 AND 23 ?ERROR SCR @ (LINE) DROP ;

;; Display the ASCII text of screen n on the selected output device. SCR
;; contains the screen number during and after this process.
OPT-FS-EDITOR-CODE? [IF]
;; cannot use aliases here, "EDIT" is not defined yet
: LIST  ( n ) EDIT ;
[ELSE]
: LIST  ( n )
  DECIMAL DUP SCR !
  ." \c\V\x01\B\x01SCR # " . 0 BRIGHT 0 INVERSE
  16 FOR
    CR 1 INVERSE I 2 .R SPACE 0 INVERSE I SCR @ .LINE
    ?TERMINAL IF LEAVE ENDIF
  LOOP CR ;
[ENDIF]
[ENDIF]


OPT-VERBOSE-LOAD? [IF]
|: START-TIMER
  DI SYS: FRAMES 1+ OFF SYS: FRAMES OFF
  SYS: FRAMES EI HALT HALT OFF ;

|: GET-TIMER  ( -- timer )  SYS: FRAMES @ ;

|: REPORT-TIMER  ( timer )
  \ DUP U.
  20 1000 */MOD ( msecs seconds )
  SWAP 100 /MOD ( seconds mslow mshi )
  SWAP 50 >= + 10 /MOD  ( seconds mshi add )
  ROT + 10 * +
  ." TIME: " 0 <# # [CHAR] . HOLD #S #> TYPE CR ;
[ENDIF]

;; Begin interpretation of screen n. Loading will terminate at the end of the
;; screen or at ;S. See: ;S -->
: LOAD  ( n )
  DUP 0= 9 ?ERROR
  [ OPT-FAST-LOAD? ] [IF] (MSPD-ON) [ENDIF]
  [ OPT-VERBOSE-LOAD? ] [IF] ." \cLSCR #" DUP . CR START-TIMER [ENDIF]
  BLK @ >R IN @ >R
  IN OFF B/SCR * BLK !
  INTERPRET
  R> IN ! R> BLK !
  [ OPT-VERBOSE-LOAD? ] [IF] GET-TIMER ." \cDONE. " REPORT-TIMER [ENDIF]
  [ OPT-FAST-LOAD? ] [IF] (MSPD-OFF) [ENDIF] ;

;; Continue interpretation with the next disc screen.
;; k8: buffer size is guaranteed to be power of 2. let's use it.
*: -->
  SYS: ?LOADING
  IN OFF B/SCR BLK @ OVER UMOD - BLK +! ;


OPT-SCREEN-WORDS? OPT-MORE-SCREEN-WORDS? forth:land [IF]
;; Print the first line of each screen over the range from, to. This is used
;; to view the comment lines of an area of text on RAM-disc screens.
: INDEX  ( from to )
  CLS 1+ SWAP DO
    CR I 3 .R SPACE 0 I .LINE ?TERMINAL IF LEAVE ENDIF
  LOOP ;

;; Display on the selected output device the three screens which include the
;; one numbered scr, beginning with a screen evenly divisible by three. Output
;; is suitable for source text records.
: TRIAD  ( scr )
  CLS 3 UU/ 3 * 3 OVER + SWAP DO
    CR I LIST ?TERMINAL IF LEAVE ENDIF
  LOOP ;
[ENDIF]

OPT-EDITOR-CODE? [IF]
;; If an error occurs during LOAD from disk, ERROR leaves these values on the
;; stack. WHERE uses these values to show the user where the error occurred by
;; printing the screen and line no.
: WHERE  \ nope ( n1 n2 )
  ERR-POS ERR-BLK
  DUP B/SCR UU/ DUP SCR ! ." SCR # " DECIMAL .
  SWAP C/L /MOD C/L * ROT BLOCK + CR
  C/L TYPE CR
  HERE C@ - SPACES [CHAR] ^ EMIT
  [COMPILE] EDITOR QUIT ;noexit
[ENDIF]

<zx-done>
