;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Z80 Abersoft fig-FORTH recompiled
;; Copyright (C) 2024 Ketmar Dark // Invisible Vector
;; Understanding is not required. Only obedience.
;; GPLv3 ONLY
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; word creation
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


<zx-definitions>

: CREATE  \ name
  ;; create LFA for the new header format.
  ;; this is because WORD copies the string to HERE
  [ OPT-NEW-ART-HEADER? ] [IF]
    [ OPT-NEW-ART-HEADER-HASHED? ] [IF] 0 C, [ENDIF] ;; hash will be put here
    ( lfa) LATEST-VLINK ,
  [ENDIF]
  -FIND IF DROP NFA ID. SPACE 4 MESSAGE ENDIF
  [ OPT-NEW-ART-HEADER? OPT-NEW-ART-HEADER-HASHED? forth:and ] [IF]
    ;; calculate hash
    HERE COUNT SYS: WIDTH @ MIN HASH-NAME HERE 3 - C!
  [ENDIF]
  HERE DUP C@ SYS: WIDTH @ MIN 1+ ALLOT DUP $A0 TOGGLE
  HERE 1- $80 TOGGLE
  ;; link
  [ OPT-NEW-ART-HEADER? ] [IF] 2- [ELSE] LATEST-VLINK , [ENDIF] CURRENT @ !
  ;; create CFA (for code word)
  [ OPT-DTC? ] [IF] ( compile JP) $C3 C, [ENDIF]
  HERE 2+ , ;

;; set latest CFA to the given code address
|: !CODE  ( code-addr )
  LATEST-CFA  ( code-addr cfa )
  [ OPT-DTC? ] [IF]
    $CD OVER C! ;; CALL
    1+
  [ENDIF] ! ;

: VOCABULARY
  CREATE
  \ IMMEDIATE   ;; k8: for some reason vocabs are immediate; this was missing in the original
  [ OPT-NEW-ART-HEADER? ] [IF]
    CURRENT @ ,
  [ELSE]
    ( reserved) 0 ,
  [ENDIF]
  ( dummy header) $A081 ,
  [ OPT-NEW-ART-HEADER? ] [IFNOT]
    ( latest) CURRENT @ CFA ,
  [ENDIF]
  ( voclink) HERE SYS: VOC-LINK @ , SYS: VOC-LINK !
  [ZX-DO-VOCAB] !CODE ;

: DEFINITIONS  CONTEXT @ CURRENT ! ;

*: [  SYS: STATE OFF ;
: ]  $C0 SYS: STATE ! ;


: CONSTANT  ( value )  \ name
  CREATE SMUDGE , [ZX-DO-CONSTANT] !CODE ;

: VARIABLE  ( value )  \ name
  CONSTANT [ZX-DO-VARIABLE] !CODE ;

\ TODO: restore this!
\ : USER  ( value )  \ name
\   CONSTANT [ZX-DO-USERVAR] !CODE ;

: 2CONSTANT  ( d )  \ name
  CREATE SMUDGE HERE 2! 4 ALLOT [ZX-DO-2CONSTANT] !CODE ;

: 2VARIABLE  ( d )  \ name
  2CONSTANT [ZX-DO-2VARIABLE] !CODE ;

: QUAN  ( value )  \ name
  CONSTANT
  [ OPT-DTC? ] [IF] ( compile CALL) $CD C, [ENDIF] [ZX-DO-QADDR] ,
  [ OPT-DTC? ] [IF] ( compile CALL) $CD C, [ENDIF] [ZX-DO-Q-TO] ,
  [ OPT-DTC? ] [IF] ( compile CALL) $CD C, [ENDIF] [ZX-DO-Q-+TO] ,
  [ZX-DO-QUAN] !CODE ;

: VECT  ( value )  \ name
  QUAN [ZX-DO-VECT] !CODE ;


: <BUILDS  0 VARIABLE ;
: DOES>  R> LATEST-PFA ! [ZX-DO-DOES] !CODE ;


: ALIAS-FOR  \ old-name IS new-name
  -FIND-RQ CFA
  BL WORD HERE UPSTR
  HERE @ $4902 - HERE 2+ C@ $53 - OR 5 ?ERROR
  CREATE SMUDGE
  ;; update cfa
  DUP SYS: CFA@ OVER CFA>PFA -
  IF  ;; not a code word
    [ OPT-DTC? ] [IF] ( compile CALL) $CD HERE 3 - C! [ENDIF]
    [ZX-DO-ALIAS] HERE 2- ! ,
  ELSE  ;; code word
    HERE 2- !
  ENDIF ;


*: :  SYS: ?EXEC SYS: !CSP CURRENT @ CONTEXT ! CREATE [COMPILE] ] [ZX-DO-COLON] !CODE ;
*: ;  SYS: ?CSP COMPILE SYS: ;S SMUDGE [COMPILE] [ ;


*: Q/V?  ( pfa -- flag )
  CFA [ OPT-DTC? ] [IF] 1+ [ENDIF] @
  DUP [ZX-DO-QUAN] = SWAP [ZX-DO-VECT] = OR ;

|: -Q/V  ( -- pfa )  \ name
  -FIND-RQ DUP Q/V? 0= 0 ?ERROR ;

*: QADDR  ( -- pfa )  \ name
  -Q/V STATE@ IF 2+ SYS: CALL, ENDIF ;

*: TO  \ name
  -Q/V STATE@ IF [ OPT-DTC? ] [IF] 5 [ELSE] 4 [ENDIF] + SYS: CALL, EXIT ENDIF ! ;

*: +TO  \ name
  -Q/V STATE@ IF [ OPT-DTC? ] [IF] 8 [ELSE] 6 [ENDIF] + SYS: CALL, EXIT ENDIF +! ;


*: (  \ )
  41 WORD ;

*: \  \ comment to the end of the line
  BLK @ 0IF RDROP EXIT ENDIF
  ;; hack: we always have a line in TIB, because of our buffer size.
  ;; so simply align IN, and call it a day.
  ;; we should use "C/L" constant here, but meh...
  ;; for non 64-columns mode, use line of 32 bytes.
  IN @ 31 + 31 ~AND IN ! ;

<zx-done>
