;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Z80 Abersoft fig-FORTH recompiled
;; Copyright (C) 2024 Ketmar Dark // Invisible Vector
;; Understanding is not required. Only obedience.
;; GPLv3 ONLY
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; interpreter and other interactive things
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


<zx-definitions>

: EXPECT  ( addr count )
  OVER + OVER DO  ( addr )
    KEY
    DUP 12 = IF DROP
      DUP I = IF ERR-BEEP 0 -1 ELSE 8 EMIT BL EMIT 8 -2 ENDIF
      R> + >R
    ELSE
      DUP 7 = IF ." \V\x01^C\V\x00\n" 0AND RDROP OVER 1- >R
      ELSE
        DUP 13 = IF LEAVE DROP BL 0
        ELSE DUP BL < IF DROP [CHAR] ? ENDIF DUP ENDIF
        I C! I 1+ OFF
      ENDIF
    ENDIF
    DUP IF EMIT ELSE DROP ENDIF
  LOOP DROP ;

: QUERY  KCUR C@ 1 AND 0IF KCUR ON ENDIF  TIB @ 80 EXPECT IN OFF ;

\ : ?STACK  SP@ (SP0) @ SWAP U< 1 ?ERROR SP@ HERE 128 + U< 7 ?ERROR ;
<zx-system>
: ?STACK  SYS: DEPTH 0< 1 ?ERROR SP@ HERE 128 + U< 7 ?ERROR ;
<zx-forth>

;; special word, name is zero byte.
*: {00}
  BLK @ 0IF RDROP EXIT ENDIF
  BLK 1+! IN OFF
  ;; check if we are done with the current screen
  BLK @
  [ zx-b/scr forth:pot 1 forth:> ] [IF] B/SCR 1- AND
  [ELSE] 0 B/SCR U/MOD DROP [ENDIF]
\ endcr ." BLK=" BLK @ . ." B/SCR=" B/SCR . ." SCRX=" DUP . CR
  ?EXIT SYS: ?EXEC RDROP ;

: INTERPRET
  BEGIN SYS: ?STACK
    -FIND IF
      STATE@ < IF CFA SYS: CALL, CONTINUE ENDIF
      CFA EXECUTE CONTINUE
    ENDIF
    HERE NUMBER
    DPL @ 1+ 0IF DROP [COMPILE] LITERAL CONTINUE ENDIF
    [COMPILE] DLITERAL
  AGAIN ;noexit


|: (.BASE)
  BASE @ DUP 10 -
  IF [CHAR] [ EMIT DUP DECIMAL 0 .R BASE ! [CHAR] ] EMIT EXIT ENDIF DROP ;

|: (.DEPTH)
  SYS: DEPTH DUP
  IF BASE @ DECIMAL [CHAR] ( EMIT SWAP 0 .R BASE ! [CHAR] ) EMIT EXIT ENDIF DROP ;

: (QUIT)
  [ OPT-FAST-LOAD? ] [IF] (MSPD-OFF) [ENDIF]
  BLK OFF [COMPILE] [
  BEGIN
    SYS: RP!
    (FIX-ATTRS) ENDCR
    STATE@ IF [CHAR] > EMIT ENDIF [CHAR] > EMIT
    QUERY ENDCR
    INTERPRET
    STATE@ 0IF
      ?SPACE
      INVERSE@ >R 1 INVERSE
      (.BASE) (.DEPTH) ." ok"
      R> INVERSE
    ENDIF
  AGAIN ;noexit

['] (QUIT) TO QUIT
['] NOOP TO MAIN

: ABORT
  SYS: RP! SYS: SP! [ OPT-ENABLE-FP? ] [IF] SYS: FP0! [ENDIF]
  DECIMAL
  [COMPILE] FORTH DEFINITIONS
  [ OPT-PARNAS? ] [IF] PARNAS-SYS-RESET [ENDIF]
  MAIN QUIT ;noexit

;; set default abort word for "ERROR"
['] ABORT TO (ABORT)

<zx-system>
;; basic error will be re-routed here
: (BERR)
  SYS: RP! SYS: SP! [ OPT-ENABLE-FP? ] [IF] SYS: FP0! [ENDIF]
  DECIMAL (FIX-ATTRS)
  ." \cBERR #" SYS: ERR-NR C@ C>S 1+ . CR
  ABORT ;noexit

code: UV0!
  push  bc
  ;; fix print values here, why no?
  ;; we need to have at least one line in the bottom area
  ld    a, # 1
  ld    sysvar-defsz (), a
  ;; open chan #2
  ld    a, # 2
  call  # $1601
  pop   bc

  ;; print whole 24 lines
  xor   a
  ld    sysvar-defsz (), a
  ;; no scroll prompt
  dec   a
  ld    sysvar-scr-ct (), a

  ld    hl, # zx-word-par-baserr
  ld    zx-baserr-stack (), hl
  ld    hl, # zx-baserr-stack
  ld    sysvar-err-sp (), hl
  ld    hl, # zx-uservars-origin
  ld    zx-uv (), hl

  ld    hl, # zx-['] (CEMIT)
  ld    zx-['pfa] EMIT (), hl
;code
<zx-forth>

: .BANNER
  \ CLS
  [ OPT-64-COLS? ] [IF]
  ." \WZX SPECTRUM fig-FORTH 1.1B, \x7F Abersoft:1983\nmodified by Ketmar Dark, 2024\n\n"
  [ELSE]
  ." \WZX SPECTRUM fig-FORTH 1.1B\n\x7F Abersoft:1983\nmodified by Ketmar Dark, 2024\n\n"
  [ENDIF] ;

: WARM
  UV0! .BANNER EMPTY-BUFFERS ABORT ;noexit

: COLD
  \ EMPTY-BUFFERS
  UV0!
  $0010 SYS: RASP !
  SYS: FIRST DUP SYS: USE ! SYS: PREV !
  SYS: DR0
  [ OPT-BLOCK-I/O-TYPE R/W-ZXEMUT = ] [IF] WARNING ON [ENDIF]
  [ OPT-BLOCK-I/O-TYPE R/W-P3DOS = ] [IF] INIT+3DOS [ENDIF]
  ;; init uservars
  [ zx-cold-uvars^ ] {#,} [ zx-uvars-origin^ ] {#,} [ zx-cold-#uvars ] {#,} CMOVE
  ;; reset DP, FENCE, VOCLINK
  [ zx-dpx-init^ ] {#,} @ SYS: (DP) !
  [ zx-dpx-init^ 2 + ] {#,} @ SYS: FENCE !
  [ zx-dpx-init^ 4 + ] {#,} @ SYS: VOC-LINK !
  ;; reset FORTH-LATEST
  [ zx-dpx-init^ 6 + ] {#,} @ [ zx-forth-vocid ] {#,} !
  ;; reset EDITOR-LATEST
  [ zx-dpx-init^ 8 + ] {#,} @ [ zx-editor-vocid ] {#,} !
  ( ABORT) WARM ;noexit


code|: (WARM)
zx-word-warm:
  ld    bc, # zx-['pfa] WARM

zx-do-warm:
  ;; setup error handler
  ld    hl, # zx-word-par-baserr
  ld    zx-baserr-stack (), hl
  ld    hl, # zx-baserr-stack
  ld    sysvar-err-sp (), hl

  ld    hl, () zx-uservars-origin
  ld    zx-uv (), hl
  ld    sp, () zx-cold-uvars
  ld    hl, () zx-cold-uvars 2+
  ld    zx-rp (), hl

  push  bc
  ld    a, # 2
  call  # $1601
  pop   bc

  ld    ix, # i-next
;code

code|: (COLD)
zx-word-cold:
  di
  ld    sp, () zx-cold-uvars

  ld    a, # 8
  ld    sysvar-flags2 (), a

  ld    hl, # zx-['pfa] +3DOS?
  ld    (hl), # 0   ;; default is no +3DOS

  ;; setup 128K state (it won't hurt 48K)
  ld    bc, # $7FFD
  ld    a, # $10    ;; ROM1 (48K), RAM 520
  out   (c), a

  ;; test for 128K
  ;; map $4000 at $C000, and trash the byte
  ld    hl, # zx-['pfa] 128?
  ld    (hl), # 0   ;; default is 48K
  ;; save byte at $C000 for 48K
  ex    de, hl
  ld    hl, # $C000
  ld    a, (hl)
  ex    af, afx
  ;; trash byte at $C000
  xor   a
  ld    (hl), a
  ;; switch memory page
  ;;ld    bc, # $7FFD
  ld    a, # $15    ;; ROM1 (48K), RAM 525
  out   (c), a      ;; now $C000 should be the screen page #5
  ld    (hl), a     ;; trash it
  ;; restore RAM mapping (just in case)
  ld    a, # $10    ;; ROM1 (48K), RAM 520
  out   (c), a
  ;; check SCR$ byte
  ld    a, () $4000
  or    a
  ;; restore $C000 byte
  ex    af, afx
  ld    (hl), a
  ex    af, afx
  jr    z, # .model-is-48k
  ;; 128K model
  ex    de, hl
  inc   (hl)
  ;; set 7FFD sysvar
  ld    a, # $10
  ld    sysvar-cur-7FFD (), a

  ;; assume that on +3 we came here from a disk loader,
  ;; so disk motor is on. it is controlled by bit 4 of
  ;; FLAGS sysvar, so let's use it. bit 4 should be reset
  ;; on 48K. dunno about 128K.

  ;; this sysvar is 0 for 48K/128K
  ld    a, () sysvar-cur-1FFD
  or    a
  jr    z, # .not-plus-n

  ;; check if +3DOS ROM is available.
  ;; 128K control port:
  xor   a
  out   (c), a
  ;; set +2A/+3 port #1FFD
  ld    b, # $1F
  ld    a, () sysvar-cur-1FFD
  or    # $04               ;; switch on +3DOS ROM
  out   (c), a

  ;; at $0000 +3DOS ROM has zeroes
  ld    hl, () $0000
  ld    a, l
  or    h
  jr    nz, # .no-plus3dos-rom
  ;; at $0008 +3DOS ROM has "PLUS3DOS" string.
  ;; check "PL" and "3D".
  ld    hl, () $0008
  ld    de, # $4C50
  sbc   hl, de
  jr    nz, # .no-plus3dos-rom
  ld    hl, () $000C
  ld    de, # $4433
  sbc   hl, de
  jr    nz, # .no-plus3dos-rom
  ;; set the flag
  ld    a, # 1
  ld    zx-['pfa] +3DOS? (), a
.no-plus3dos-rom:
  ;; switch to 48K BASIC ROM
  ld    a, () sysvar-cur-1FFD
  and   # $F8
  or    # $04       ;; bit 2: BASIC ROM
  out   (c), a
  ld    sysvar-cur-1FFD (), a
  ;; as machines without #1FFD may map this to #7FFD, restore #7FFD value
  ld    b, # $7F
  ld    a, () sysvar-cur-7FFD
  out   (c), a
.not-plus-n:

.model-is-48k:
  ;; interrupts are disabled here

  OPT-IM2-HANDLER? [IF]
  ;; setup interrupt table
  ;; this is using an old trick:
  ;; 48K ROM has a lot of #FF (more than 256),
  ;; so we can use it as interrupt table.
  ;; this way, the interrupt will jump at #FFFF.
  ;; this is not that useful... but we can put "JR"
  ;; opcode there, and it will jump to #FFF4
  ;; (due to first ROM byte being DI opcode aka #F3).
  ;; and we can put a real jump to our interrupt
  ;; routine at #FFF4.
  ;; so we cannot use last UDG, but meh... not a big deal.

  ;; check if ROM starts with "#F3" (this will be JR disp)
  ld    a, () 0
  cp    # $F3
  jr    nz, # .cannot_use_im2

  ld    hl, # $3BFF
  ld    a, (hl)
  inc   (hl)      ;; smaller than load and compare, and ROM ignores changes anyway
  jr    nz, # .cannot_use_im2
  inc   hl
  inc   (hl)      ;; smaller than load and compare, and ROM ignores changes anyway
  jr    nz, # .cannot_use_im2

  \ di
  ld    a, # $3B  ;; ROM contains a lot of #FF there
  ;; I register points to area filled with #FF here
  ld    i, a
  ;; set "IM2?" flag (active)
  ld    a, # 1
  ld    zx-['pfa] IM2? (), a
  ;; write JR opcode to #FFFF
  ld    a, # $18
  ld    $FFFF (), a   ;; jr
  ;; write JP opcode to #FFF4
  ld    a, # $C3      ;; jp
  ld    $FFF4 (), a
  ;; write interrupt handler address to #FFF5
  ld    hl, # .IntrHandler
  ld    $FFF5 (), hl
  ;; switch to second interrupt mode
  im    # 2
  [ENDIF]

.im-setup-done:
  ;; we can enable interrupts now
  ei

  ;; go on
  ld    ix, # i-next
  ld    bc, # zx-['pfa] COLD
  jp    # zx-do-warm

  OPT-IM2-HANDLER? [IF]
.IntrHandler:
  push  af
  push  hl
  ;; call ROM IM1 handler before user handler?
  ;; 0: no; 1: before; 2: after
  ld    a, # 1    ;; self-modifying code
;; values: 0 -- don't call; 1: before; 2: after
$here 1- @def: zx-im2-rom-im1
  dec   a
  call  z, # $0038
  di              ;; because ROM enabled it
  ;; call user handler
  call  # dummy-im2-userproc  ;; self-modifying code
$here 2- @def: zx-im2-userproc-addr
  ;; call ROM IM1 handler after user handler?
  ld    a, () zx-im2-rom-im1
  cp    # 2
  call  z, # $0038
  pop   hl
  pop   af
  ei
  ret   ;; it should be reti, but meh

.cannot_use_im2:
  xor   a
  ld    zx-['pfa] IM2? (), a
  jr    # .im-setup-done

@dummy-im2-userproc:
  ret
  [ENDIF]
;code-no-next

;; return from basic error
code|: (BASERR)
zx-word-par-baserr:
  ld    bc, # zx-sys['pfa] (BERR)
  jp    # zx-do-warm
flush!
  ;; reserve some room for error stack.
  ;; this in in case we'll get an interrupt here.
  0 dw, 0 dw, 0 dw, 0 dw,
  0 dw, 0 dw, 0 dw, 0 dw,
  0 dw, 0 dw, 0 dw, 0 dw,
  0 dw, 0 dw,
zx-baserr-stack:
  0 dw, 0 dw,
;code-no-next


<zx-done>
