;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Z80 Abersoft fig-FORTH recompiled
;; Copyright (C) 2024 Ketmar Dark // Invisible Vector
;; Understanding is not required. Only obedience.
;; GPLv3 ONLY
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; EDITOR words
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


<zx-definitions>
<zx-editor>
zx-here

;; From the cursor position, determine the line-no n2 and the offset into line n1.
: #LOCATE ( -- n1 n2 )  R# @ C/L /MOD ;
: #LEAD   ( -- line-address offset-to-cursor )  #LOCATE LINE SWAP ;
: #LAG    ( -- cursor-address count-after-cursor-till-EOL )  #LEAD DUP >R + C/L R> - ;

;; Move a line of text from addr to line of current screen.
: -MOVE  ( addr line-no -- )  LINE C/L CMOVE UPDATE ;

;; Hold numbered line at PAD.
: H  ( n )  LINE PAD 1+ C/L DUP PAD C! CMOVE ;

;; Erase line n with blanks.
: E  ( n )  LINE C/L BLANKS UPDATE ;

;; Spread. Lines n and following move down. n becomes blank.
: S  ( n )  DUP ( 1 -)1- 14 DO I LINE I 1+ -MOVE -1 +LOOP E ;

;; Delete line n, but hold in PAD.
: D  ( n )
  DUP H 15 DUP ROT DO I 1+ LINE I -MOVE LOOP E ;

;; Move cursor by a signed amount and print its line.
: M  ( n )  R# +! CR SPACE #LEAD TYPE 95 EMIT #LAG TYPE #LOCATE . DROP ;

;; Type line n and save in PAD.
: T  ( n )  DUP C/L * R# ! DUP H 0 M ;

;; List the current screen.
: L  SCR @ LIST 0 M ;

;; Replace line n with the text in PAD.
: R  ( n )  PAD 1+ SWAP -MOVE ;

;; Put the following text on line n.
: P  ( n ) ( text )  1 TEXT R ;

;; Spread at line n and insert text from PAD.
: I  ( n )  DUP S R ;

;; Position editing cursor at top of screen.
: TOP  0 R# ! ;

;; Clear screen n, can be used to select screen n for editing.
: CLEAR  ( n )  SCR ! 16 FOR I E LOOP ;

;; Copy screen n1 to screen n2.
: COPY  ( n1 n2 )
  B/SCR * OFFSET @ + SWAP B/SCR * B/SCR OVER + SWAP
  DO DUP I BLOCK ( 2 -)2- ! 1+ UPDATE LOOP
  DROP FLUSH ;

;; Match the string at add2 with all strings on the cursor line forward from
;; the cursor. If found leaves n3 bytes until the end of the matching string,
;; else leaves n4 bytes to EOL.
;; k8: WTF?!
: -TEXT  ( add1 count add2 -- ??? )
  SWAP -DUP IF
    OVER + SWAP DO DUP C@ I C@ - IF 0= LEAVE ELSE 1+ ENDIF LOOP
  ELSE DROP 0= ENDIF ;

: MATCH  ( add1 n1 add2 n2 -- tf n3 )
  >R >R 2DUP R> R> 2SWAP OVER + SWAP DO
    2DUP I -TEXT IF >R 2DROP R> - I SWAP - 0 SWAP 0 0 LEAVE ENDIF
  LOOP 2DROP SWAP 0= SWAP ;

;; Scan the cursor line for a match to PAD text. Return flag and update R#
;; to the end of matching text, or to the start of the next line it no match
;; found.
: 1LINE  ( -- flag )  #LAG PAD COUNT MATCH R# +! ;

;; Search for a match to the string at PAD, from the cursor position until the
;; end of screen. If no match found issue an error message and reposition the
;; cursor at the top of the screen.
: FIND
  BEGIN 1023 R# @ < IF TOP PAD HERE C/L 1+ CMOVE 0 ERROR ENDIF 1LINE UNTIL ;

;; Delete n characters prior to the cursor.
: DELETE  ( n )
  >R #LAG + R - #LAG R MINUS R# +! #LEAD + SWAP CMOVE R> BLANKS UPDATE ;

;; Find next occurrence of PAD text.
: N  FIND 0 M ;

;; Input following text to PAD and search for match from cursor position till
;; end of screen.
: F  1 TEXT N ;

;; Backup cursor by text in PAD.
: B  PAD C@ MINUS M ;

;; Delete next occurrence of following text.
: X  1 TEXT FIND PAD C@ DELETE 0 M ;

;; Delete on cursor line from cursor to the end of the following text.
: TILL  #LEAD + 1 TEXT 1LINE 0= 0 ?ERROR #LEAD + SWAP - DELETE 0 M ;

;; Spread at cursor and copy the following text into the cursor line.
: C
  1 TEXT PAD COUNT #LAG ROT OVER MIN >R R R# +! R - >R
  DUP HERE R CMOVE HERE #LEAD + R> CMOVE R> CMOVE UPDATE 0 M ;

zx-here swap - endcr ." line editor size: " .bytes cr

<zx-forth>
<zx-done>
