;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Z80 Abersoft fig-FORTH recompiled
;; Copyright (C) 2024 Ketmar Dark // Invisible Vector
;; Understanding is not required. Only obedience.
;; GPLv3 ONLY
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; FULLSCREEN EDITOR words
;; WARNING! this code is not good Forth code at all!
;;          i had to sacrifice some clarity for speed and size.
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


<zx-definitions>
<zx-editor>

zx-here

<zx-hidden>
OPT-FS4-64? [IF]
16 constant SCR-LINES
16 constant VIS-LINES
64 constant COLS  -- WARNING! DO NOT CHANGE!
[ELSE]
32 constant SCR-LINES
23 constant VIS-LINES
32 constant COLS  -- WARNING! DO NOT CHANGE!
0 quan TOP-LINE
[ENDIF]
0 quan CUR-LINE   ;; current line, counting from TOP-LINE
0 quan CUR-COL    ;; current offset in line
0 quan STMSG?     ;; status message shown? countdown to full update
false quan DONE?  ;; "exit main editor loop" flag

;; refresh types
\ zx-['] noop vect REFRESH
0 quan RF-FROM  ;; fefresh from this visible line
0 quan RF-COUNT ;; refresh this number of lines

editor-shadows:VIS-LINES 1- constant LAST-VIS-LINE
editor-shadows:SCR-LINES 1- constant LAST-SCR-LINE
editor-shadows:COLS 1- constant LAST-COL
\ <zx-normal>


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; useful primitives (coded in asm for speed)

code: MEMCHR  ( addr count ch -- addr1 count1 TRUE // FALSE )
  exx
  pop   de        ;; ch
  pop   bc        ;; count
  pop   hl
  ld    a, b
  or    c
  jr    z, # .not-found
  ld    a, e
  cpir
  jr    nz, # .not-found
  dec   hl
  inc   bc
  push  hl
  push  bc
@zx-memchr-pop-bc-exit-push-true:
  ld    hl, # 1
  jr    # .done-push-hl
.not-found:
@zx-memchr-pop-bc-push-false:
  ld    hl, # 0
.done-push-hl:
  push  hl
  exx
;code


code: MEMEQU  ( a0 a1 count -- flag )
  exx
  pop   bc        ;; count
  pop   de        ;; a1
  pop   hl        ;; a0
  ld    a, b
  or    c
  jr    z, # zx-memchr-pop-bc-push-false
.loop:
  ld    a, (de)
  cp    (hl)
  jr    nz, # zx-memchr-pop-bc-push-false
  inc   hl
  inc   de
  dec   bc
  ld    a, b
  or    c
  jp    nz, # .loop
  jr    # zx-memchr-pop-bc-exit-push-true
;code-no-next


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; read/write the whole RAM-DISC
;; ZXEmuT-specific!

OPT-BLOCK-I/O-TYPE R/W-TAPE = [IF]
code|: ZXEMUT-SAVE-RAMDISC
  push  bc
  ld    hl, # zxemut-disc-fname
  ld    de, # zx-ramdisc^
  ld    bc, # zx-#ramdisc
flush!
  ;; ZXEmuT trap
  $ED db, $FE db, $18 db,
  2 db, $13 db, $14 db,
flush!
  pop   bc
  next
flush!
zxemut-disc-fname:
  [CHAR] D db,
  [CHAR] I db,
  [CHAR] S db,
  [CHAR] C db,
  0 db,
;code-no-next

code|: ZXEMUT-LOAD-RAMDISC
  push  bc
  ld    hl, # zxemut-disc-fname
  ld    de, # zx-ramdisc^
  ld    bc, # zx-#ramdisc
flush!
  ;; ZXEmuT trap
  $ED db, $FE db, $18 db,
  2 db, $13 db, $12 db,
flush!
  pop   bc
;code
[ENDIF]


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; some ASM routines, for speed

;; check if the given block contains only 0 or BL chars
|: BUF-EMPTY?  ( buf# -- flag )
  TRUE SWAP BLOCK
  B/BUF FOR C@++ BL > IF 0AND SWAP LEAVE ENDIF LOOP
  DROP ;

|: SCR-EMPTY?  ( scr# -- flag )
  TRUE SWAP B/SCR * DUP B/SCR + SWAP DO
    I BUF-EMPTY? 0IF 0AND LEAVE ENDIF
  LOOP ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; mid-level block routines

;; find empty screen. used in "insert screen" code.
\ FIXME: obey block limit!
|: SCR-FIND-EMPTY  ( scr#-from -- scr# TRUE // FALSE )
  BEGIN DUP SCR-EMPTY? 0WHILE 1+ REPEAT TRUE ;

;; copy screen.
;; this uses the hack: it changes block number directly.
|: SCR-COPY  ( from to )
  B/SCR * SWAP B/SCR * B/SCR OVER + SWAP
  ( to-block )
  DO DUP I BLOCK 2- ! 1+ UPDATE LOOP
  DROP ;

;; fill screen with the given byte.
|: SCR-FILL  ( scr# byte )
  SWAP B/SCR * DUP B/SCR + SWAP DO
    I BLOCK OVER B/BUF SWAP FILL UPDATE
  LOOP DROP ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; high-level block routines (used as kb handlers)

|: SCR-BLANK  SCR @ BL SCR-FILL RF-FULL ;
|: SCR-ERASE  SCR @  0 SCR-FILL RF-FULL ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; wait for keypress, blink cursor

|: WAIT-KEY  ( -- char )
  KCUR C@ ALLOW-G? C@
  $81 KCUR C! ALLOW-G? OFF
  CUR-LINE CUR-COL AT KEY-NC
  NROT ALLOW-G? C! KCUR C! ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; find key handler in a key table

|: FIND-KEY  ( key list^ -- cfa TRUE // FALSE )
  SWAP >R BEGIN ( list^ | key )
    C@++ DUP R@ - 0IF DROP RDROP @ TRUE EXIT ENDIF
  SWAP 2+ SWAP 0UNTIL DROP RDROP FALSE ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; various helper words

;; after PAD
|: FIND-BUF  ( -- addr )  PAD 128 + ;

|: SVPOS^  ( idx -- addr )
  2* 2* [ OPT-FS4-64? ] [IFNOT] 2* ( waste some memory, meh) [ENDIF]
  FIND-BUF 68 + + ;

|: SET-DONE  TRUE TO DONE? ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; setup initial attrs and such

|: NORM-ATTRS
  \ 0 INK 7 PAPER
  INK@ INK PAPER@ PAPER
  0 DUP DUP DUP INVERSE GOVER FLASH BRIGHT
  RF-NONE ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; buffer helpers

;; get line addr, and cache it.
|: LADDR  ( line# -- addr )
 [ OPT-FS4-64? ] [IF]
  SCR @ (LINE) DROP
 [ELSE]
  COLS B/BUF */MOD  ( ofs bdelta )
  SCR @ B/SCR * + BLOCK +
 [ENDIF] ;

OPT-FS4-64? [IF]
ALIAS-FOR CUR-LINE IS CL#
[ELSE]
|: CL#  ( -- line# )  TOP-LINE CUR-LINE + ;
[ENDIF]

|: CUR-LADDR  ( -- addr )  CL# LADDR ;

;; WARNING! `pos` must be valid!
|: C-ADDR  ( pos -- addr )  CUR-LADDR + ;

|: COL-C@  ( pos -- char )
  DUP COLS U< IF C-ADDR C@ ELSE 0AND ENDIF ;

|: COL-C!  ( char pos )
  DUP COLS U< IF C-ADDR C! UPDATE ELSE 2DROP ENDIF ;

|: CC-ADDR  ( -- addr )  CUR-COL C-ADDR ;
|: CUR-C@   ( -- char )  CUR-COL COL-C@ ;

|: #REST-CH  ( -- cols-cur-ofs )  COLS CUR-COL - ;

|: LINE-LEN    ( lidx -- len )  LADDR COLS -TRAILING NIP ;
|: EMPTY-LINE? ( lidx -- flag ) LINE-LEN 0= ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; print one screen line with correct handling of unprintables

|: XBRIGHT  ( -- old-attr-t )  ATTR-T C@ DUP $40 OR ATTR-T C! ;
|: XBREST   ( old-attr-t )     ATTR-T C! ;

;; redraw the line
|: XLINE  ( addr )
  [ 1 ] [IF]
  COLS OVER @ $203A - 0IF ( addr count )
    OVER 2+ OVER 2- BL MEMCHR IF
      ( addr count restaddr restcount )
      1- [ OPT-UNDER-MATH? ] [IF] UNDER+1 [ELSE] SWAP 1+ SWAP [ENDIF]
      >R >R ( addr count | restcount restaddr )
      DROP R@ OVER - XBRIGHT NROT XTYPE XBREST
      R> R>
    ENDIF
  ENDIF
  [ELSE] COLS [ENDIF]
  XTYPE
  [ OPT-32-COLS? ] [IFNOT]
    [ OPT-FS4-64? ] [IFNOT] XBRIGHT [CHAR] | EMIT XBREST [ENDIF]
  [ENDIF] ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; screen refreshing

|: REFRESH-FROM-VLINE  ( vline# )
  0 TO RF-COUNT
  0MAX  DUP VIS-LINES >= IF DROP EXIT ENDIF
  DUP 0 AT  ;; start from here
  [ OPT-FS4-64? ] [IF]
    VIS-LINES
  [ELSE]
    TOP-LINE + TOP-LINE VIS-LINES + SCR-LINES MIN
  [ENDIF]
  SWAP DO I LADDR XLINE CR LOOP ;

|: REFRESH
  RF-COUNT NOT?EXIT
  RF-COUNT 1- IF RF-FROM REFRESH-FROM-VLINE EXIT ENDIF
  0 TO RF-COUNT
  RF-FROM DUP 0 VIS-LINES WITHIN 0IF DROP EXIT ENDIF
  DUP 0 AT
  [ OPT-FS4-64? ] [IFNOT]
    TOP-LINE + DUP SCR-LINES < 0IF DROP EXIT ENDIF
  [ENDIF]
  LADDR XLINE ;

|: RF-NONE  0 TO RF-COUNT ;
|: RF-FULL  0 TO RF-FROM VIS-LINES TO RF-COUNT ;

|: RF-VLINE#  ( vlidx )
  DUP 0 VIS-LINES WITHIN 0IF DROP EXIT ENDIF
  RF-COUNT 0IF TO RF-FROM 1 TO RF-COUNT EXIT ENDIF
  DUP RF-FROM - 0IF DROP EXIT ENDIF
  RF-FROM MIN TO RF-FROM VIS-LINES TO RF-COUNT ;

|: RF-CLINE  CUR-LINE RF-VLINE# ;
|: RF-PLINE  CUR-LINE ?DUP IF 1- RF-VLINE# ENDIF ;

|: RF-FROM-CLINE
  CUR-LINE RF-COUNT IF RF-FROM MIN ENDIF TO RF-FROM
  VIS-LINES TO RF-COUNT ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; save and restore current text position
;; used in incremental search code

;; used in incremental find
|: SAVE-POS  ( idx )
  SVPOS^
  CUR-COL OVER ! 2+
  [ OPT-FS4-64? ] [IF]
    CUR-LINE SWAP !
  [ELSE]
    CUR-LINE OVER ! 2+
    TOP-LINE SWAP !
  [ENDIF] ;

;; used in incremental find
|: RESTORE-POS  ( idx )
  SVPOS^
  DUP @ TO CUR-COL 2+
  [ OPT-FS4-64? ] [IF]
    @ TO CUR-LINE
  [ELSE]
    DUP @ TO CUR-LINE 2+
    @ DUP TOP-LINE - IF TO TOP-LINE RF-FULL ELSE DROP ENDIF
  [ENDIF] ;

|: SAVE-CUR-POS     FIND-BUF C@ SAVE-POS ;
|: RESTORE-CUR-POS  FIND-BUF C@ RESTORE-POS ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; print various status lines

|: .N2  ( n base )  BASE @ SWAP BASE ! SWAP 0 <# # # #> TYPE BASE ! ;

|: .ST-LINE  VIS-LINES 5 AT CL# 10 .N2 ;
|: .ST-POS   VIS-LINES 12 AT CUR-COL 10 .N2 ;
|: .ST-CHAR  VIS-LINES 20 AT CUR-C@ 16 .N2 ;

|: START-FULL-STATUS  VIS-LINES 0 AT 1 INVERSE OUT OFF 0 TO STMSG? ;
|: CLEAR-REST-STATUS  BEGIN OUT @ #COLS < WHILE SPACE REPEAT 0 INVERSE ;

|: REFRESH-STATUS
  1 INVERSE .ST-LINE .ST-POS .ST-CHAR 0 INVERSE ;

|: DRAW-FULL-STATUS
  START-FULL-STATUS
  ." LINE:   POS:    CH:$   SCR#"
  BASE @ DECIMAL SCR @ . BASE !
  CLEAR-REST-STATUS
  REFRESH-STATUS ;

|: UPDATE-STATUS
  STMSG? 0IF REFRESH-STATUS EXIT ENDIF
  -1 +TO STMSG? STMSG? ?EXIT DRAW-FULL-STATUS ;


|: IC-TYPE  ( addr count )
  OVER + SWAP DO
    I C@ DUP [CHAR] ~ - IF EMIT 1 ELSE 0AND ENDIF INVERSE
  LOOP ;

|: SMSG  ( addr count )
  START-FULL-STATUS IC-TYPE CLEAR-REST-STATUS
  2 TO STMSG? ;

|: MARK-FULL-STATUS-UPDATE  1 TO STMSG? ;


|: DRAW-EXT-STATUS
  START-FULL-STATUS
  [ OPT-64-COLS? ] [IF]
  "  ~DEL ~INS ~JOIN ~SPLIT ~^~YANK ~^~INS-PAD ~FIND  ~NEXT ~PREV"
  [ELSE] [ OPT-42-COLS? ] [IF]
  "  ~DEL ~INS ~JOIN ~SPLIT ~FIND  ~NEXT ~PREV"
  [ELSE]
  "  ~DEL ~INS ~JOIN ~SPLIT  ~NEXT ~PREV"
  [ENDIF] [ENDIF]
  IC-TYPE
  CLEAR-REST-STATUS MARK-FULL-STATUS-UPDATE ;

|: INIT-FIND-STATUS  START-FULL-STATUS CLEAR-REST-STATUS ;

|: .CH-FIND-STATUS  ( ch )
  VIS-LINES FIND-BUF C@ AT 1 INVERSE EMIT 0 INVERSE ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; simple cursor movement

|: LINE-UP
 [ OPT-FS4-64? ] [IF]
  CUR-LINE 1- 0MAX TO CUR-LINE
 [ELSE]
  CUR-LINE 1- DUP 0>= IF TO CUR-LINE EXIT ENDIF DROP
  TOP-LINE NOT?EXIT
  TOP-LINE 1- TO CUR-LINE  0 TO TOP-LINE
  RF-FULL
 [ENDIF] ;

|: LINE-DOWN
 [ OPT-FS4-64? ] [IF]
  CUR-LINE 1+ LAST-SCR-LINE MIN TO CUR-LINE
 [ELSE]
  CUR-LINE 1+ DUP VIS-LINES < IF TO CUR-LINE EXIT ENDIF
  TOP-LINE + SCR-LINES U< NOT?EXIT
  TOP-LINE VIS-LINES 2/ +  SCR-LINES VIS-LINES - MIN
  DUP TOP-LINE - CUR-LINE SWAP - 1+ TO CUR-LINE
  TO TOP-LINE
  RF-FULL
 [ENDIF] ;

|: LEFT/RIGHT  ( dir )  CUR-COL + 0MAX LAST-COL MIN TO CUR-COL ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; set line and column, used in "WHERE"

|: SET-LINE  ( lidx )
  0MAX LAST-SCR-LINE MIN TO CUR-LINE
 [ OPT-FS4-64? ] [IFNOT]
  0 TO TOP-LINE
  BEGIN CUR-LINE VIS-LINES < 0WHILE
  1 +TO TOP-LINE -1 +TO CUR-LINE REPEAT
 [ENDIF] ;

|: SET-POS  ( pos )  TO CUR-COL 0 LEFT/RIGHT ;


|: CHAR-LEFT   -1 LEFT/RIGHT ;
|: CHAR-RIGHT   1 LEFT/RIGHT ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; insert and delete one char

|: INS-CH  ( ch )
  DUP BL < IF DROP EXIT ENDIF
  LAST-COL COL-C@ BL - IF ERR-BEEP DROP EXIT ENDIF
  CC-ADDR DUP 1+ LAST-COL CUR-COL - CMOVE>
  CUR-COL COL-C!
  CHAR-RIGHT UPDATE
  RF-CLINE ;

|: DEL-CH
  CC-ADDR DUP 1+ SWAP  LAST-COL CUR-COL - CMOVE
  BL LAST-COL COL-C!
  RF-CLINE ;

|: BS-CHAR  CUR-COL IF CHAR-LEFT DEL-CH ELSE ERR-BEEP ENDIF ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; fill current line with blanks

|: WIPE-LINE
  CUR-LADDR COLS BLANK UPDATE
  RF-CLINE ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; move to prev/next word.
;; the idea is like this:
;;   back: always move to word start.
;;   forward: after word end if inside a word, otherwise to word start.


|: WCHECK-BL?     ( char -- flag )  BL = ;
|: WCHECK-NOT-BL? ( char -- flag )  BL <> ;

;; move while the next char is passed the check.
|: WORD-SKIP  ( dir checker-cfa )
  >R BEGIN  ( dir | checker-cfa )
    DUP CUR-COL + COL-C@ ?DUP WHILE ;; always stop at \x00
    R@ EXECUTE WHILE
  DUP +TO CUR-COL REPEAT
  DROP RDROP ;

|: WSKIP  ( dir n/p-char-addr )
  COL-C@ BL = IF ['] WCHECK-BL? ELSE ['] WCHECK-NOT-BL? ENDIF
  WORD-SKIP ;

|: WORD-LEFT
  CUR-COL NOT?EXIT
  -1 CUR-COL 1- WSKIP ;

|: WORD-RIGHT
  CUR-COL LAST-COL - NOT?EXIT
  1 CUR-COL WSKIP CHAR-RIGHT ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; insert and delete the whole line.
;; assume that we have at least 2 usable buffers.

|: DEL-LINE
  CL# LAST-SCR-LINE
  2DUP - IF RF-FROM-CLINE
    SWAP DO
      I 1+ LADDR I LADDR COLS CMOVE UPDATE
    LOOP
  ELSE 2DROP RF-CLINE ENDIF
  LAST-SCR-LINE LADDR COLS BLANK UPDATE ;


|: CAN-INS-LINE?  ( -- flag )
  CL# LAST-SCR-LINE - 0IF FALSE EXIT ENDIF
  LAST-SCR-LINE EMPTY-LINE? ;

|: INS-LINE
  CAN-INS-LINE? 0IF ERR-BEEP EXIT ENDIF
  RF-FROM-CLINE ;; anyway
  CL# LAST-SCR-LINE DO I 1- LADDR I LADDR COLS CMOVE UPDATE -1 +LOOP
  WIPE-LINE ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; split line at the current cursor position

;; after pad
|: TMP-BUF  ( -- addr )  FIND-BUF 69 + ;

|: SPLIT-LINE
  CUR-COL 0IF INS-LINE LINE-DOWN 0 TO CUR-COL EXIT ENDIF
  CAN-INS-LINE? 0IF ERR-BEEP EXIT ENDIF
  ;; save current line to temp buffer, we'll need it
  CL# LADDR TMP-BUF CUR-COL CMOVE
  INS-LINE
  ;; restore partial line
  TMP-BUF CUR-LADDR CUR-COL CMOVE UPDATE
  ;; rewrite next line
  LINE-DOWN WIPE-LINE
  TMP-BUF CUR-COL + CUR-LADDR #REST-CH CMOVE UPDATE
  0 TO CUR-COL ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; join lines

;; this will collect chars from the next line.
;; if the next line is empty, delete next line.
;; assume that we have at least 2 buffers.
|: JOIN-LINE
  CL# LAST-SCR-LINE - 0IF ERR-BEEP EXIT ENDIF
  LAST-COL COL-C@ BL - IF ERR-BEEP EXIT ENDIF
  CL# LINE-LEN CUR-COL MAX TO CUR-COL
  ;; copy chars from the beginning of the next line
  CL# 1+ LADDR  ;; from
  CC-ADDR       ;; to
  #REST-CH CMOVE UPDATE ;; copy
  ;; now delete first chars at the next line
  CL# 1+ LADDR    ;; to
  DUP #REST-CH +  ;; from
  SWAP COLS #REST-CH - CMOVE
  ;; blank rest of the line
  CL# 1+ LADDR COLS + #REST-CH -  ;; from
  #REST-CH BLANK UPDATE
  ;; check if the next line is empty
  CL# 1+ LINE-LEN IF  ;; not empty, fast refresh
    [ OPT-FS4-64? ] [IF]
      CUR-LINE 0 AT  CUR-LADDR XLINE CR
      CL# 1+ LADDR XLINE
    [ELSE]
      \ FIXME: optimise this!
      RF-FROM-CLINE
    [ENDIF]
  ELSE  ;; empty, delete it
    RF-CLINE
    LINE-DOWN DEL-LINE LINE-UP
  ENDIF ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; yank to PAD, and insert yanked line

;; to PAD
|: YANK
  CUR-LADDR PAD COLS CMOVE
  PAD COLS + OFF
  DEL-LINE ;

|: INS-PAD
  CAN-INS-LINE? 0IF ERR-BEEP EXIT ENDIF
  INS-LINE
  PAD CUR-LADDR COLS CMOVE UPDATE ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; move to prev/next screen

|: SHIFT-TO-SCREEN  ( dir )
  SCR @ +
  DUP $8000 AND IF DROP ERR-BEEP EXIT ENDIF
  SCR !
  DRAW-FULL-STATUS
  RF-FULL ;

|: PREV-SCREEN  -1 SHIFT-TO-SCREEN ;
|: NEXT-SCREEN   1 SHIFT-TO-SCREEN ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; main key handlers

|: KK-CR    0 TO CUR-COL LINE-DOWN ;

|: KK-SS+CR SPLIT-LINE ;

|: GO-HOME  0 TO CUR-COL ;
|: GO-END   CL# LINE-LEN LAST-COL MIN TO CUR-COL ;

|: GO-PG-TOP    0 TO CUR-LINE ;
|: GO-PG-BOTTOM LAST-VIS-LINE TO CUR-LINE ;

OPT-BLOCK-I/O-TYPE R/W-TAPE = [IF]
|: ZXE-SAVE  FLUSH ZXEMUT-SAVE-RAMDISC ;
|: ZXE-LOAD  EMPTY-BUFFERS ZXEMUT-LOAD-RAMDISC RF-FULL ;
[ENDIF]

|: GO-LEFT
  CUR-COL IF CHAR-LEFT EXIT ENDIF
  CL# NOT?EXIT
  LINE-UP CL# LINE-LEN LAST-COL MIN TO CUR-COL ;

|: GO-RIGHT
  CUR-COL LAST-COL - IF CHAR-RIGHT EXIT ENDIF
  CL# LAST-SCR-LINE - NOT?EXIT
  0 TO CUR-COL LINE-DOWN ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; incremental search

|: FBUF+LEN!  ( n )
  FIND-BUF C@ + FIND-BUF C! ;

|: INC-BS
  FIND-BUF C@ 0IF ERR-BEEP EXIT ENDIF
  -1 FBUF+LEN! BL .CH-FIND-STATUS
  RESTORE-CUR-POS ;


|: INC-CP-EQU  ( -- flag )
  FIND-BUF C@ DUP NOT?EXIT
  #REST-CH > IF FALSE EXIT ENDIF
  CC-ADDR FIND-BUF COUNT MEMEQU ;

|: INC-LINE-NO-MORE?  ( -- flag )
  #REST-CH  FIND-BUF C@  < ;

;; from CUR-COL
|: (INC-LFIND-NEXT)  ( -- pos TRUE // FALSE )
  #REST-CH FIND-BUF C@ - DUP 0< IF 0AND EXIT ENDIF 1+
  CC-ADDR SWAP  FIND-BUF 1+ C@ MEMCHR 0IF FALSE EXIT ENDIF
  DROP CUR-LADDR - TRUE ;


|: INC-RIGHT  ( -- ok? )
  INC-LINE-NO-MORE? 0IF 1 +TO CUR-COL TRUE EXIT ENDIF
  CL# LAST-SCR-LINE - 0IF FALSE EXIT ENDIF
  0 TO CUR-COL LINE-DOWN TRUE ;

|: INC-LEFT  ( -- ok? )
  CUR-COL IF -1 +TO CUR-COL TRUE EXIT ENDIF
  CL# DUP NOT?EXIT DROP LINE-UP
  COLS FIND-BUF C@ - TO CUR-COL TRUE ;


|: (INC-FIND-NEXT)  ( -- found-flag )
  BEGIN INC-RIGHT WHILE
    (INC-LFIND-NEXT) IF
      TO CUR-COL INC-CP-EQU -DUP ?EXIT
    ELSE COLS TO CUR-COL ENDIF
  REPEAT RF-NONE FALSE ;

|: INC-FIND-PREV
  SAVE-CUR-POS
  BEGIN INC-LEFT WHILE INC-CP-EQU ?EXIT REPEAT
  RESTORE-CUR-POS RF-NONE ERR-BEEP ;

|: INC-FIND-NEXT  ( -- ok-flag )
  FIND-BUF C@ 0IF ERR-BEEP EXIT ENDIF
  SAVE-CUR-POS (INC-FIND-NEXT)
  0IF ERR-BEEP RESTORE-CUR-POS ENDIF ;

|: INC-ADD-CH  ( ch )
  FIND-BUF C@ LAST-COL - 0IF ERR-BEEP EXIT ENDIF
  SAVE-CUR-POS
  DUP .CH-FIND-STATUS
  FIND-BUF COUNT + C!  1 FBUF+LEN!
  INC-CP-EQU ?EXIT
  (INC-FIND-NEXT) 0IF ERR-BEEP INC-BS ENDIF ;


\ <zx-hidden>
CREATE INC-KEYS
  12 c, ['] INC-BS ,        ;; CS+0
   8 c, ['] INC-FIND-PREV , ;; CS+5
  11 c, ['] INC-FIND-PREV , ;; CS+7
   9 c, ['] INC-FIND-NEXT , ;; CS+8
  10 c, ['] INC-FIND-NEXT , ;; CS+6
   7 c, ['] SET-DONE ,      ;; CS+1
0 c, CREATE;
\ <zx-normal>

|: INC-FIND-LOOP
  SAVE-CUR-POS
  BEGIN ?TERMINAL ?EXIT
    WAIT-KEY DUP INC-KEYS FIND-KEY IF NIP EXECUTE
    ELSE DUP BL >= IF INC-ADD-CH ELSE DROP ENDIF ENDIF
    REFRESH
  DONE? UNTIL
  FALSE TO DONE? MARK-FULL-STATUS-UPDATE ;

|: INC-FIND-START
  FIND-BUF OFF  ;; cstr
  INIT-FIND-STATUS
  INC-FIND-LOOP ;

|: INC-FIND-CONT
  INIT-FIND-STATUS
  FIND-BUF C@ 1- COLS U> IF 0 FIND-BUF C! ENDIF
  VIS-LINES 0 AT 1 INVERSE FIND-BUF C@++ TYPE 0 INVERSE
  INC-FIND-LOOP ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; working with whole screens

|: WIPE-BUFS
  "  ~BLANK ~ERASE  ~WIPE" SMSG
  MARK-FULL-STATUS-UPDATE
  WAIT-KEY UPCHAR CASE
    [CHAR] W OF EMPTY-BUFFERS RF-FULL " WEPT" SMSG ENDOF
    [CHAR] B OF SCR-BLANK ENDOF
    [CHAR] E OF SCR-ERASE ENDOF
  ENDCASE ;

|: FLUSH-BUFS
  FLUSH " FLUSHED" SMSG ;

|: INSERT-SCR
  SCR @ 1+ SCR-FIND-EMPTY 0IF ERR-BEEP EXIT ENDIF
  BEGIN SCR @ OVER < WHILE DUP 1- OVER SCR-COPY 1- REPEAT DROP
  SCR-BLANK
  RF-FULL ;

|: REMOVE-SCR
  SCR @ 1+ SCR-FIND-EMPTY 0IF ERR-BEEP EXIT ENDIF
  DUP >R  ;; we'll wipe it
  SCR @ DO I 1+ I SCR-COPY LOOP
  R> 0 SCR-FILL
  RF-FULL ;

|: TO-LAST-SCR
  SCR @ SCR-FIND-EMPTY NOT?EXIT 1- 0MAX SCR !
  MARK-FULL-STATUS-UPDATE RF-FULL ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; block commands

CREATE BC-KEYS
  [CHAR] F c, ['] FLUSH-BUFS ,
  [CHAR] W c, ['] WIPE-BUFS ,
  [CHAR] I c, ['] INSERT-SCR ,
  [CHAR] D c, ['] REMOVE-SCR ,
  [CHAR] L c, ['] TO-LAST-SCR ,
  7 c, ['] noop ,
0 c, CREATE;


|: BLOCK-CMDS
  "  ~FLUSH ~INSERT ~DELETE ~WIPE(RST)" SMSG
  MARK-FULL-STATUS-UPDATE
  WAIT-KEY UPCHAR BC-KEYS FIND-KEY
  IF EXECUTE ELSE ERR-BEEP ENDIF ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; lists of key handlers

\ <zx-hidden>
;;  DOWN: 10
;;    UP: 11
;;  LEFT: 8
;; RIGHT: 9
;;  CS+1: 7
;;  CS+3: 4
;;  CS+4: 5
;;  CS+9: 15
;;  CS+0: 12

;; extended mode keys (after CS+SS)
CREATE X-KEYS
  [CHAR] H c, ['] GO-HOME ,
  [CHAR] E c, ['] GO-END ,
  [CHAR] > c, ['] GO-PG-TOP ,     ;; SS+T
  [CHAR] * c, ['] GO-PG-BOTTOM ,  ;; SS+B
  [CHAR] D c, ['] DEL-LINE ,
  [CHAR] I c, ['] INS-LINE ,
  [CHAR] J c, ['] JOIN-LINE ,
  [CHAR] S c, ['] SPLIT-LINE ,
  [CHAR] P c, ['] PREV-SCREEN ,
  [CHAR] N c, ['] NEXT-SCREEN ,
  [CHAR] [ c, ['] YANK ,      ;; SS+Y
       127 c, ['] INS-PAD ,   ;; SS+I
OPT-BLOCK-I/O-TYPE R/W-TAPE = [IF]
  [CHAR] | c, ['] ZXE-SAVE ,  ;; SS+S
  [CHAR] = c, ['] ZXE-LOAD ,  ;; SS+L
[ENDIF]

  [CHAR] F c, ['] INC-FIND-START ,
  [CHAR] { c, ['] INC-FIND-CONT ,

  [CHAR] B c, ['] BLOCK-CMDS ,

         8 c, ['] GO-HOME ,
         9 c, ['] GO-END ,
        11 c, ['] GO-PG-TOP ,
        10 c, ['] GO-PG-BOTTOM ,

  7 c, ['] noop ,
0 c, CREATE;

;; main extended mode keypress handler
|: DO-EXT-MODE  ( ch )
  DRAW-EXT-STATUS WAIT-KEY
  UPCHAR X-KEYS FIND-KEY
  IF EXECUTE ELSE ERR-BEEP ENDIF ;


;; normal editing keys
CREATE E-KEYS
   8 c, ['] GO-LEFT ,     ;; 8 -- CS+5
   9 c, ['] GO-RIGHT ,    ;; 9 -- CS+8
  10 c, ['] LINE-DOWN ,   ;; 10 -- CS+6
  11 c, ['] LINE-UP ,     ;; 11 -- CS+7
  12 c, ['] BS-CHAR ,     ;; 12 -- CS+0
  14 c, ['] DO-EXT-MODE , ;; 14 -- SS+CS

   2 c, ['] KK-SS+CR ,    ;; 2 -- SS+CR
  13 c, ['] KK-CR ,       ;; 13 -- CR

   4 c, ['] WORD-LEFT ,   ;; 4 -- CS+3
  15 c, ['] WORD-RIGHT ,  ;; 15 -- CS+9

   7 c, ['] SET-DONE ,    ;; 7 -- CS+1
0 c, CREATE;
\ <zx-normal>

;; main editor loop
|: ED-LOOP
  0 SYS: LAST-K C!
  BEGIN ?TERMINAL ?EXIT
    WAIT-KEY
    DUP E-KEYS FIND-KEY IF NIP EXECUTE ELSE INS-CH ENDIF
    UPDATE-STATUS REFRESH
  DONE? UNTIL CR ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; editor entry point

<zx-normal>
: RUN
  NORM-ATTRS CLS
  FALSE TO DONE?
  DRAW-FULL-STATUS
  RF-FULL REFRESH
  ED-LOOP FLUSH ;

zx-here swap - endcr ." FS editor size: " .bytes cr

<zx-forth>
<zx-normal>

: ED  EDITOR: RUN ;
: EDIT  ( scr# )  SCR ! ED ;

: WHERE
  ERR-BLK B/BUF U* ERR-POS 0 D+ #SCR U/MOD SCR !
  EDITOR: COLS /MOD
  EDITOR: SET-LINE EDITOR: SET-POS
  ED ;


<zx-done>
