;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Z80 Abersoft fig-FORTH recompiled
;; Copyright (C) 2024 Ketmar Dark // Invisible Vector
;; Understanding is not required. Only obedience.
;; GPLv3 ONLY
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; input ("stick") scanner
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


<zx-definitions>
zx-here


;; detect kempston joystick
;; interrupts must be enabled
code: DETECT-KJOY ( -- bool )
  call  # detect-kjoy
  next-push-hl
;; AF, E: dead; HL: result
detect-kjoy:
  ;; high byte of the port number
  xor   a
  ;; 2 bytes, 8ts vs 3 bytes, 10ts
  ld    h, a
  ld    l, a
  ;; wait for the interrupt to avoid floating bus issue
  halt
  in    a, () $1F
  ;; $FF? nothing's here
  cp    # $FF
  ret   z
  ;; test for some invalid values
  ld    e, a
  ;; bits 5-7 should be reset
  and   # $E0
  ret   nz
  ld    a, e
  ;; check if left & right pressed simultaneously
  and   # $03
  cp    # $03
  ret   z
  ld    a, e
  ;; check if up & down pressed simultaneously
  and   # $0C
  cp    # $0C
  ret   z
  inc   l           ;; change the flag
  ret
;code-no-next


$01 constant SMASK-LEFT
$02 constant SMASK-RIGHT
$04 constant SMASK-UP
$08 constant SMASK-DOWN
$10 constant SMASK-FIRE
$20 constant SMASK-FIRE2

6 constant #INP-TBL

;; various predefined keys
;; fire,down,up,right,left
;; scancode format:
;;   bits 0-2: bit index (0 is LSB), counting from 1
;;   bits 3-6: index in `input-port-table`
create INP-TBL-STICK  ;; working table
  7 8 *  0 1+  + c,   ;; space
  7 8 *  2 1+  + c,   ;; "m"
  1 8 *  0 1+  + c,   ;; "a"
  2 8 *  0 1+  + c,   ;; "q"
  5 8 *  0 1+  + c,   ;; "p"
  5 8 *  1 1+  + c,   ;; "o"
create;

create INP-TBL-OPQAM
  7 8 *  0 1+  + c,   ;; space
  7 8 *  2 1+  + c,   ;; "m"
  1 8 *  0 1+  + c,   ;; "a"
  2 8 *  0 1+  + c,   ;; "q"
  5 8 *  0 1+  + c,   ;; "p"
  5 8 *  1 1+  + c,   ;; "o"
create;

create INP-TBL-KJOY
  7 8 *  0 1+  + c,   ;; space
  8 8 *  4 1+  + c,   ;; fire
  8 8 *  2 1+  + c,   ;; down
  8 8 *  3 1+  + c,   ;; up
  8 8 *  0 1+  + c,   ;; right
  8 8 *  1 1+  + c,   ;; left
create;

create INP-TBL-CURSOR
  7 8 *  0 1+  + c,   ;; space
  4 8 *  0 1+  + c,   ;; "0"
  4 8 *  4 1+  + c,   ;; "6"
  4 8 *  3 1+  + c,   ;; "7"
  4 8 *  2 1+  + c,   ;; "8"
  3 8 *  4 1+  + c,   ;; "5"
create;

create INP-TBL-SCL1
  7 8 *  0 1+  + c,   ;; space
  4 8 *  0 1+  + c,   ;; "0"
  4 8 *  2 1+  + c,   ;; "8"
  4 8 *  1 1+  + c,   ;; "9"
  4 8 *  3 1+  + c,   ;; "7"
  4 8 *  4 1+  + c,   ;; "6"
create;


: STICK-KJOY!  inp-tbl-kjoy inp-tbl-stick #inp-tbl cmove ;


;; read input key, return "scancode" suitable for using in stick table
code: STICK-INKEY  ( -- scancode // 0 )
  call  # input-inkey-scancode
  ld    l, a
  jr    # zx-word-read-stick-xexit
;code-no-next

;; translate *VALID* scancode to ASCII. to use in "redefine keys".
;; note that 0 is valid scancode too, and returns 0.
code: SCAN>ASCII  ( scancode -- ascii // 0 )
  pop   hl
  ld    a, l
  call  # input-scancode-to-ascii
  ld    l, a
  jr    # zx-word-read-stick-xexit
;code-no-next


;; read stick input using input table (see `input_table_XXX` above).
;; stick state: XFDURL; F is $10.
;; "X" is "space pressed".
code: STICK-READ  ( -- stick-state )
  push  bc
  call  # read-stick
  ld    l, c
  pop   bc
zx-word-read-stick-xexit:
  ld    h, # 0
  next-push-hl

;; this is used to convert the result of `xte_input_inkey_scancode`
;; to ASCII via `xte_input_inkey_ascii`
;; 13: enter
;; 30: caps shift
;; 31: symbol shift
input-keynames:
        30 db, [CHAR] Z db, [CHAR] X db, [CHAR] C db, [CHAR] V db, 0 dw, 0 db,
  [CHAR] A db, [CHAR] S db, [CHAR] D db, [CHAR] F db, [CHAR] G db, 0 dw, 0 db,
  [CHAR] Q db, [CHAR] W db, [CHAR] E db, [CHAR] R db, [CHAR] T db, 0 dw, 0 db,
  [CHAR] 1 db, [CHAR] 2 db, [CHAR] 3 db, [CHAR] 4 db, [CHAR] 5 db, 0 dw, 0 db,
  [CHAR] 0 db, [CHAR] 9 db, [CHAR] 8 db, [CHAR] 7 db, [CHAR] 6 db, 0 dw, 0 db,
  [CHAR] P db, [CHAR] O db, [CHAR] I db, [CHAR] U db, [CHAR] Y db, 0 dw, 0 db,
        13 db, [CHAR] L db, [CHAR] K db, [CHAR] J db, [CHAR] H db, 0 dw, 0 db,
        32 db,       31 db, [CHAR] M db, [CHAR] N db, [CHAR] B db, 0 dw, 0 db,


;; table of all ports (keyboard and kempston)
;; low part of the port number is always "#fe" for kb, and "#00" for kempston
input-port-table:
  $FE db, ;; #1111_1110 -- Czxcv  (0)
  $FD db, ;; #1111_1101 -- asdfg  (1)
  $FB db, ;; #1111_1011 -- qwert  (2)
  $F7 db, ;; #1111_0111 -- 12345  (3)
  $EF db, ;; #1110_1111 -- 09876  (4)
  $DF db, ;; #1101_1111 -- poiuy  (5)
  $BF db, ;; #1011_1111 -- Elkjh  (6)
  $7F db, ;; #0111_1111 -- _Smnb  (7)
  $00 db, ;; special flag: kempston joystick -- right,left,down,up,fire


;; check for a pressed key, return scancode.
;; returned scancode can be used to built the table for `read-stick`.
;;
;; IN:
;;   none
;; OUT:
;;   A: key scancode, or 0
;;   F: dead
;;   CARRY: set if a key was pressed
;;
input-inkey-scancode:
  push  bc
  ld    bc, # $FE00     ;; B is keyboard port, C will contain scancode
.nextrow:
  ;; two-byte IN
  ld    a, b            ;; high byte of the keyboard port
  in    a, () $FE       ;; read keyboard
  cpl                   ;; invert (because for keyboard, "0" means "pressed")
  and   # $1F           ;; throw away non-keyboard bits
  jr    nz, # .gotkey
  ;; no key pressed, move to the next port
  ;; increment keyboard port bit index
  ld    a, c
  add   a, # 8
  ld    c, a
  ;; modify keyboard port
  sll   b               ;; sets bit 0 to 1, moves bit 7 to carry
  jr    c, # .nextrow
  ;; alas, nothing was pressed at all
  xor   a               ;; set result, reset carry flag
  pop   bc
  ret
.gotkey:
  ;; find the number of the first set bit
  ;; C contains high part of the scancode
  inc   c
  rra
  jr    nc, # .gotkey
  ;; done
  ld    a, c
  pop   bc
  ret


;; translate *VALID* scancode to ASCII
;;
;; IN:
;;   A: *VALID* scancode from `input-inkey-scancode`
;; OUT:
;;   A: ASCII code or 0 if A is 0 on enter
;;   F: dead
;;   CARRY: set if A is not 0
input-scancode-to-ascii:
  ;; just in case, check for zero scancode (it means "no key")
  or    a
  ret   z
  push  hl
  dec   a               ;; we are indexing from 1
  add   a, # input-keynames lo-byte
  ld    l, a
  [@@] input-keynames lo-byte 0= [IF]
    ;; it is page-aligned
    ld    h, # input-keynames hi-byte
  [ELSE]
    ;; not page-aligned
    [@@] input-keynames lo-byte 60 + 255 > [IF]
      ;; need to fix high byte
      ld    a, # input-keynames hi-byte
      adc   a, # 0
      ld    h, a
    [ELSE]
      ld    h, # input-keynames hi-byte
    [ENDIF]
  [ENDIF]
  ld    a, (hl)
  pop   hl
  scf                   ;; set carry flag
  ret


;; IN:
;;   HL: ptr to input table
;; OUT:
;;   AF,DE: destroyed
;;   HL: ptr to next input table (who cares), or the same
;;   B: zero or 6
;;   C: stick state (FDURL; F is $10)
read-stick:
  ld    hl, # zx-['pfa] INP-TBL-STICK
read-stick-hl:
  ld    bc, # $0600     ;; 6 "keys" per channel, and init C to 0
.stick-loop:
  ld    a, (hl)         ;; load "scancode"
  or    a               ;; key may be missing
  jr    z, # .key-missing
  ;; port reading
  ;; shift away bit index
  rra
  rra
  rra
  and   # $0F
  ;; calculate addres in `input-port-table`, put it to DE
  add   a, # input-port-table lo-byte
  ld    e, a
  [@@] input-port-table lo-byte 0= [IF]
    ;; it is page-aligned
    ld    d, # input-port-table hi-byte
  [ELSE]
    ;; not page-aligned
    [@@] input-port-table lo-byte 7 + 255 > [IF]
      ;; need to fix high byte
      ld    a, # input-port-table hi-byte
      adc   a, # 0
      ld    d, a
    [ELSE]
      ;; high byte is always the same
      ld    d, # input-port-table hi-byte
    [ENDIF]
  [ENDIF]
  ;; load high port byte
  ld    a, (de)
  or    a
  jp    nz, # .read-keyboard
  ;; read kempston
  in    a, () $1F       ;; high port byte is in A
  jp    # .read-complete
.read-keyboard:
  ;; read keyboard
  in    a, () $FE       ;; high port byte is in A
  cpl                   ;; invert (because for keyboard, "0" means "pressed")
.read-complete:
  ;; check bit
  ld    e, a            ;; save read value to E
  ld    a, (hl)         ;; read "scancode" again
  and   # $07           ;; extract bit number (it is never 0)
.shift-loop:
  rr    e               ;; shift bit 0 to carry
  dec   a               ;; this doesn't modify carry
  jp    nz, # .shift-loop
.key-missing:
  ;; carry now contains "pressed" bit
  ;; shift it into C
  rl    c
  inc   hl              ;; move to the next scancode
  ;; and repeat everything
  djnz  # .stick-loop
  ;; done
  ret
;code-no-next


zx-here swap - endcr ." stick library size: " .bytes cr
<zx-done>
