;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Z80 Abersoft fig-FORTH recompiled
;; Copyright (C) 2024 Ketmar Dark // Invisible Vector
;; Understanding is not required. Only obedience.
;; GPLv3 ONLY
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; 16x16 XOR sprites, no runtime shifting
;; test code, directly included from "zx-80-libs-30-spr16.f"
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


\ $include "zx-80-libs-16-sfx-turner.f"
$include "zx-80-libs-17-sfx-joffa.f"

<zx-definitions>

;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; sprites

zx-here
push-ctx voc-ctx: spr16-loader
" zx-80-libs-30-spr16-30-sprites.f" false (include)
pop-ctx
zx-here swap - endcr ." SPR16 sprites size: " .bytes cr


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; test tile printer

0 quan TILE-BASE

code: PR-TILE  ( x y idx )
  pop   hl      ;; tile index
  ld    e, l
  add   hl, hl  ;; *2
  add   hl, hl  ;; *4
  add   hl, hl  ;; *8
  ld    d, # 0
  add   hl, de  ;; *9

  ld    de, () zx-['pfa] TILE-BASE
  add   hl, de

  pop   de      ;; y
  ld    a, d
  or    a
  jr    nz, # .bad-coord
  ld    a, e
  cp    # 24
  jr    nc, # .bad-coord
  pop   de
  push  hl
  ld    l, a    ;; L=y
  ld    a, d
  or    a
  jr    nz, # .bad-coord
  ld    a, e
  cp    # 32
  jr    nc, # .bad-coord

  ;; calculate scr$ address
  ld    a, l
  and   # $18
  or    # $40
  ld    h, a
  ld    a, l
  rrca
  rrca
  rrca
  and   # $E0
  or    e
  ld    l, a
  ;; HL=scr$

  pop   de

  ;; blit
  ld a, (de)  inc de  ld (hl), a  inc h   ;; 0
  ld a, (de)  inc de  ld (hl), a  inc h   ;; 1
  ld a, (de)  inc de  ld (hl), a  inc h   ;; 2
  ld a, (de)  inc de  ld (hl), a  inc h   ;; 3
  ld a, (de)  inc de  ld (hl), a  inc h   ;; 4
  ld a, (de)  inc de  ld (hl), a  inc h   ;; 5
  ld a, (de)  inc de  ld (hl), a  inc h   ;; 6
  ld a, (de)  inc de  ld (hl), a          ;; 7

  ;; move to attrs
  ld    a, h
  or    # $87
  rra
  rra
  srl   a     ;; rra for #C000 screen
  ld    h, a

  ld    a, (de)
  ld    (hl), a
.exit:
  next

.bad-coord:
  pop   de
  jr    # .exit
;code-no-next


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; tiles

\ zx-here
$8000 $1000 - tloader-zx-addr:!
tloader-zx-addr TO TILE-BASE
push-ctx voc-ctx: tile8-loader
" zx-80-libs-30-spr16-32-tiles.f" false (include)
pop-ctx
\ zx-here
tloader-zx-addr tloader-zx-addr-start -
endcr ." SPR16 tiles size: " .bytes ."  (" tloader-#tiles . ." tiles).\n"


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; test code


: MARK-THIRDS
   8  0 DO 0 I CXY>ATTR 32 @070 FILL LOOP
  16  8 DO 0 I CXY>ATTR 32 @060 FILL LOOP
  24 16 DO 0 I CXY>ATTR 32 @050 FILL LOOP ;


: EMPTY-CHAR?  ( x y -- flag )
  CXY>SCR$ 0 SWAP 8 FOR  ( flag addr )
    TUCK C@ OR SWAP
  256 + LOOP DROP 0= ;

: FILL-PROP-MAP
  TPROP-MAP [ 32 24 * ] {#,} ERASE
  24 0 DO
    32 0 DO
      I J EMPTY-CHAR? 0IF
        $80 J 1 AND 0IF $40 OR ENDIF  J 32* I + TPROP-MAP + C!
        $40 I J CXY>ATTR XOR!C
      ENDIF
    LOOP
  LOOP ;


;; test several sprites
create spi
  ( x)  0 c, ( y)  17 c, ( phase) 0 c, ( dummy) 0 c,
  ( x)  4 c, ( y)  98 c, ( phase) 1 c, ( dummy) 0 c,
  ( x) 12 c, ( y)  42 c, ( phase) 2 c, ( dummy) 0 c,
  ( x) 16 c, ( y)  21 c, ( phase) 3 c, ( dummy) 0 c,

  ( x)  3 c, ( y)  19 c, ( phase) 0 c, ( dummy) 0 c,
  ( x)  7 c, ( y)  98 c, ( phase) 1 c, ( dummy) 0 c,
  ( x) 15 c, ( y)  42 c, ( phase) 2 c, ( dummy) 0 c,
  ( x) 19 c, ( y)  32 c, ( phase) 3 c, ( dummy) 0 c,

  ( x)  0 c, ( y)  41 c, ( phase) 0 c, ( dummy) 0 c,
  ( x)  4 c, ( y) 118 c, ( phase) 1 c, ( dummy) 0 c,
  ( x) 12 c, ( y)  82 c, ( phase) 2 c, ( dummy) 0 c,
  ( x) 16 c, ( y)  92 c, ( phase) 3 c, ( dummy) 0 c,

  ( x)  3 c, ( y)  51 c, ( phase) 0 c, ( dummy) 0 c,
  ( x)  7 c, ( y)  48 c, ( phase) 1 c, ( dummy) 0 c,
  ( x) 15 c, ( y)  92 c, ( phase) 2 c, ( dummy) 0 c,
  ( x) 19 c, ( y)  62 c, ( phase) 3 c, ( dummy) 0 c,

  ( x) 23 c, ( y) 162 c, ( phase) 2 c, ( dummy) 0 c,
create;

: xz-move  ( spi^ )
  >R  ( | spi^ )
  ;; phase
  R@ 2+ C@ 1+ 3 AND DUP R@ 2+ C!
  0IF ;; x
    R@ C@ 1+ DUP 29 > IF DROP 0 ENDIF R@ C! ;; x
  ENDIF
  R@ 1+ C@ 1+ DUP 191 > IF DROP 15 ENDIF R@ 1+ C! ;; y
  RDROP ;

: xz-update  ( spi^ sidx )
  >R C@++ SWAP C@++ SWAP C@ 4* ( x y phase | sidx )
  R@ 6 > IF SPR-SOLDIER-WALK ELSE SPR-SKULL ENDIF +
  R> SPR-UPDATE ;

: xz
  1 BORDER
  SPR-INIT
  CLS CR CR CR
  MARK-THIRDS
  BEGIN
    spi 0 xz-update
    spi 4 + 1 xz-update
    HALT HALT
    HALT HALT
    SPR-IM2-HANDLER
    spi xz-move
    spi 4 + xz-move
  ?TERMINAL UNTIL ;

;; we can update about 5 sprites per frame
17 constant #XI
true quan xi-moving
true quan xi-player?

1 quan pl-x
3 quan pl-p
75 quan pl-y
true quan pl-left?

: .pl
  EXIT
  0 0 AT ." L: PL-P=" PL-P . ." X=" PL-X . ;

: pl-update
  pl-x pl-y
  \ pl-left? IF LREX ELSE RREX ENDIF
  pl-p pl-left? 4* +  4* SPR-REX-WALK +
  #XI SPR-UPDATE ;

: pl-turn
  pl-left? 1 XOR TO pl-left?
  \ 3 pl-p - TO pl-p
;

: pl-left
  pl-left? 0IF pl-turn EXIT ENDIF
  pl-p 1- 3 AND DUP TO pl-p 3 - ?EXIT
  pl-x 1- ( 0MAX) TO pl-x ;

: pl-right
  pl-left? IF pl-turn EXIT ENDIF
  pl-p 1+ 3 AND DUP TO pl-p ?EXIT
  pl-x 1+ ( 29 MIN) TO pl-x ;

: pl-up
  pl-y 1- 15 MAX TO pl-y ;

: pl-down
  pl-y 1+ 191 MIN TO pl-y ;

1 quan sfx-next
0 quan fire-down

: do-sfx
  fire-down ?EXIT
  JSFX-PLAYING? IF 5 254 OUTP EXIT ENDIF
  sfx-next DUP JSFX-START
  1+ 15 AND DUP 0= + sfx-next:!
  TRUE fire-down:! ;

: player-control
  STICK@
  DUP SMASK-LEFT AND IF pl-left .pl ( pl-update)
  ELSE DUP SMASK-RIGHT AND IF pl-right .pl ( pl-update) ENDIF ENDIF
  DUP SMASK-UP AND IF pl-up
  ELSE DUP SMASK-DOWN AND IF pl-down ENDIF ENDIF
  DUP SMASK-FIRE AND IF do-sfx ELSE 0 fire-down:! ENDIF
  DROP ;

;; with interrupts
: xi
  1 BORDER
  DETECT-KJOY IF STICK-KJOY! ENDIF
  SPR-SETUP
  \ CLS CR CR CR
  ." \cXS16 Engine demo.\n"
  ." \cpress SPACE to pause animation.\nuse kempston to move.\n"
  #XI . ." sprites, and the player."
  xi-player? 0IF CLS ENDIF
  MARK-THIRDS
  FILL-PROP-MAP
  SPR-ON
  0 LAST-K C!
  .pl
  \ pl-update
  BEGIN
    \ DI
    xi-player? IF player-control pl-update ENDIF
    #XI FOR I 4* SPI + I xz-update LOOP
    \ EI
    xi-player? 0IF
      HALT HALT
      HALT HALT
    ENDIF
    xi-moving IF
      SPI DUP #XI 4* + SWAP DO I xz-move 4 +LOOP
    ENDIF
    CS/SS? DUP 1 AND IF 0 TO xi-moving ENDIF
    2 AND IF 1 TO xi-moving ENDIF
    JSFX-PLAY
  ?TERMINAL UNTIL
  #XI 1+ FOR I SPR-ERASE LOOP
  HALT HALT HALT HALT
  \ SPR-OFF
  SPR-UNSETUP
;

\ ['] xi TO MAIN

1 [IF]
40 quan curr-spr-buf
0 quan curr-spr-buf-ofs

(*
16x16 sprite loader.
sprite format:
  db mask    ;; mask for the first byte ($FF -- put full first byte)
  32 bytes of data, upside down

sprite definition:
  db attr
  dw gfx-addr
  db reserved
*)

|: adv-spr-ofs
  curr-spr-buf-ofs 34 +
  b/buf over - 34 < ?< drop 0  1 +to curr-spr-buf >?
  to curr-spr-buf-ofs ;

|: write-one-sprite  ( gfx-addr )
  curr-spr-buf ( buffer) block curr-spr-buf-ofs +
  ( gfx-addr dest-addr )
  over c@ over c! ;; mask
  1+ 0 over c!    ;; reserved
  under+1 1+ 32 cmove
  update
  adv-spr-ofs ;

|: write-sprite-set  ( set-addr count )
  under+1 for dup @ write-one-sprite 4 + loop drop ;

: wrs
  SPR-REX-WALK 8 write-sprite-set
  SPR-REX-FALL 8 write-sprite-set
  SPR-SOLDIER-WALK 8 write-sprite-set
  SPR-SOLDIER-SHOOT 8 write-sprite-set
  SPR-SOLDIER-DYING 8 write-sprite-set
  SPR-FUN-ARROW 8 write-sprite-set
  SPR-SKULL 4 write-sprite-set
  flush
  endcr ." EBUF=" curr-spr-buf . ." EOFS=" curr-spr-buf-ofs . cr
;
[ENDIF]


<zx-done>
