;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Z80 Abersoft fig-FORTH recompiled
;; Copyright (C) 2024 Ketmar Dark // Invisible Vector
;; Understanding is not required. Only obedience.
;; GPLv3 ONLY
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; room (and tile) editor
;; fast 8x8 char printer
;; included directly from "zx-90-apps-20-room-edit.f"
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; draw text chars
zx-here

7 quan E8-ATTR
0 quan E8-XY
;; so we could print with pixel precision by Y.
;; note that attrs will not be set in this case.
0 quan E8-YOFS

;; window
$0000 quan E8WIN-XY0
$1820 quan E8WIN-XY1


code|: (E8)  ( char )
  ;; call screen address
  ld    hl, () zx-['pfa] E8-XY
  ld    a, h
  and   # $18
  or    # $40
  ld    d, a
  ld    a, h
  rrca
  rrca
  rrca
  and   # $E0
  or    l
  ld    e, a

  pop   hl        ;; char code
  push  bc        ;; save it here

  xor   a         ;; normal output
  bit   7, l
  jr    z, # .not-inv
  ld    a, # $2F  ;; CPL
.not-inv:
  ld    .inv-mode-instr (), a
  ;; normalize char
  ld    a, l
  and   # $7F
  cp    # 32
  jr    nc, # .char-ok
  cp    # 4
  jr    c, # .special-chars
  ld    a, # 63
.char-ok:

  ;; DE: scr$
  ;;  A: char code
  ld    l, a       ;; first part of ROM char address calculation
  cp    # 52
  jr    z, # .distortNot
  cp    # 96
  ld    a, # 7
  jr    c, # .distortOk
.distortNot:
  ld    a, # $ff
.distortOk:
  ld    e6-distort-and-oper (), a
  add   hl, hl    ;; *2
  ld    h, # 15
  add   hl, hl    ;; *4
  add   hl, hl    ;; *8
  ;; DE: scr$
  ;; HL: char-addr
.char-addr-hl-done:
  ex    de, hl
  ;; HL: scr$
  ;; DE: char-addr

  ld    b, # 8

  ld    a, () zx-['pfa] E8-YOFS
  and   # $07
  jr    nz, # .pixel-print

.line-loop:
  call  # .load-char-byte
  ld    (hl), a
  inc   h
  djnz  # .line-loop

.set-attrs:
  ;; set attr
  ld    a, h
  dec   a
  or    # $87
  rra
  rra
  srl   a     ;; rra for #C000 screen
  ld    h, a

  ld    a, () zx-['pfa] E8-ATTR
  cp    # $FF
  jr    z, # .skip-attr
  ld    (hl), a
.skip-attr:

  pop   bc
  next

.load-char-byte:
  ld    a, (de)
  inc   de
  ld    c, a
  and   # $7C
  rlca
  or    c
  ;; distort a little
  ld    c, a
  ld    a, b
  cp    # 3
  ld    a, c
  jr    nz, # .dontfuck
  and   # $07
$here 1- @def: e6-distort-and-oper
.dontfuck:
.inv-mode-instr:
  nop             ;; or CPL ($2F)
  ret

  ;; HL: scr$
  ;; DE: char-addr
.pixel-print:
  add   a, h
  ld    h, a
  jr    # .pixel-print-no-down
.pixel-print-loop:
  inc   h
  ld    a, h
  and   # $07
  jr    nz, # .pixel-print-no-down
  ld    a, l
  sub   # -32
  ld    l, a
  sbc   a, a
  and   # -8
  add   a, h
  ld    h, a
.pixel-print-no-down:
  call  # .load-char-byte
  ld    (hl), a
  djnz  # .pixel-print-loop
  jr    # .skip-attr

  ;; 0: up arrow
  ;; 1: down arrow
  ;; 2: left arrow
  ;; 3: right arrow
.special-chars:
  ;; DE: scr$
  ;;  A: char code
  add   a, a    ;; *2
  add   a, a    ;; *4
  add   a, a    ;; *8
  ld    hl, # .special-chars-data
  ld    c, a
  ld    b, # 0
  add   hl, bc
  ex    de, hl

  ld    a, () .inv-mode-instr
  ld    .inv-mode-instr-spec (), a
  ld    b, # 8
.line-loop-spec:
  ld    a, (de)
.inv-mode-instr-spec:
  nop             ;; or CPL ($2F)
  ld    (hl), a
  inc   de
  inc   h
  djnz  # .line-loop-spec
  jr    # .set-attrs

.special-chars-data:
;; up arrow
%00000000 db,
%00011000 db,
%00111100 db,
%01111110 db,
%01011010 db,
%00011000 db,
%00011000 db,
%00000000 db,
;; down arrow
%00000000 db,
%00011000 db,
%00011000 db,
%01011010 db,
%01111110 db,
%00111100 db,
%00011000 db,
%00000000 db,
;; left arrow
%00000000 db,
%00011000 db,
%00110000 db,
%01111110 db,
%01111110 db,
%00110000 db,
%00011000 db,
%00000000 db,
;; right arrow
%00000000 db,
%00011000 db,
%00001100 db,
%01111110 db,
%01111110 db,
%00001100 db,
%00011000 db,
%00000000 db,
;code-no-next


code|: (E8-ADV)
  ld    hl, () zx-['pfa] OUT
  inc   hl
  ld    zx-['pfa] OUT (), hl

  ld    hl, () zx-['pfa] E8-XY
  ld    de, () zx-['pfa] E8WIN-XY1
  ;; L=x
  ;; H=y
  inc   l
  ld    a, l
  cp    e
  jr    c, # .no-x-wrap
  ld    a, () zx-['pfa] E8WIN-XY0
  ld    l, a
  inc   h
  ld    a, h
  cp    d
  jr    c, # .no-y-wrap
  ld    a, () zx-['pfa] E8WIN-XY0 1+
  ld    h, a
.no-y-wrap:
.no-x-wrap:
  ld    zx-['pfa] E8-XY (), hl
;code


: XY/WH-SPLIT  ( xy/wh -- x/w y/h )  dup lo-byte swap hi-byte ;
: XY/WH-JOIN   ( xy/wh -- x/y-w/h )  256 * + ;


: WTYPE  ( addr count )  for c@++ (e8) (e8-adv) loop drop ;

: WIN-W  ( -- width )  E8WIN-XY1 lo-byte E8WIN-XY0 lo-byte - ;
: WIN-H  ( -- height)  E8WIN-XY1 hi-byte E8WIN-XY0 hi-byte - ;

: WIN-X0  ( -- x0 )  E8WIN-XY0 lo-byte ;
: WIN-Y0  ( -- y0 )  E8WIN-XY0 hi-byte ;
: WIN-XY0  ( -- x0 y0 )  win-x0 win-y0 ;

;; note that this as *after* the last col/row
: WIN-X1  ( -- x1 )  E8WIN-XY1 lo-byte ;
: WIN-Y1  ( -- y1 )  E8WIN-XY1 hi-byte ;
: WIN-XY1  ( -- x1 y1 )  win-x1 win-y1 ;


: WIN-AT  ( y x )
  win-w 1- min 0max win-x0 +
  swap win-h 1- min 0max win-y0 + 256* +
  to e8-xy ;

: WIN-AT@  ( -- y x )  e8-xy xy/wh-split swap ;


: WCR      E8WIN-XY1 lo-byte qaddr E8-XY c! (e8-adv) ;
: WENDCR   E8-XY lo-byte E8WIN-XY0 lo-byte - ?< wcr >? ;
: WLSTART  E8WIN-XY0 lo-byte qaddr E8-XY c! ;

: WBACK
  e8-xy lo-byte ?dup ?exit< 1- qaddr e8-xy c! >?
  e8-xy hi-byte dup ?< 1- e8win-xy1 lo-byte 1- || 0 >?
  swap 256* to e8-xy ;

: WDEMIT  ( char )
  dup $1F ~and ?exit< (e8) (e8-adv) >?
  << 4 of?v| wendcr |?
     7 of?v| wlstart |?
     8 of?v| wback |?
    13 of?v| wcr |?
  else| (e8) (e8-adv) >> ;

|: (WCLS)   (CLS) e8win-xy0 e8-xy:! ;
|: (.WCUR)  KCUR c@ 1 and not?exit [char] _ (e8) ;
|: (.WCUR0) KCUR c@ 1 and not?exit bl (e8) ;

: SETUP-E8-DRIVER
  ['] WDEMIT TO EMIT
  ['] (WCLS) TO CLS
  ['] (.WCUR) TO .CUR
  ['] (.WCUR0) TO .CUR0
  ['] WIN-AT TO AT
  ['] WIN-AT@ TO AT@ ;


: FS-WIN
  0 to e8win-xy0 $1820 to e8win-xy1
  0 to e8-xy 0 to e8-yofs ;


: ?ED-DEPTH0
  depth ?< fs-win @017 e8-attr:! 1 0 at ." INVALID DEPTH: " depth . cr quit >? ;


zx-here swap - endcr ." win-base library size: " .bytes cr
