;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Z80 Abersoft fig-FORTH recompiled
;; Copyright (C) 2024 Ketmar Dark // Invisible Vector
;; Understanding is not required. Only obedience.
;; GPLv3 ONLY
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; room (and tile) editor
;; simple windowing system
;; included directly from "zx-90-apps-20-room-edit.f"
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


zx-here

FREE-BUF-END quan wsave-buf-top

(*
: FS-WIN
  0 to e8win-xy0 $1820 to e8win-xy1
  0 to e8-xy 0 to e8-yofs ;
*)

;; set fullscreen window
: WIN-REINIT
  fs-win
  free-buf-end to wsave-buf-top ;


;; FIXME: too lazy to check args
: MK-WINDOW  ( x y w h )
  2over ( x y w h x y )
  rot + 256*  nrot + + to e8win-xy1
  256* + dup to e8win-xy0 to e8-xy
  0 to e8-yofs ;


;; fill current window with the current attr
: ATTR-FILL-WINDOW
  win-w >r  win-xy0 cxy>attr  win-h
  for dup j e8-attr fill 32 + loop
  drop rdrop ;


|: (FR-HLINE-LR)  ( scr$ width fill-byte lbyte rbyte )
  >r >r >r
  2dup r> fill
  ( left byte) over r> swap c!
  ( right byte) + 1- r> swap c! ;

;; clear current window, draw frame. doesn't set attr
: FRAME-WINDOW
  win-w >r  win-xy0 cxy>scr$
  dup r@ erase scr$v
  dup r@ $FF $7F $FE (fr-hline-lr) scr$v
  dup r@ $00 $70 $0E (fr-hline-lr) scr$v
  dup r@ $FF $7F $FE (fr-hline-lr) scr$v
  dup r@ $00 $58 $1A (fr-hline-lr) scr$v
  win-h 1- 8* 2- for dup j $00 $50 $0A (fr-hline-lr) scr$v loop
  dup r@ $00 $58 $1A (fr-hline-lr) scr$v
  dup r@ $FF $7F $FE (fr-hline-lr) scr$v
  dup r@ $00 $70 $0E (fr-hline-lr) scr$v
  dup r@ $FF $7F $FE (fr-hline-lr) scr$v
  r> erase ;

;; shrink current window, so output will not corrupt the frame.
;; also, reset otput position to (0, 0).
;; FIXME: too lazy to check args.
: SHRINK-WINDOW
  $0101 +to e8win-xy0
  -$0101 +to e8win-xy1
  e8win-xy0 to e8-xy ;

: UNSHRINK-WINDOW
  -$0101 +to e8win-xy0
  $0101 +to e8win-xy1 ;


;; save current window size, position, at-coords, attrs
: SAVE-WSP
  -7 +to wsave-buf-top
  e8win-xy0 wsave-buf-top !
  e8win-xy1 wsave-buf-top 2+ !
  e8-xy wsave-buf-top 4 + !
  e8-attr wsave-buf-top 6 + c! ;

;; restore window size, position, at-coords, attrs saved with "SAVE-WSP"
: RESTORE-WSP
  wsave-buf-top @ to e8win-xy0
  wsave-buf-top 2+ @ to e8win-xy1
  wsave-buf-top 4 + @ to e8-xy
  wsave-buf-top 6 + c@ to e8-attr
  7 +to wsave-buf-top ;


;; size of buffer required for saving scr$ under the window
|: SVW-BUF#  ( -- size )
  win-w win-h 2dup 8* *  nrot * + ;

;; save current window area to temporary buffer
: SAVE-WINDOW
  svw-buf# negate +to wsave-buf-top  ;; alloc room
  ;; save screen$ data
  win-xy0 cxy>scr$ wsave-buf-top  ( scr$ dest )
  win-w >r win-h 8* for ( scr$ dest | w )
    2dup r2@ cmove
    r2@ +  swap scr$v swap
  loop nip
  ;; save attr data
  win-xy0 cxy>attr swap
  win-h for ( attr dest | w )
    2dup r2@ cmove
    r2@ +  32 under+
  loop
  2drop rdrop ;

;; restore window saved with "SAVE-WINDOW"
: RESTORE-WINDOW
  wsave-buf-top win-xy0 cxy>scr$  ( src scr$ )
  win-w >r win-h 8* for ( src scr$ | w )
    2dup r2@ cmove
    r2@ under+  scr$v
  loop drop
  ;; restore attr data
  win-xy0 cxy>attr
  win-h for ( src attr | w )
    2dup r2@ cmove
    r2@ under+  32 +
  loop
  2drop rdrop
  ;; adjsust buffer pointer
  svw-buf# +to wsave-buf-top ;


;; use "CLOSE-WINDOW" to close it
: SHOW-INFO-WINDOW  ( addr count )
  save-wsp
  0max 30 min
  30 over - 0
  2over nip 2+ 3
\ endcr >r >r ." x=" swap . ." y=" . r> r> ." w=" swap . ." h=" . cr quit
  mk-window
  arrow-hide
  save-window
  @117 to e8-attr attr-fill-window frame-window shrink-window
  @116 to e8-attr wtype ;


;; close window opened by "OPEN-WINDOW" or by "SHOW-INFO-WINDOW"
: CLOSE-WINDOW
  unshrink-window
  arrow-hide
  restore-window
  restore-wsp ;

;; saves previous window size and scr$ occupied by the new window
: OPEN-WINDOW  ( x y w h attr )
  arrow-hide
  save-wsp
  to e8-attr mk-window
  save-window
  attr-fill-window frame-window shrink-window ;


;; doesn't save anything
: MAKE-WINDOW  ( x y w h attr )
  to e8-attr mk-window
  attr-fill-window frame-window shrink-window ;


: KM-WIN-XY@  ( -- x y )  km-xy@ win-y0 8* - swap win-x0 8* - swap ;
: KM-WIN-CXY@ ( -- x y )  km-cxy@ win-y0 - swap win-x0 - swap ;


zx-here swap - endcr ." win-high library size: " .bytes cr
