;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Z80 Abersoft fig-FORTH recompiled
;; Copyright (C) 2024 Ketmar Dark // Invisible Vector
;; Understanding is not required. Only obedience.
;; GPLv3 ONLY
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; room (and tile) editor
;; hot areas (to use with the arrow)
;; included directly from "zx-90-apps-20-room-edit.f"
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


zx-here
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; hot areas (maybe with text)

(*
  hot area format:
    db x, y   ;; in window; negative means "from the right / bottom"
    db count, text...   ;; text to print
      special case: negative count means "length in chars", and height
      in chars follows
    dw udata
  end with $80
*)


 0 quan HCUR-LIST   ;; current active list
-1 quan HCUR-XY     ;; current char abs coords (init to impossible)
 0 quan HCUR-WH     ;; current width and height (in chars)
 0 quan HCUR-UDATA  ;; current udata
 0 quan HCUR-ITEM^  ;; current item

;; frame areas without text?
true quan HOT-FRAME-AREAS


;; reset current item (but not list)
: HOT-RESET
  -1 to hcur-xy 0 to hcur-wh 0 to hcur-udata 0 to hcur-item^ ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; item accessors

|: HOT-X>ABS  ( x -- x )  dup $80 and ?< c>s win-x1 || win-x0 >? + ;
|: HOT-Y>ABS  ( y -- y )  dup $80 and ?< c>s win-y1 || win-y0 >? + ;


\ : HOT-END?  ( item^ -- end-flag )
\   dup 0?exit< drop true >? c@ $80 = ;
;; it is used on each iteration, let's make it slightly faster
code: HOT-END?  ( item^ -- end-flag )
  pop   hl
  ld    a, h
  or    l
  jr    z, # .done
  ld    a, (hl)
  ld    hl, # 0
  cp    # $80
  jr    nz, # .done
  inc   l
.done:
;code-push-hl

;; move to the next hot item
\ : HOT-NEXT  ( item^ -- item^ )
\   2+ c@++ dup $80 and ?< drop 3 || 2+ >? + ;
;; it is used on each iteration, let's make it slightly faster
code: HOT-NEXT  ( item^ -- item^ )
  pop   hl
  inc   hl
  inc   hl
  ld    a, (hl)
  inc   hl
  ld    d, # 0
  ld    e, a
  cp    # $80
  jr    c, # .not-a-bar
  ld    e, # 1
.not-a-bar:
  add   hl, de
  inc   hl
  inc   hl
;code-push-hl

: HOT-XY-ABS  ( item^ -- xabs yabs )
  c@++ hot-x>abs swap c@ hot-y>abs ;

: HOT-WH  ( item^ -- w h )
  2+ c@++ dup $80 and ?exit< c>s negate swap c@ >?
  nip 2+ 1 ;

: HOT-BAR?  ( item^ -- bar? )
  2+ c@ $7F u> ;

: HOT-TEXT?  ( item^ -- bar? )
  hot-bar? not ;

;; for bars, "random-addr 0"
: HOT-TEXT  ( item^ -- addr count )
  2+ c@++ dup $7F u> ?< 0and >? ;

: HOT-UDATA  ( item^ -- udata )
  2+ c@++ dup $7F u> ?< drop 1+ || + >? @ ;

;; coords are absolute char coords
: HOT-INSIDE?  ( cx cy item^ -- inside? )
  \ dup 0?exit< 2drop 0and >?
  dup >r hot-xy-abs ( cx cy ix iy | item^ )
  rot swap -  nrot -  ( dy dx | item^ )
  2dup or 0< ?exit< rdrop drop 0and >?
  r> hot-wh ( dy dx w h )
  nrot u< nrot u< and ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; utilities

;; print text of all items
: HOT-PRINT  ( hot-list^ )
  hot-reset << dup hot-end? not?^|
    dup >r hot-xy-abs under+1 xy/wh-join to e8-xy
    r@ hot-text wtype
    r> hot-next |?
  else| drop >> ;

;; find item by index
: HOT-FIND  ( idx hot-list^ -- item^ TRUE // FALSE )
  << dup hot-end? not?^|
       over 0?exit< nip true >?
     hot-next |?
  else| drop 0and >> ;

;; find item at the given absolute char coords
: HOT-FIND-AT  ( cx cy hot-list^ -- item^ TRUE // FALSE )
  << dup hot-end? not?^|
    >r 2dup r@ hot-inside? ?exit< 2drop r> true >?
    r> hot-next |?
  else| 2drop 0and >> ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; draw frames around items

|: (hbar-or-lr)  ( scr$ w )
  ( left) over 1- $01 swap or!c  ( right) + $80 swap or!c ;

|: (frame-hbar)  ( absx absy w h )
  dup 0?exit< 2drop 2drop >?
  swap >r >r cxy>scr$  ( scr$ | w h )
  ;; top frame
  dup scr$^ r1@ 2dup (hbar-or-lr) $ff fill
  ;; mid frame
  r> 8* for ( scr$ | w ) dup j (hbar-or-lr) scr$v loop
  ;; bottom frame
  r> 2dup (hbar-or-lr) $ff fill ;

: HOT-DRAW-FRAMES  ( hot-list^ )
  << dup hot-end? not?^|
      hot-frame-areas over hot-text? or ?<
        dup >r hot-xy-abs
        r@ hot-wh (frame-hbar)
        r> >?
     hot-next |?
  else| drop >> ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; check if the arrow is inside any item, and mark it

|: (inv-hbar)  ( absxy abswh )
  over 1+ not?exit< 2drop >?
  swap xy/wh-split cxy>scr$
  swap xy/wh-split
  8* for  ( scr$ w )
    2dup over + swap do $FF i xor!c loop
    \ 2dup over + swap do over $0100 and ?< $AA || $55 >? i xor!c loop
  swap scr$v swap loop 2drop ;

;; hide selection (if there is any).
: HCUR-HIDE
  hcur-xy 1+ not?exit
  hcur-xy hcur-wh arrow-visible? >r
  arrow-hide (inv-hbar) r> ?< arrow-show >?
  hot-reset ;

: HCUR-MARK-ITEM  ( item^ )
  dup hcur-item^ = ?exit< drop >? ;; just in case
  hcur-hide
  dup hot-wh dup 0?exit< 2drop drop >? xy/wh-join to hcur-wh
  dup hot-xy-abs xy/wh-join to hcur-xy
  dup hot-udata to hcur-udata
  to hcur-item^
  hcur-xy hcur-wh arrow-visible? >r
  arrow-hide (inv-hbar) r> ?< arrow-show >? ;

|: (HCUR-CHECK)
  km-cxy@
  hcur-item^ ?< 2dup hcur-item^ hot-inside? ?exit< 2drop >? >?
  hcur-list hot-find-at not?exit< hcur-hide >?
  hcur-mark-item ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; main loop

;; called after each "halt" by waiters and loop
['] noop vect ARROW-IDLE

;; mask for mouse buttons (on which buttons we should react)
1 quan hot-loop-bmask
;; acutally pressed buttons
0 quan hot-loop-butt

: WAIT-BUTT-RELEASE
  km-butt@ not?exit
  arrow-pressed!
  arrow-show-force << halt arrow-idle km-butt@ ?^|| else| >>
  arrow-default! arrow-show-force ;

: WAIT-BUTT-PRESS
  wait-butt-release 0
  arrow-default! arrow-show-force
  << drop halt arrow-idle km-butt@ dup 0?^||
  else| hot-loop-butt:! >> ;


: LASTK-OFF  0 last-k c! ;
: LASTK@  ( -- char )  last-k c@ ;

: HOT-LOOP-BT0?   ( -- flag )  hot-loop-butt 1 and ;
: HOT-LOOP-BT1?   ( -- flag )  hot-loop-butt 2u/ 1 and ;
: HOT-LOOP-BT0|1? ( -- flag )  hot-loop-butt 3 and 0<> ;

: HOT-BT0!  $01 hot-loop-bmask:! ;
: HOT-BT1!  $02 hot-loop-bmask:! ;
: HOT-BT01! $03 hot-loop-bmask:! ;

: HOT-MENU-LOOP
  arrow-show wait-butt-release
  lastk-off
  << halt arrow-idle
     km-butt@ hot-loop-butt:!
     (hcur-check)
     lastk@ 7 = ?v||
     hot-loop-butt hot-loop-bmask and 0?^||
  else| >>
  arrow-hide ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; setup new hotlist

: HOT-SETUP  ( hot-list^ )
  hot-reset
  dup to hcur-list
  dup not?exit< drop >?
  dup hot-print hot-draw-frames
  (hcur-check) ;


zx-here swap - endcr ." hot-menu library size: " .bytes cr
