;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Z80 Abersoft fig-FORTH recompiled
;; Copyright (C) 2024 Ketmar Dark // Invisible Vector
;; Understanding is not required. Only obedience.
;; GPLv3 ONLY
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; 16x16 XOR sprites, no runtime shifting
;; room (and tile) editor
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


<zx-definitions>
zx-here


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; editor
<zx-hidden>

;; 16 tile types are allowed. kept in the low nibble.

;; if set, attribute of this tile will override sprite attribute
$80 constant PMASK-A-FROZEN
;; if set, it is not allowed to print any sprites over this tile
$40 constant PMASK-S-FROZEN
;; if set, tile type is "frozen"
$20 constant PMASK-T-FROZEN
;; if set, tile attr is overriden
$10 constant PMASK-U-ATTR

;; current tile map (the one we are editing).
;; WARNING! should always come in this order!
$6000 constant PROP-MAP   ;; 768 bytes, 3 buffers
$6300 constant TILE-MAP   ;; 768 bytes, 3 buffers
$6600 constant ATTR-MAP   ;; 768 bytes, 3 buffers

;; tile definitions.
;; first, gfx for all 256 tiles (2048 bytes, 8 buffers).
;; second, attributes for each tile (256 bytes, 1 buffer).
;; third, property flags for each tile (256 bytes, 1 buffer).
;; WARNING! should always come in this order!
$6900 constant TILE-GFX
$7100 constant TILE-ATTR
$7200 constant TILE-PROP

;; this is used to save scr$ under windows, if necessary.
;; it goes from "free-buf-end" up.
;; 3,328 bytes.
$7300 constant FREE-BUF-START
$8000 constant FREE-BUF-END

;; buffer #60: 16 lines of 16 chars -- property names
60 constant PNAMES-BUF

;; buffer with the current room.
;; the layout is:
;;   768 bytes of prop map (3 buffers)
;;   768 bytes of tile map (3 buffers)
;;   768 bytes of attr map (3 buffers)
;; 9 buffers
16 1024 256 */ constant WK-ROOM-BUF     ;; buffer #64

;; tile bitmaps (8 buffers), tile attrs (1 buffer), tile props (1 buffer)
WK-ROOM-BUF 10 + constant WK-TILES-BUF  ;; buffer #74

;; room buffers (10 buffers per room, 1 reserved)
100 constant WK-ROOMS-STORE-BUF


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; utilities

<zx-normal>
$include "zx-90-apps-20-room-edit-10-emit8.f"
$include "zx-90-apps-20-room-edit-12-mk-win.f"
$include "zx-90-apps-20-room-edit-14-hot-areas.f"

<zx-normal>
: PNAME"  ( idx )  \ name
  dup 0 16 within not 23 ?error \ FIXME: bad error!
  16* >r 34 text
  pad count dup 1 17 within not 23 ?error \ FIXME: bad error!
  pnames-buf block
  ( saddr scount baddr | bofs )
  r> + dup 16 blank
  ( saddr scount baddr )
  swap cmove update
  flush ;


<zx-hidden>


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; draw tile

code: (DRAW-TILE)  ( x y attr idx )
  pop   hl      ;; tile index
  ld    e, l
  add   hl, hl    ;; *2
  add   hl, hl    ;; *4
  add   hl, hl    ;; *8
  \ ld    d, # 0
  \ add   hl, de    ;; *9

  ld    de, # TILE-GFX
  add   hl, de

  pop   de
  ld    a, e
  ex    af, afx   ;; attr

  pop   de        ;; y
  ld    a, d
  or    a
  jr    nz, # .bad-coord
  ld    a, e
  cp    # 24
  jr    nc, # .bad-coord
  pop   de        ;; x
  push  hl        ;; save tile address
  ld    l, a      ;; L=y
  ld    a, d
  or    a
  jr    nz, # .bad-coord
  ld    a, e
  cp    # 32
  jr    nc, # .bad-coord

  ;; calculate scr$ address
  ld    a, l
  and   # $18
  or    # $40
  ld    h, a
  ld    a, l
  rrca
  rrca
  rrca
  and   # $E0
  or    e
  ld    l, a
  ;; HL=scr$

  pop   de      ;; tile address

  ;; blit
  ld a, (de)  inc e  ld (hl), a  inc h   ;; 0
  ld a, (de)  inc e  ld (hl), a  inc h   ;; 1
  ld a, (de)  inc e  ld (hl), a  inc h   ;; 2
  ld a, (de)  inc e  ld (hl), a  inc h   ;; 3
  ld a, (de)  inc e  ld (hl), a  inc h   ;; 4
  ld a, (de)  inc e  ld (hl), a  inc h   ;; 5
  ld a, (de)  inc e  ld (hl), a  inc h   ;; 6
  ld a, (de)         ld (hl), a          ;; 7

  ;; move to attrs
  ld    a, h
  or    # $87
  rra
  rra
  srl   a     ;; rra for #C000 screen
  ld    h, a

  ex    af, afx   ;; attrs
  cp    # $FF
  jr    z, # .exit
  ld    (hl), a
.exit:
  next

.bad-coord:
  pop   de    ;; x
  jr    # .exit
;code-no-next


;; from the working tilemap. doesn't check coords.
: DRAW-TILE  ( x y )
  2dup  32* +  dup ATTR-MAP + c@  swap TILE-MAP + c@ (draw-tile) ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; load/save various data

: load-n-blocks  ( block dest-addr block-count )
  for 2dup swap block swap 256 cmove  256 + under+1 loop 2drop ;

: save-n-blocks  ( block src-addr block-count )
  for 2dup swap ( buffer) block 256 cmove update  256 + under+1 loop 2drop ;

: load-wk-room  wk-room-buf prop-map 9 load-n-blocks ;
: save-wk-room  wk-room-buf prop-map 9 save-n-blocks flush ;

: load-tile-gfx-attrs  wk-tiles-buf tile-gfx 10 load-n-blocks ;
: save-tile-gfx-attrs  wk-tiles-buf tile-gfx 10 save-n-blocks flush ;

: load-room-to-wk  ( room# )
  dup 1 128 within not?exit< drop 200 400 beep >?
  1- 10 * WK-ROOMS-STORE-BUF + prop-map 9 load-n-blocks ;

: save-wk-to-room  ( room# )
  dup 1 128 within not?exit< drop 200 400 beep >?
  1- 10 * WK-ROOMS-STORE-BUF + prop-map 9 save-n-blocks flush ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; draw working room

: draw-wk-tile-row  ( x y len )
  dup 0<= ?exit
  >r
  over 0< over 0< or ?exit< 2drop rdrop >?
  over 31 > over 23 > or ?exit< 2drop rdrop >?
  over 32 swap - r> min for 2dup draw-tile under+1 loop
  2drop ;

: draw-wk-room
  22 for 0 i 32 draw-wk-tile-row loop ;


|: fix-tile-attrs/props  ( mapofs )
  dup >r
  tile-map + c@ dup >r prop-map + c@ ( prop | mapofs tidx )
  ;; change attr?
  dup pmask-u-attr and 0?< r@ tile-attr + c@  r1@ attr-map + c! >?
  ;; change props?
  pmask-t-frozen and 0?< r@ tile-prop + c@  r1@ prop-map + c! >?
  rdrop rdrop ;

;; fix room for the given tile (attrs, props).
;; called from tile editor when tile attrs/props are changed.
: fix-wk-tile-attrs  ( tidx )
  lo-byte 768 for
    dup i tile-map + c@ - 0?< i fix-tile-attrs/props >?
  loop drop ;

: fix-wk-all-tile-attrs
  768 for i fix-tile-attrs/props loop ;


;; erase current window by drawing tiles over it
: erase-window
  win-y0 dup win-h + swap do
    win-x0 i win-w draw-wk-tile-row
  loop ;


: clear-wk-room
  prop-map 768 erase
  tile-map 768 erase
  attr-map 768 @007 fill ;


: load-all-wk-data
  " LOADING WORKING ROOM" show-info-window
  load-wk-room
  close-window
  " LOADING TILE GFX" show-info-window
  load-tile-gfx-attrs
  close-window ;

: init-editor
  1 border 0 paper 7 ink 0 flash 0 bright 0 inverse 0 gover cls
  win-reinit setup-e8-driver init-kmouse
  [ OPT-BLOCK-I/O-TYPE R/W-P3DOS = ] [IF]
    \ FREE-BUF-END +3DOS-SP0 !
    " PREPARING BLK FILE" show-info-window
    " ROOMEDIT.BLK" +3BLK-OPEN
    close-window
  [ENDIF]
  load-all-wk-data ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; bottom bar with tiles

0 quan bbar-first-tile
0 quan bbar-cur-tile-xy


|: print-def-tile  ( x y tidx )
  lo-byte dup tile-attr + c@ swap (draw-tile) ;

|: print-bb-tile  ( x y tofs )
  bbar-first-tile + print-def-tile ;

|: adv-bbar-first  ( delta )
  bbar-first-tile + qaddr bbar-first-tile c! ;


: bbar-up/down  ( -- down? )
  ?< 16 || -16 >? adv-bbar-first ;

: bbar-printer
  31 for i 1+ 22 i 2* print-bb-tile loop
  31 for i 1+ 23 i 2* 1+ print-bb-tile loop ;

: bbar-ctile  ( -- tidx )
  bbar-cur-tile-xy xy/wh-split
  swap 2* + lo-byte
  bbar-first-tile + lo-byte ;


: draw-bbar
  save-wsp arrow-hide
  fs-win @005 to e8-attr
  22 0 at $80 (e8) 23 0 at $81 (e8)
  bbar-printer
  bbar-cur-tile-xy xy/wh-split under+1 22 + cxy>attr $80 toggle
  restore-wsp ;


|: tbar-set-xy  ( tidx )
  bbar-first-tile -
  dup 2u/ swap 1 and xy/wh-join bbar-cur-tile-xy:! ;

: tbar-set-tidx  ( tidx )
  lo-byte
  dup bbar-first-tile dup 62 + within not?<
    dup $E0 and bbar-first-tile:!
    bbar-first-tile 15 > ?< 0 bbar-up/down >?
  >? tbar-set-xy draw-bbar ;


;;  0: not on tbar
;;  1: scrolled
;; -1: new tile selected
: km-tbar?  ( -- 0/1/-1 )
  km-cxy@ 22 - dup 0< ?exit< drop 0and >?
  over ?< 256* swap 1- + bbar-cur-tile-xy:! -1 >r || nip bbar-up/down 1 >r >?
  arrow-hide draw-bbar
  r> ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; editors

;; property editor
$include "zx-90-apps-20-room-edit-20-prop-editor.f"
;; tile editor
$include "zx-90-apps-20-room-edit-30-tile-editor.f"
;; room editor
$include "zx-90-apps-20-room-edit-40-room-editor.f"


<zx-normal>
: ee
  init-editor
  false to pe-names-loaded?
  room-editor
  arrow-hide
  fs-win ;


zx-here swap - endcr ." total room/tile editor size: " .bytes cr
<zx-done>
