;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Z80 Abersoft fig-FORTH recompiled
;; Copyright (C) 2024 Ketmar Dark // Invisible Vector
;; Understanding is not required. Only obedience.
;; GPLv3 ONLY
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; target compiler basics
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


extend-module TCOM

;; addresses of important ZX variables
0 quan zx-dp^
0 quan zx-fence^
0 quan zx-voclink^
0 quan zx-current^
0 quan zx-context^
0 quan zx-state^

0 quan zx-do-colon^
0 quan zx-do-constant^
0 quan zx-do-2constant^
0 quan zx-do-quan^
0 quan zx-do-vect^
0 quan zx-do-quan-addr^
0 quan zx-do-quan-to^
0 quan zx-do-quan-add-to^
0 quan zx-do-variable^
0 quan zx-do-2variable^
0 quan zx-do-uservar^
0 quan zx-do-does^
0 quan zx-do-vocab^
0 quan zx-do-alias^

OPT-ENABLE-FP? [IF]
0 quan zx-do-fp^
0 quan zx-do-fp-xmem^
[ENDIF]

;; user variables area
0 quan zx-uservar-start^
0 quan zx-uservar-end^
0 quan zx-uservar-curr^

0 quan zx-cold-uvars^
0 quan zx-cold-#uvars
0 quan zx-uvars-origin^

0 quan zx-dpx-init^
0 quan zx-dpx-#init


: .round-1000  ( kb kb-mod )
  100 /mod 10 mod 5 >= -
  dup 9 > ?< swap 1+ swap 10 - >? ;

: .bytes  ( n )
  dup ., ." bytes (~"
  1024 /mod .round-1000 swap 0.r, [char] . emit . ." kb)" ;


0 [IF]
: mem-tracer  ( byte addr -- byte addr )
  dup 10 < ?error" WUT?!" exit
  \ dup $765D $765F within not?exit
  dup $765D = not?exit
  over lo-byte $92 = ?exit
  endcr ." WRITING BYTE $" over .hex2 ."  to $" dup .hex4 cr
  abort ;
['] mem-tracer zxa:mem:trace-c!:!
[ENDIF]

: zx>real  ( zx-addr -- addr )  zxa:mem:ram^ ;

: zx-c@  ( zx-addr -- byte )  zxa:mem:c@ ;
: zx-c!  ( val zx-addr )  zxa:mem:c! ;

: zx-w@  ( zx-addr -- word )  zxa:mem:w@ ;
: zx-w!  ( val zx-addr )  zxa:mem:w! ;

;; set zxasm ORG to DP
: zx-fix-org  zx-dp^ zx-w@ zxa:org! ;

;; set DP to zxasm ORG
: zx-fix-dp  zxa:org@ zx-dp^ zx-w! ;

: zx-uservar-nallot  ( n )
  dup 1 65536 within not?error" invalid uservar size"
  dup zx-uservar-curr^ + zx-uservar-end^ u> ?error" out of uservar area"
  dup zx-uservar-curr^ << dup 0 swap zx-c! 1+ 1 under- over ?^|| else| 2drop >>
  zx-uservar-curr^ dup rot + zx-uservar-curr^:! ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; ZX dictionary allocation and data compile

: zx-here  ( -- zx-addr )  zx-dp^ zx-w@ ;

: zx-nallot ( n -- zx-start-address )
  dup 0< ?error" negative zx-allot"
  zx-dp^ zx-w@ dup >r
  + dup hi-word ?error" out of ZX memory"
  zx-dp^ zx-w! r> ;

: zx-allot ( n )  zx-nallot drop ;

: zx-c,  ( byte )  1 zx-nallot zx-c! ;
: zx-w,  ( word )  2 zx-nallot zx-w! ;

: zx-allot0 ( n )  << dup +?^| 0 zx-c, 1- |? else| drop >> ;

: zx-align-256  << zx-here lo-byte ?^| 0 zx-c, |? else| >> ;

;; compile byte-counted string
: zx-raw-bstr,  ( addr count )
  dup 0 256 within not?error" invalid zx string length"
  dup zx-c, swap
  << over ?^| c@++ zx-c, 1 under- |? else| 2drop >> ;

;; make sure that the buffer of size `n` is completely inside one 256-byte page.
;; fill unused bytes with 0.
: zx-ensure-page  ( n )
  dup 1 256 bounds not?error" invalid buffer size"
  zx-here  dup rot + 1- hi-byte  swap hi-byte = ?exit
  ;; need to align
  zx-here dup $FF or 1+ swap - zx-allot0 ;

: zx-ensure-page-with-report  ( addr count n )
  dup 1 256 bounds not?error" invalid buffer size"
  zx-here  dup rot + 1- hi-byte  swap hi-byte = ?exit< 2drop >?
  ;; need to align
  2>r zx-here dup $FF or 1+ swap -
  2r> type ."  align: " dup .bytes ."  wasted.\n"
  zx-allot0 ;

;; return number of bytes left in the current page
: zx-page-left  ( -- n )  $100 zx-here lo-byte - ;


module str-unescaper
<disable-hash>

: c,  ( ch )  zx-c, ;

;; string literals
: ?hex-digit  ( ch -- digit )
  dup [char] 0 [char] 9 bounds ?exit< [char] 0 - >?
  dup [char] A [char] F bounds ?exit< [char] A - 10 + >?
  dup [char] a [char] f bounds ?exit< [char] a - 10 + >?
  error" invalid hex escape" ;

;; addr and count already advanced past the escaped char
: decode-hex-escape  ( count addr -- count addr ch )
  over 2 < ?error" invalid hex escape" 2 under-
  c@++ ?hex-digit 16 * swap c@++ ?hex-digit rot + ;

: decode-escape  ( count addr ch -- count addr ch )
  << [char] r of?v|  7 |? ;; ROM wants it this way
     [char] n of?v| 13 |? ;; ROM wants it this way
     [char] t of?v|  6 |? ;; ROM wants it this way
     [char] c of?v|  4 |? ;; ENDCR
     \ [char] e of?v| 27 |?
     \ [char] a of?v|  7 |?
     [char] b of?v|  8 |?
     [char] W of?v|  5 |? ;; CLS
     \ [char] z of?v|  0 |?
     [char] \ of?v| [char] \ |?
     [char] ` of?v| [char] " |?
     [char] ' of?v| [char] " |?
     [char] I of?v| 16 |? ;; INK
     [char] P of?v| 17 |? ;; PAPER
     [char] F of?v| 18 |? ;; FLASH
     [char] B of?v| 19 |? ;; BRIGHT
     [char] V of?v| 20 |? ;; INVERSE
     [char] O of?v| 21 |? ;; OVER
     [char] x of?v| decode-hex-escape |?
  else| error" invalid escape" >> ;

;; decode to sl-dpos
: unescape  ( addr count )
  << dup +?^|
      1- swap c@++ dup 92 = ?< drop
        over not?error" invalid escape"
        1 under- c@++ decode-escape
      >? c, swap
  |? else| 2drop >> ;

end-module

;; compile byte-counted string (and unescape it)
: zx-bstr,  ( addr count )
  \ dup 0 256 within not?error" invalid zx string length"
  ( reserve length byte) 0 zx-c,
  zx-here >r
  str-unescaper:unescape
  zx-here r@ -
  dup 255 u> ?error" invalid zx string length"
  r> 1- zx-c! ;


: zx-incbin  ( addr count )
  file:open-r/o >r
  << here 1 r@ file:read 1 = ?^| here c@ zx-c, |?
  else| r> file:close >> ;

: <zx-asm>  zx-fix-org [\\] <asm> ;


end-module
