;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Z80 Abersoft fig-FORTH recompiled
;; Copyright (C) 2024 Ketmar Dark // Invisible Vector
;; Understanding is not required. Only obedience.
;; GPLv3 ONLY
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; some basic words
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(*
new zx article header:
  dw  lfa         ;; link to the previous LFA, or 0
  db  nlen-flags  ;; name length, flags
    name (with bit 7 set at the last char)
itc:
  dw  cfa
  pfa is here
dtc:
  db  opcode  (JP/CALL) ($C3/$CD)
  dw  dest
  pfa is here

nlen bits:
  bit 7: always set   $80
  bit 6: immediate    $40
  bit 5: smudge       $20

vocabulary format:
  cfa is "zx-do-vocab"
  dw latest-nfa
  $A081 -- dummy name
  dw voc-link -- points to the previous voc-link, or 0

last vocab word is linked to the previous vocab dummy name.
*)

extend-module TCOM

;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; ZX statistics

;; number of corresponding words
0 quan zx-stats-code
0 quan zx-stats-colon
0 quan zx-stats-var
0 quan zx-stats-const
0 quan zx-stats-uvar
0 quan zx-stats-quan
0 quan zx-stats-vect
0 quan zx-stats-alias

;; number of headers created
0 quan zx-stats-headers
;; number of words created
0 quan zx-stats-words

;; bytes wasted on headers
0 quan zx-stat-header-bytes
;; bytes wasted on CFAs
0 quan zx-stat-cfa-bytes
;; bytes wasted on machine code
0 quan zx-stat-mcode-bytes
;; bytes wasted on threaded code
0 quan zx-stat-tcode-bytes

;; number of forward references
0 quan zx-stat-forwards

0 quan zx-stat-clit-bytes-saved
0 quan zx-stat-blit-bytes-saved


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; ZX utilities

 0 constant zxc-none
-1 constant zxc-code
 1 constant zxc-colon
zxc-none quan zx-compile-mode

$40 constant zx-immediate-mask
$20 constant zx-smudge-mask

$C3 constant zx-jp-opcode
$CD constant zx-call-opcode

OPT-DTC? [IF] 3 [ELSE] 2 [ENDIF] constant zx-#cfa

: zx-traverse  ( zx-addr dir -- zx-addr )
  >r << r@ + dup zx-c@ $80 < ?^|| else| rdrop >> ;

: zx-cfa>pfa  ( zx-cfa -- zx-pfa ) zx-#cfa + ;


: zx-context@  ( -- zx-addr )  zx-context^ zx-w@ ;
: zx-current@  ( -- zx-addr )  zx-current^ zx-w@ ;

: zx-context!  ( value )  zx-context^ zx-w! ;
: zx-current!  ( value )  zx-current^ zx-w! ;

;; set in "zx-mk-header"
0 quan zx-latest-cfa  ;; always valid
0 quan zx-latest-nfa  ;; can be 0 if the current word has no header

: zx-latest-pfa  ( -- zx-pfa )  zx-latest-cfa zx-cfa>pfa ;

|: ?zx-last-nfa  ( -- zx-nfa )
  zx-latest-nfa dup not?error" last defined ZX word is hidden" ;

: zx-immediate  ?zx-last-nfa dup zx-c@ zx-immediate-mask xor swap zx-c! ;
: zx-smudge     zx-latest-nfa dup ?< dup zx-c@ zx-smudge-mask xor swap zx-c! || drop >? ;


: zx-comp!  $C0 zx-state^ zx-w! ;
: zx-exec!  0 zx-state^ zx-w! ;

: zx-comp?  ( -- flag )  zx-state^ zx-w@ 0<> ;
: zx-exec?  ( -- flag )  zx-state^ zx-w@ 0= ;

: zx-?comp  zx-comp? not?error" ZX compilation mode expected" ;
: zx-?exec  zx-exec? not?error" ZX execution mode expected" ;

: zx-?in-colon
  system:?exec zx-?comp
  zx-compile-mode zxc-colon = not?error" ZX colon word was not started yet" ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; compile call to another word, and CFA field

OPT-DTC? [IF]
: zx-resolve-cfa  ( zx-cfa -- zx-cfa )
  dup not?exit  ;; forward
  << dup zx-c@ zx-jp-opcode = ?^| 1+ zx-w@ |? else| >> ;
[ELSE]
: zx-resolve-cfa  ( zx-cfa -- zx-cfa )  ;
[ENDIF]

: zx-call,  ( zx-cfa )  zx-resolve-cfa zx-w, ;

;; compile CFA field.
;; special value: -1 means "code word CFA"
: zx-cfa,  ( zx-addr )
  zx-here >r  ;; for stats
  [ OPT-DTC? ] [IF]
    dup 1+ ?< zx-call-opcode zx-c, zx-w,
    || drop zx-jp-opcode zx-c, zx-here 2+ zx-w, >?
  [ELSE]
    dup 1+ not?< drop zx-here 2+ >? zx-w,
  [ENDIF]
  ;; statistics
  zx-here r> - zx-stat-cfa-bytes:+! ;

OPT-DTC? [IF]
: zx-jp-cfa,  ( zx-addr )
  zx-here >r  ;; for stats
  zx-jp-opcode zx-c, zx-w,
  ;; statistics
  zx-here r> - zx-stat-cfa-bytes:+! ;
[ENDIF]


*: @asm-label:  ( -- value )  \ name
  parse-name z80-labman:@get [\\] {#,} ;


end-module
