;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Z80 Abersoft fig-FORTH recompiled
;; Copyright (C) 2024 Ketmar Dark // Invisible Vector
;; Understanding is not required. Only obedience.
;; GPLv3 ONLY
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; fixup list
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


extend-module TCOM

enum{
  def: fix-literal
  def: fix-call
}

;; fixup chain item
0
new-field fix>next    -- 0: no more
new-field fix>zxaddr  -- zx address
new-field fix>type    -- lit or call
new-field fix>fline   -- source line
new-field fix>fname   -- pointer to file name nfa
constant #fixup

;; file names (for forwards)
module FILES
<case-sensitive>
end-module (private)


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; forward references

0 quan last-used-file-nfa   (private)
0 quan last-used-file-uid0  (private)
0 quan last-used-file-uid1  (private)

|: fname-last?  ( uid0 uid1 -- fname-nfa TRUE // FALSE )
  last-used-file-uid1 = swap last-used-file-uid0 = and not?exit&leave
  last-used-file-nfa true ;

|: fname-find  ( addr count -- TRUE // addr count FALSE )
  2dup vocid: files vocid-find not?exit&leave
  nrot 2drop
  inc-uidx last-used-file-uid1:! last-used-file-uid0:!
  dart:cfa>nfa dup last-used-file-nfa:! true ;

|: fname-ptr  ( -- fname-nfa )
  inc-uidx  2dup or not?error" not in include!"
  fname-last? ?exit
  inc-fname fname-find ?exit
  system:default-ffa >r <public-words>
  push-cur voc-cur: files system:mk-create-4
  r> system:default-ffa !
  inc-uidx swap
  dup last-used-file-uid0:! ,
  dup last-used-file-uid1:! ,
  system:latest-nfa pop-cur
  dup last-used-file-nfa:! ;


|: mk-fixup  ( zx-addr chain-head^ type )
  nrot fname-ptr >r
  align-here-4 here >r
  ( next) dup @ , r> swap !
  ( zx-addr) lo-word ,
  ( type) ,
  ( file line) inc-line# ,
  ( file name) r> , ;


|: dump-chain-item  ( chain^ )
  ." CITEM=$" dup .hex8
  ."  NEXT=$" dup fix>next:@ .hex8
  ."  ZXADDR=$" dup fix>zxaddr:@ .hex4
  ."  TYPE=" dup fix>type:@ base @ decimal swap 0.r base !
  ."  FLINE=" dup fix>fline:@ base @ decimal swap 0.r base !
  ."  FNAME=" dup fix>fname:@ debug:.id
  drop ;


|: fix-one  ( chain^ zx-addr )
\ endcr ." FIXING: chain=$" over .hex8 ."  zx-addr=$" dup .hex4
\ ."  type=" over fix>type:@ . ."  [zx-addr]=$" dup zx-c@ .hex2 cr
  >r fix>zxaddr:@ dup zx-w@ r> + swap zx-w! ;

|: fix-chain  ( zx-addr chain^ )
  dup not?< 2drop exit >?
\ endcr ." FIX-CHAIN: zx-addr=$" over .hex8 ."  chead=$" dup .hex8 cr
  swap >r <<  ( chain | zx-addr )
    dup ?^| \ endcr ."   ITEM:  " dup dump-chain-item cr
            dup r@ fix-one
            fix>next:@ |?
  else| drop rdrop >> ;


;; optimise primitive calls for DTC
|: fix-prim-calls  ( shadow-cfa )
  [ OPT-DTC? ] [IF]
    dup dart:cfa>pfa ss>chain:@ not?exit< drop >?
    \ endcr ." refix: " dup dart:cfa>nfa debug:.id cr
    dart:cfa>pfa ss>chain:^ << @ dup not?v||
      \ dup dump-chain-item
      dup fix>type:@ fix-call = ?<
        dup fix>zxaddr:@ zx-w@ zx-c@ zx-jp-opcode = ?<
          \ endcr ."   PRIM-FIX at $" dup fix>zxaddr:@ .hex4 cr
          dup fix>zxaddr:@ dup zx-w@ zx-resolve-cfa
          ( chain^ zx-addr zx-dest )
          swap zx-w!
        >?
      >?
    ^|| >>
  [ENDIF]
  drop ;

false quan undefined-forwards? (private)

|: dump-chain-files  ( chain^ )
  << dup ?^|
    endcr ."   used around "
    dup fix>fname:@ debug:.id
    dup fix>fline:@ [char] : emit 0.r cr
  fix>next:@ |?
  else| drop >> ;

|: check-forward  ( shadow-cfa -- res )
\ endcr ." CHECKING: \'" dup dart:cfa>nfa debug:.id
\ ." \': zxcfa=$" dup dart:cfa>pfa ss>zxcfa:@ .hex8
\ ."  chain=$" dup dart:cfa>pfa ss>chain:@ .hex8 cr
  dup dart:cfa>pfa ss>zxcfa:@ 1+ ?exit< fix-prim-calls 0 >?
  dup dart:cfa>pfa ss>chain:@ not?exit< drop 0 >?
  endcr ." TCOM: unresolved forward \'" dup dart:cfa>nfa debug:.id ." \'\n"
  dart:cfa>pfa ss>chain:@ dump-chain-files
  undefined-forwards?:!t 0 ;

@: check-forwards
  undefined-forwards?:!f
  vocid: forth-shadows ['] check-forward vocid-foreach drop
  vocid: editor-shadows ['] check-forward vocid-foreach drop
  [ OPT-SYS-VOCAB? ] [IF]
    vocid: system-shadows ['] check-forward vocid-foreach drop
  [ENDIF]
  undefined-forwards? ?error" TCOM: found some unresolved forwards" ;


end-module
