;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Z80 Abersoft fig-FORTH recompiled
;; Copyright (C) 2024 Ketmar Dark // Invisible Vector
;; Understanding is not required. Only obedience.
;; GPLv3 ONLY
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; create shadow words
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


extend-module TCOM

;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; shadow words (creation, doers, etc.)

;; zx-vocabulary sets this
0 quan zx-latest-vocid

0 quan zx-forth-vocid
0 quan zx-editor-vocid
OPT-SYS-VOCAB? [IF]
0 quan zx-system-vocid
[ENDIF]

0 quan latest-shadow-cfa


0 quan zx-default-shadow-vocid

: <zx-shadow-forth>  vocid: forth-shadows zx-default-shadow-vocid:! ;

: <zx-shadow-system>
  [ OPT-SYS-VOCAB? ] [IF] vocid: system-shadows [ELSE] vocid: forth-shadows [ENDIF]
  zx-default-shadow-vocid:! ;


: <zx-forth>   zx-forth-vocid dup zx-current! zx-context! ;
: <zx-editor>  zx-editor-vocid dup zx-current! zx-context! ;

: <zx-system>
  [ OPT-SYS-VOCAB? ] [IF] zx-system-vocid [ELSE] zx-forth-vocid [ENDIF]
  dup zx-current! zx-context! ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; forwards support mechanics

;; register fixup chain address at "zx-here", if necessary.
: ss-want-zx-any-cfa-here  ( shadow-pfa fix-type -- zx-cfa )
  >r dup ss>zxcfa:@ dup +0?< nip rdrop exit >? drop
  zx-stat-forwards:1+!
  zx-here swap ss>chain:^ r> mk-fixup 0 ;

;; compile call to shadow word zx-cfa.
;; takes care of forward fixup chain.
: ss-call,  ( shadow-pfa )
  fix-call ss-want-zx-any-cfa-here zx-call, ;

;; compile call to shadow word zx-cfa.
;; takes care of forward fixup chain.
: ss-ofs-call,  ( shadow-pfa ofs )
  swap fix-call ss-want-zx-any-cfa-here + zx-call, ;

;; compile address of shadow word.
;; takes care of forward fixup chain.
;; compiles with offset (to allow compile PFA literals, for example)
: ss-ofs,  ( shadow-pfa ofs )
  swap fix-literal ss-want-zx-any-cfa-here + zx-w, ;

;; fail on forwards
: ss-cfa-zx-cfa@  ( shadow-cfa -- zx-cfa )
  dup dart:cfa>pfa ss>zxcfa:@ dup -?< drop
    " cannot use undefined ZX forward word \'" string:>pad
    dart:cfa>nfa idcount string:pad+cc
    " \'!" string:pad+cc
    string:pad-cc@ error >?
  nip ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; shadow words support

|: (zsys)  ( addr count -- shadow-cfa )
  2dup
  [ OPT-SYS-VOCAB? ] [IF] vocid: system-shadows [ELSE] vocid: forth-shadows [ENDIF]
  vocid-find not?<
    " cannot find ZX SYSTEM word \'" string:>pad
    string:pad+cc " \'!" string:pad+cc
    string:pad-cc@ error >?
  nrot 2drop ;

*: zsys:  ( -- shadow-cfa ) \ name
  parse-name (zsys)
  [\\] {#,} ;

*: zsys-run:  \ name
  parse-name (zsys) \, ;


: curr-shadow-vocid  ( -- vocid )
  zx-current@ <<
    0 of?v| zx-default-shadow-vocid |?
    zx-forth-vocid of?v| vocid: forth-shadows |?
    zx-editor-vocid of?v| vocid: editor-shadows |?
    [ OPT-SYS-VOCAB? ] [IF]
      zx-system-vocid of?v| vocid: system-shadows |?
    [ENDIF]
  else| drop zx-default-shadow-vocid >> ;


;; doers are used to check word kind
: ss-forth-doer     ( pfa )  system:?exec zx-?comp ss-call, ;
: ss-code-doer      ( pfa )  system:?exec zx-?comp ss-call, ;
: ss-variable-doer  ( pfa )  system:?exec zx-?comp ss-call, ;
: ss-2variable-doer ( pfa )  system:?exec zx-?comp ss-call, ;
: ss-uservar-doer   ( pfa )  system:?exec zx-?comp ss-call, ;
: ss-does-doer      ( pfa )  system:?exec zx-?comp ss-call, ;
: ss-vocab-doer     ( pfa )  system:?exec zx-?comp ss-call, ;
: ss-alias-doer     ( pfa )  system:?exec zx-?comp ss-call, ;

OPT-ENABLE-FP? [IF]
: ss-fp-doer        ( pfa )  system:?exec zx-?comp ss-call, ;
: ss-fp-xmem-doer   ( pfa )  system:?exec zx-?comp ss-call, ;
[ENDIF]

: ss-constant-doer ( pfa )
  zx-comp? ?exit< system:?exec ss-call, >?
  ss>zxcfa:@ zx-cfa>pfa zx-w@ [\\] {#,} ;

: ss-quan-doer ( pfa )  ss-constant-doer ;
: ss-vect-doer ( pfa )  system:?exec zx-?comp ss-call, ;

#ss
new-field ss>quan-scfa
drop

module zx-quan-mth
<disable-hash>

\ |: zx-quan-doer  ( pfa )  ss-constant-doer ;

|: ?zx-quan  ( -- shadow-cfa )
  system:?exec
  ['] ss-quan-doer system:vocobject-this?
  ['] ss-vect-doer system:vocobject-this?
  or not?error" not a quan/vect"
  system:vocobject-this
  ss>quan-scfa:@ ;
end-module


: zx-checker:  ( shadow-cfa )  \ name
  <builds ,
 does> ( cfa pfa )
  over system:does? not?exit< 2drop false >?
  @ swap system:doer@ = ;

['] ss-forth-doer zx-checker: zx-forth?
['] ss-code-doer zx-checker: zx-code?
['] ss-constant-doer zx-checker: zx-constant?
['] ss-quan-doer zx-checker: zx-quan?
\ ['] ss-quan-mth::zx-quan-doer zx-checker: zx-quan?
['] ss-vect-doer zx-checker: zx-vect?
['] ss-variable-doer zx-checker: zx-variable?
['] ss-2variable-doer zx-checker: zx-2variable?
['] ss-uservar-doer zx-checker: zx-uservar?
['] ss-vocab-doer zx-checker: zx-vocab?
['] ss-does-doer zx-checker: zx-does?
['] ss-alias-doer zx-checker: zx-alias?
OPT-ENABLE-FP? [IF]
['] ss-fp-doer zx-checker: zx-fp?
['] ss-fp-xmem-doer zx-checker: zx-fp-xmem?
[ENDIF]

: zx-forward?  ( shadow-cfa -- flag )
  dup system:does? not?exit< drop false >?
  \ FIXME: better checks!
  dart:cfa>pfa ss>zxcfa:@ 0< ;


;; define undefined forward value
|: ss-define-forward  ( zx-addr shadow-cfa )
  over hi-word ?error" wut?!"  -- just in case
  dup dart:cfa>pfa ss>zxcfa:@ +0?<
    " duplicate definition of forward \'" string:>pad
    dart:cfa>nfa idcount string:pad+cc
    " \'!" string:pad+cc
    string:pad-cc@ error >?
\ endcr ." *** FIX FORWARD \'" dup dart:cfa>nfa debug:.id ." \'\n"
  dart:cfa>pfa
  2dup ss>zxcfa:!
  ss>chain:@ fix-chain ;

|: check-existing-shadow  ( zx-cfa addr count -- zx-cfa addr count FALSE // TRUE )
  2dup curr-shadow-vocid vocid-find not?exit&leave
  dup ss>zxcfa:@ +0?< drop
    " duplicate definition of forward \'" string:>pad
    string:pad+cc " \'!" string:pad+cc
    string:pad-cc@ error >?
  nrot 2drop  ;; drop addr and count
  dup latest-shadow-cfa:!
  over ?< ss-define-forward || 2drop >?
  true ;

: (mk-shadow-header)  ( zx-cfa addr count doer )
  >r check-existing-shadow ?exit< rdrop >?
  push-cur curr-shadow-vocid current! system:mk-builds
  system:latest-cfa latest-shadow-cfa:!
  ( zx-cfa) ,
  ( chain) 0 ,
  system:latest-cfa pop-cur r> system:!doer ;

: mk-shadow-header  ( addr count doer )
  >r zx-here nrot r> (mk-shadow-header) ;


;; create forward reference to Forth word
: ss-mk-forth-forward  ( addr count -- shadow-cfa )
  -1 nrot ['] ss-forth-doer (mk-shadow-header)
  curr-shadow-vocid system:vocid-latest dart:lfa>cfa ;


;; automatically creates Forth forward
: (-zx-find-ss-cfa)  ( -- shadow-cfa )
  parse-name
  2dup curr-shadow-vocid vocid-find ?exit< nrot 2drop >?
  2dup vocid: forth-shadows vocid-find ?exit< nrot 2drop >?
  ss-mk-forth-forward ;

;; doesn't allow forwards
: (-zx-find-ss)  ( -- shadow-cfa zx-cfa )
  (-zx-find-ss-cfa) dup ss-cfa-zx-cfa@ ;

;; doesn't allow forwards
: (-zx-find)  ( -- zx-cfa )
  (-zx-find-ss) nip ;

;; doesn't create Forth forward
: (zx-find-ss-cfa-no-forward)  ( addr count -- shadow-cfa TRUE // FALSE )
  2dup curr-shadow-vocid vocid-find ?exit< nrot 2drop true >?
  vocid: forth-shadows vocid-find ;


;; this is used to search for ZX constants
module zx-consts
end-module

|: (voc-zx-const-finder)  ( addr count skip-hidden? vocid -- cfa TRUE // FALSE )
  2drop (zx-find-ss-cfa-no-forward) not?exit&leave
  dup zx-constant? over zx-quan? or not?exit< drop false >?
  true ;
['] (voc-zx-const-finder) vocid: zx-consts system:vocid-find-cfa!

|: (voc-zx-const-execcomp)  ( cfa -- ... TRUE // cfa FALSE )
  dup zx-constant? over zx-quan? or not?error" wut?! (not a ZX constant)"
\ endcr dup dart:cfa>nfa debug:.id cr
  execute true ;
['] (voc-zx-const-execcomp) vocid: zx-consts system:vocid-execcomp-cfa!


end-module
