;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Z80 Abersoft fig-FORTH recompiled
;; Copyright (C) 2024 Ketmar Dark // Invisible Vector
;; Understanding is not required. Only obedience.
;; GPLv3 ONLY
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; create zx headers
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


extend-module TCOM

;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; create ZX word header

true quan zx-headers-allowed?

: <zx-hidden>  zx-headers-allowed?:!f ;
: <zx-normal>  zx-headers-allowed?:!t ;

|: error-bad-name  ( addr count )
  0 max 1024 min
  " invalid ZX word name: \'" string:>pad string:pad+cc
  " \'!" string:pad+cc
  string:pad-cc@ error ;

|: ?good-zx-name  ( addr count )
  dup 0 32 within not?< error-bad-name >?
  2dup swap <<
    over ?^| c@++ 32 128 within not?< 2drop error-bad-name >?
            1 under- |?
  else| 4drop >> ;

;; max length is 31
|: zx-hash-name  ( addr count -- u8 )
  $1F and swap 0 >r
  << over ?^| c@++ $7F and ( $20 or) string:upchar
              r> 1 rol8 swap - lo-byte >r 1 under- |? else| 2drop >>
  r> ;

|: (zx-put-wname)  ( addr count )
  dup $80 or zx-c,
  swap << over ?^| c@++ string:upchar zx-c, 1 under- |? else| 2drop >>
  zx-here 1- dup zx-c@ $80 or swap zx-c! ;

;; "{00}" is the special name -- word with zero name byte. yes, it is used.
|: zx-{00}-name?  ( addr count -- flag )
  4 = not?exit< drop false >?
  @ $7D_30_30_7B = ;

;; up to CFA.
: zx-mk-header  ( addr count )
  zx-stats-words:1+!
  2dup zx-{00}-name? ?< 2drop " \x00" || 2dup ?good-zx-name >?
  zx-headers-allowed? not?exit< 2drop
    zx-here zx-latest-cfa:!
    zx-latest-nfa:!0 >?
\ endcr ." ZX: new header \'" 2dup type ." \' at $" zx-here .hex4 cr
  zx-here >r  ;; for stats
  [ OPT-NEW-ART-HEADER? ] [IF]
    [ OPT-NEW-ART-HEADER-HASHED? ] [IF]
\ endcr ." ZX: new header \'" 2dup type ." \' at $" zx-here .hex4 ."  hash=$" 2dup zx-hash-name .hex2 cr
      2dup zx-hash-name zx-c,
    [ENDIF]
    ( lfa) zx-here zx-current@ zx-w@ zx-w, zx-current@ zx-w!
  [ENDIF]
  zx-here zx-latest-nfa:! (zx-put-wname)
  [ OPT-NEW-ART-HEADER? ] [IFNOT]
    ( lfa) zx-latest-nfa zx-current@ zx-w@ zx-w, zx-current@ zx-w!
  [ENDIF]
  zx-here zx-latest-cfa:!
  ;; update stats
  zx-here r> - zx-stat-header-bytes:+!
  zx-stats-headers:1+! ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; branch chain compilers/resolvers

OPT-REL-BRANCH? [IF]
: zx-brdest@    ( zx-addr )  dup zx-w@ + lo-word ;
: zx-brdest!    ( zx-dest zx-addr )  tuck - swap zx-w! ;
[ELSE]
: zx-brdest@    ( zx-addr )  zx-w@ ;
: zx-brdest!    ( zx-dest zx-addr )  zx-w! ;
[ENDIF]

: zx-brdest,  ( zx-dest )  zx-here 0 zx-w, zx-brdest! ;

: zx-<mark    ( -- zx-dest )  zx-here ;
: zx-chain>   ( chain -- chain-new )  zx-here swap zx-w, ;
: zx-mark>    ( -- chain-new )  0 zx-chain> ;
: zx-<resolve ( zx-dest )  zx-brdest, ;

: zx-resolve> ( chain )
  dup not?exit< drop >? zx-here >r
  << dup zx-w@ r@ rot zx-brdest! dup ?^|| else| drop rdrop >> ;


end-module
