;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Z80 Abersoft fig-FORTH recompiled
;; Copyright (C) 2024 Ketmar Dark // Invisible Vector
;; Understanding is not required. Only obedience.
;; GPLv3 ONLY
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; create other ZX words
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


extend-module TCOM

;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; create constants, vars, etc.

OPT-ENABLE-FP? [IF]
*: zx-fp:  ( cmd )  \ name
  system:?exec zx-?exec
  parse-name
  2dup 2>r zx-mk-header
  2r> ['] ss-fp-doer mk-shadow-header
  ( cfa) zx-do-fp^ zx-cfa,
  ( cmd) zx-c,
  zx-fix-org ;

*: zx-fp-xmem:  ( cmd )  \ name
  system:?exec zx-?exec
  parse-name
  2dup 2>r zx-mk-header
  2r> ['] ss-fp-xmem-doer mk-shadow-header
  ( cfa) zx-do-fp-xmem^ zx-cfa,
  ( cmd) zx-c,
  zx-fix-org ;
[ENDIF]

: zx-mk-constant  ( value addr count )
  system:?exec zx-?exec
  2dup 2>r zx-mk-header
  2r> ['] ss-constant-doer mk-shadow-header
  ( cfa) zx-do-constant^ zx-cfa,
  ( value) zx-w,
  zx-stats-const:1+!
  zx-fix-org ;

*: zx-constant  ( value )  \ name
  parse-name
  zx-mk-constant ;

*: zx-label  ( value )  \ name
  system:?exec
  parse-name z80-labman:define ;

*: zx-constant-and-label  ( value )  \ const-name label-name
  dup [\\] zx-constant [\\] zx-label ;

*: zx-create  \ name
  system:?exec zx-?exec
  parse-name
  2dup 2>r zx-mk-header
  2r> ['] ss-variable-doer mk-shadow-header
  ( cfa) zx-do-variable^ zx-cfa,
  zx-fix-org ;

(* ZX QUAN/VECT methods (after PFA):
  QADDR
  TO
  +TO
*)
*: zx-quan  ( value )  \ name
  system:?exec zx-?exec
  parse-name
  2dup 2>r zx-mk-header
  2r> ['] ss-quan-doer mk-shadow-header
  vocid: zx-quan-mth latest-shadow-cfa dart:cfa>vocid !
  ( shadow-cfa in the host) latest-shadow-cfa ,
  ( cfa) zx-do-quan^ zx-cfa,
  ( value) zx-w,
  ( addr-cfa) zx-do-quan-addr^ zx-cfa,
  ( to-cfa) zx-do-quan-to^ zx-cfa,
  ( +to-cfa) zx-do-quan-add-to^ zx-cfa,
  zx-stats-quan:1+!
  zx-fix-org ;

*: zx-vect  ( value )  \ name
  system:?exec zx-?exec
  parse-name
  2dup 2>r zx-mk-header
  2r> ['] ss-vect-doer mk-shadow-header
  vocid: zx-quan-mth latest-shadow-cfa dart:cfa>vocid !
  ( shadow-cfa in the host) latest-shadow-cfa ,
  ( cfa) zx-do-vect^ zx-cfa,
  ( value) zx-w,
  ( addr-cfa) zx-do-quan-addr^ zx-cfa,
  ( to-cfa) zx-do-quan-to^ zx-cfa,
  ( +to-cfa) zx-do-quan-add-to^ zx-cfa,
  zx-stats-vect:1+!
  zx-fix-org ;

*: zx-variable  ( value )  \ name
  [\\] zx-create
  ( value) zx-w,
  zx-stats-var:1+!
  zx-fix-org ;

*: zx-ud-buffer  ( size )  \ name
  system:?exec zx-?exec
  parse-name
  2dup 2>r zx-mk-header
  2r> ['] ss-variable-doer mk-shadow-header
  ( cfa) zx-do-variable^ zx-cfa,
  zx-allot
  zx-fix-org ;

*: zx-2variable  ( value )  \ name
  system:?exec zx-?exec
  parse-name
  2dup 2>r zx-mk-header
  2r> ['] ss-2variable-doer mk-shadow-header
  ( cfa) zx-do-2variable^ zx-cfa,
  ( value) dup zx-cfa, hi-word zx-cfa,
  zx-fix-org ;

*: zx-uservar  ( value )  \ name
  system:?exec zx-?exec
  parse-name
  2dup 2>r zx-mk-header
  2r> ['] ss-uservar-doer mk-shadow-header
  ( cfa) zx-do-uservar^ zx-cfa,
  2 zx-uservar-nallot
  ( offset) dup zx-uservar-start^ - zx-c,
  zx-w! ;; set value
  zx-stats-uvar:1+!
  zx-fix-org ;

*: zx-uvar-ofs  ( -- value ) \ name
  system:?exec
  parse-name 2dup vocid: forth-shadows vocid-find not?<
    [ OPT-SYS-VOCAB? ] [IF] vocid: system-shadows vocid-find [ELSE] 2drop false [ENDIF]
    not?error" unknown uservar"
  || nrot 2drop >?
  dup zx-uservar? not?error" not a uservar"
  dart:cfa>pfa ss>zxcfa:@ dup 1+ not?error" undefined uservar"
  zx-cfa>pfa zx-c@ ;

*: zx-vocabulary  \ name
  system:?exec zx-?exec
  parse-name
  2dup 2>r zx-mk-header
  2r> ['] ss-vocab-doer mk-shadow-header
  ( cfa) zx-do-vocab^ zx-cfa,
  [ OPT-NEW-ART-HEADER? ] [IF]
    zx-here zx-latest-vocid:!
    ( latest-lfa) zx-current@ zx-w,
  [ELSE]
    ( reserved) 0 zx-w,
  [ENDIF]
  ( dummy name) $A081 zx-w,
  [ OPT-NEW-ART-HEADER? ] [IFNOT]
    zx-here zx-latest-vocid:!
    ( latest-nfa) zx-current@ 2- zx-w,
  [ENDIF]
  ( voc-link) zx-here zx-voclink^ zx-w@ zx-w, zx-voclink^ zx-w!
  zx-fix-org ;


*: zx-[']  ( -- zx-cfa )  \ name
  (-zx-find) [\\] {#,} ;

*: zx-['pfa]  ( -- zx-pfa )  \ name
  (-zx-find) zx-cfa>pfa [\\] {#,} ;

|: (zx-exec-in-system)  ( cfa )
  zx-current@ >r zx-context@ >r <zx-system>
  execute r> zx-context! r> zx-current! ;

*: zx-sys[']  ( -- zx-cfa )  \ name
  ['] zx-['] (zx-exec-in-system) ;

*: zx-sys['pfa]  ( -- zx-pfa )  \ name
  ['] zx-['pfa] (zx-exec-in-system) ;


0 quan (zals-scfa)

;; WARNING! doesn't copy "immediate" bit!
*: zx-alias-for  \ oldname IS newname
  system:?exec
  zx-?exec
  (-zx-find-ss) swap
  dup (zals-scfa):!
  dup zx-alias? ?error" cannot create alias to alias yet (FIXME)"
  zx-code? ?< [ OPT-DTC? ] [IF] 1+ [ENDIF] zx-w@ false || zx-do-alias^ true >?
  ( orig-cfa do-alias TRUE // code-cfa FALSE )
  parse-name 2 = swap w@ $20_20 or $73_69 = and not?error" `IS` expected"
  parse-name
  [ 0 ] [IF]
    endcr ." NEW ZX ALIAS FOR \'" (zals-scfa) dart:cfa>nfa debug:.id ." \'"
    (zals-scfa) zx-code? ?< ."  [CODE]" || ."  [FORTH]" >?
    ."  is \'" 2dup type ." \'\n"
  [ENDIF]
  2dup 2>r zx-mk-header
  2r> ['] ss-alias-doer mk-shadow-header
  ( cfa)
  ?< zx-cfa, zx-w,
  || [ OPT-DTC? ] [IF] zx-jp-cfa, [ELSE] zx-cfa, [ENDIF] >?
  zx-stats-alias:1+!
  zx-fix-org ;


end-module
