;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Z80 Abersoft fig-FORTH recompiled
;; Copyright (C) 2024 Ketmar Dark // Invisible Vector
;; Understanding is not required. Only obedience.
;; GPLv3 ONLY
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; simple ZX decompiler
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


extend-module TCOM



module decompiler
<disable-hash>

0 quan forth-name-cfa (private)
false quan forth-prim-call (private)

OPT-DTC? [IF]
|: forth-name-checker  ( zx-cfa shadow-cfa -- zx-cfa res )
  2dup dart:cfa>pfa ss>zxcfa:@
  ( zx-cfa shadow-cfa zx-cfa zx-shadow-cfa )
  2dup = ?exit< 2drop forth-prim-call:!f forth-name-cfa:! 1 >?
  3 + = ?exit< forth-prim-call:!t forth-name-cfa:! 1 >?
  drop 0 ;
[ELSE]
|: forth-name-checker  ( zx-cfa shadow-cfa -- zx-cfa res )
  2dup dart:cfa>pfa ss>zxcfa:@ = ?< forth-name-cfa:! 1 || drop 0 >? ;
[ENDIF]

: zx-forth-name  ( zx-addr -- addr count )
  forth-name-cfa:!0 forth-prim-call:!f
  dup not?exit< forth-prim-call:!f " <forward>" >?
  vocid: forth-shadows ['] forth-name-checker vocid-foreach nip
  ?< forth-name-cfa dart:cfa>nfa idcount || " <???>" >? ;


0 quan zx-addr
0 quan zx-addr-start
0 quan zx-addr-end

: cfetch  ( -- ubyte )  zx-addr zx-c@ zx-addr:1+! ;
: wfetch  ( -- word )   cfetch cfetch 256 * or ;
: brfetch ( -- bdest )  zx-addr zx-brdest@ 2 zx-addr:+! ;

: .lit  ( n )
  dup .
  dup ." [$" .hex4 ." ]"
  dup $8000 < ?exit< drop >?
  ."  (" $FF_FF_00_00 or 0.r ." )" ;

: op-wlit  bl emit wfetch .lit ;
: op-clit  bl emit cfetch .lit ;
: op-blit  bl emit cfetch $FF00 or .lit ;

: op-s8lit
  ."  \'" cfetch
  << dup ?^| 1-
      cfetch
      dup 34 = ?< drop ." \\'"
      || dup 32 127 within ?< emit
      || ." \\x" .hex2 >? >? |?
  else| drop >> ." \'" ;

: op-branch  brfetch ."  $" .hex4 ;


;; word call? for DTC, address may point after CFA
: wcall?  ( zx-call-cfa zx-shadow-cfa -- flag )
  [ OPT-DTC? ] [IF] 2dup = nrot 3 + = or
  [ELSE] =
  [ENDIF] ;

enum{
  def: cfa-normal ;; should be first!
  def: cfa-exit
  def: cfa-branch
  def: cfa-log-branch
  def: cfa-loop
  def: cfa-wlit
  def: cfa-clit
  def: cfa-blit
  def: cfa-s8lit
  def: cfa-forward
}

create zxcfa-list
  zx-sys['] BRANCH  , cfa-branch ,
  zx-sys['] 0BRANCH , cfa-log-branch ,
  zx-sys['] TBRANCH , cfa-log-branch ,
  zx-sys['] (FOR)   , cfa-log-branch ,
  zx-sys['] (LOOP)  , cfa-loop ,
  zx-sys['] (+LOOP) , cfa-loop ,
  zx-sys['] LIT     , cfa-wlit ,
  OPT-CLIT? [IF]
  zx-sys['] CLIT    , cfa-clit ,
  [ENDIF]
  OPT-BLIT? [IF]
  zx-sys['] BLIT    , cfa-blit ,
  [ENDIF]
  zx-sys['] (.")    , cfa-s8lit ,
  zx-sys['] ;S      , cfa-exit ,
  zx-['] EXIT       , cfa-exit ,
  zx-['] QUIT       , cfa-exit ,
  zx-['] ABORT      , cfa-exit ,
  zx-['] ERROR      , cfa-exit ,
  zx-['] (ABORT)    , cfa-exit ,
  zx-['] COLD       , cfa-exit ,
  zx-['] WARM       , cfa-exit ,
0 , create;

: zxcfa-classify  ( zx-call-cfa -- class )
  dup not?exit< drop cfa-forward >?
  >r zxcfa-list << ( list^ | zx-call-cfa )
    @++
    dup not?v| 2drop rdrop cfa-normal |?
    r@ swap wcall? ?v| @ rdrop |?
  ^| 4 + | >> ;


0 quan highest-jump

|: trace-end-addr?  ( -- end-flag )
  zx-addr highest-jump u> ;

|: trace-branch  ( -- brdest )
\ endcr ." BRX at $" zx-addr .hex4 cr
  brfetch dup highest-jump max highest-jump:! ;


|: trace-cfa-normal  ( -- done-flag )  false ;
|: trace-cfa-exit    ( -- done-flag )  trace-end-addr? ;
|: trace-cfa-branch  ( -- done-flag )
  trace-branch zx-addr u< ?< trace-end-addr? || false >? ;

|: trace-cfa-log-branch  ( -- done-flag )  trace-branch drop false ;
|: trace-cfa-loop        ( -- done-flag )  trace-branch drop false ;

|: trace-cfa-wlit  ( -- done-flag )  wfetch drop false ;
|: trace-cfa-clit  ( -- done-flag )  cfetch drop false ;
|: trace-cfa-blit  ( -- done-flag )  cfetch drop false ;
|: trace-cfa-s8lit ( -- done-flag )  cfetch zx-addr:+! false ;

|: trace-cfa-forward  ( -- done-flag )  true ;

;; must be synced with CFA classes
;; handler:  ( -- done-flag )
create trace-handlers
  ['] trace-cfa-normal ,
  ['] trace-cfa-exit ,
  ['] trace-cfa-branch ,
  ['] trace-cfa-log-branch ,
  ['] trace-cfa-loop ,
  ['] trace-cfa-wlit ,
  ['] trace-cfa-clit ,
  ['] trace-cfa-blit ,
  ['] trace-cfa-s8lit ,
  ['] trace-cfa-forward ,
create;

: trace-end
  zx-addr-start zx-addr:!
  zx-addr highest-jump:!
  << wfetch zxcfa-classify trace-handlers dd-nth @execute not?^|| else| >>
  zx-addr zx-addr-end:! ;


: .one
  endcr zx-addr .hex4 ." : "
  wfetch
  dup zx-forth-name forth-prim-call ?< ." PCALL " >? type
\ bl emit bl emit ."  CFA=$" dup .hex4 ."  class=" dup zxcfa-classify .
\ ." 0BRN=$" zx-sys['] 0BRANCH .hex4 3 bl #emit
  zxcfa-classify
  << cfa-branch of?v| op-branch |?
     cfa-log-branch of?v| op-branch |?
     cfa-loop of?v| op-branch |?
     cfa-wlit of?v| op-wlit |?
     cfa-clit of?v| op-clit |?
     cfa-blit of?v| op-blit |?
     cfa-s8lit of?v| op-s8lit |?
     cfa-forward of?v||
   else| drop >> cr ;


: (see)  ( cfa )
  [ OPT-DTC? ] [IF]
    dup 1+ zx-w@ zx-do-colon^ = not?error" not a Forth word"
    3 +  ;; move to PFA
  [ELSE]
    dup zx-w@ zx-do-colon^ = not?error" not a Forth word"
    2+  ;; move to PFA
  [ENDIF]
  zx-addr-start:! trace-end
  endcr ." === START:$" zx-addr-start .hex4 ."  END:$" zx-addr-end .hex4 ."  ===\n"
  zx-addr-start zx-addr:!
  << zx-addr zx-addr-end u< ?^| .one |? else| >> ;

end-module

: zx-see  \ name
  (-zx-find) decompiler:(see) ;

end-module
